/***************************************************************************
 *
 * Copyright (c) 2002 Promise Technology, Inc.  All rights reserved.
 * No part of this document may be reproduced or transmitted in any form or
 * by any means, electronic or mechanical, for any purpose, without the
 * express written permission of Promise Technology, Inc.
 *
 *
 * cam_ata.c - CAM ATA module functions
 *
 ***************************************************************************/
#include "cam_con.h"
#include "cam_def.h"
#include "cam_imp.h"
#include "cam_macr.h"
#include "cam_gb.h"
#include "cam_exp.h"

U8 IssueIdentify(U8 bChannelNumber, U8 bID, PU8 pBuf, U8 command);
U8 CAMInitDevice(PDEV_CONFIG pDevConfig);
void prepareCamPacket(PCAM_PACKET CAM_Packet,PCAM_ATA_PACKET CAM_ATAPacket,PCAM_ATA_SG CAM_ATASG,PATA_TASK pATATask);
U8 WaitOnBusy(PTR_SIZE IDEBaseAddress);
U8 WaitOnBaseBusy(PTR_SIZE IDEBaseAddress);
void AtapiSoftReset(U8 bChannelNumber, U8 DeviceNumber);
U8 AtaSoftReset(U8 bChannelNumber, U8 DeviceNumber);
U8 AtaSoftReset_PM(U8 bChannelNumber, U8 DeviceNumber);
U8 WaitForDrq(PTR_SIZE IDEBaseAddress);
U8 WaitOnReady(U8 bChannelNumber);
U8 WaitDelayOnReady(U8 bChannelNumber, U32 delay);
U8 WaitForINT(PTR_SIZE IDEBaseAddress);
U8 WaitForATAPIINT(U8 bChannelNumber);
void prepareATAPacket(PCAM_ATA_PACKET CAM_ATAPacket, PATA_TASK pATATask);

/* for polling device ready */
void CAM_InitATAChnl_CallBack(U8 bChannelNumber, U8 status);
void CAM_Reset_ReinitATA_CallBack(U8 bChannelNumber, U8 status);
void ATA_Timeout_CallBack(U8 bChannelNumber, U8 status);
void CAM_ResetATAChannel_CallBack(U8 bChannelNumber, U8 status);
void ATA_Polling_Channel(U8 bChannelNumber);

void InternalATAReturn(PATA_TASK pATATask)
{
	return;
}

U8 CAM_ConfigATADevice(PDEV_CONFIG pDevConfig)
{
	U8 			bChannelNumber=CAM_ATA_Chnl_InMAP(pDevConfig->bCh);
	U8			bID=pDevConfig->bID;
	PDEV_CONFIG pCurDevConfig = &gDrvConfig[bChannelNumber][bID];
	PATA_TASK pATATask = &InternalATATask[bChannelNumber];
	
	PCAM_ATA_PACKET CAM_ATAPacket;
	/*U8 Adapter_ID = bChannelNumber/MAX_ATA_CHNL;*/
	
	if (bID > MAX_ID_PER_CHNL)
		return(camFAIL);
		
	if (!(pCurDevConfig->DevFlag & DEV_EXIST))
		return(camFAIL); /* no drive */
		
	if( ( gChnlConfig[bChannelNumber].CableType < CABLE_80 )
			  ||(  gChnlConfig[bChannelNumber].ChnlSpeed < CHNL_U100 ) )	/* the channle support upto Ultra DMA mode 2 */
	{
		if (pDevConfig->UDMAMode > 2)
			pDevConfig->UDMAMode = 2;
	}
			
	if (  gChnlConfig[bChannelNumber].ChnlSpeed < CHNL_U133 )	/* the channel support upto Ultra DMA mode 5 */
	{
		if (pDevConfig->UDMAMode > 5)
			pDevConfig->UDMAMode = 5;
	}		
		
	if (pCurDevConfig->UDMAMode != 0xFF) 
	{
		if (pCurDevConfig->UDMAMode == pDevConfig->UDMAMode)
			return (camSUCCESS);	/* no change of udma mode */
	}
	else if (pCurDevConfig->MDMAMode != 0xFF) 
	{
		if (pCurDevConfig->MDMAMode == pDevConfig->MDMAMode)
			return (camSUCCESS);	/* no change of dma mode(not support udma mode) */		
	}
			
	if (pCurDevConfig->UDMAMode != 0xFF) 
		pCurDevConfig->UDMAMode = pDevConfig->UDMAMode;	
		
	if (pCurDevConfig->MDMAMode != 0xFF) 
		pCurDevConfig->MDMAMode = pDevConfig->MDMAMode;	
	
	if ( (pCurDevConfig->UDMAMode != 0xFF) || (pCurDevConfig->MDMAMode != 0xFF) )
	{
		/* set feature mode for DMA/UDMA */
		pATATask->bCh = pDevConfig->bCh;
		pATATask->bID = bID;
		pATATask->bCh2Chip = bChannelNumber;
		pATATask->callback = InternalATAReturn;		

		pATATask->ATACmdFlag = ATA_DEVICE;
		pATATask->Cmd.ATACmd.bCmd = IDE_COMMAND_SET_FEATURES;
		pATATask->Cmd.ATACmd.bCmd2Chip = pATATask->Cmd.ATACmd.bCmd;
		pATATask->Cmd.ATACmd.bFeature=0x03;

		if ( pCurDevConfig->UDMAMode != 0xFF )
			pATATask->Cmd.ATACmd.bCount = (U8)(pCurDevConfig->UDMAMode | 0x40);
		else
			pATATask->Cmd.ATACmd.bCount = (U8)(pCurDevConfig->MDMAMode | 0x20);
		
		pATATask->Cmd.ATACmd.LBAl = 0;
		pATATask->Cmd.ATACmd.LBAh = 0;
	

		/* get free packet address */
		CAM_ATAPacket = (PCAM_ATA_PACKET)CAMAllocATAPacket(bChannelNumber);

		if (CAM_ATAPacket==(PVOID)PTR_NULL){	/* no free ata packet. */
			return(camFAIL);
		}	


		/* prepare ata packet */
		prepareATAPacket(CAM_ATAPacket, pATATask);
	
		/* put the atapacket to the haed of ata queue */
		Put2ATAQueue(CAM_ATAPacket,QUEUEHead);

		/* start ata module */
		StartATAOperation(bChannelNumber);
	}

	return (camSUCCESS);
}

/* this function will do AtaSoftReset to Port_Multiplier, if Port_Multiplier exist , it will update status to host. */
U8 CAM_SATA_Port_Multiplier_Update_status(U8 bChannelNumber)
{
	U8 	bCh =  bChannelNumber%MAX_ATA_CHNL;
	U8	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;
	U32 ldata;
	PTR_SIZE	BaseExpIOAddress = ATA_ExpBaseAddr[bChannelNumber];

	/* check SATA connect well */
	ReadPortUlong( HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG, ldata );
	ldata >>= bCh;
	if (!(ldata & 0x100))
		return (camFAIL);

	/* 1. SRST=1 then 0  */
	/* issue a software reset with the PM Port field set to the control port.*/
	/* PM port : Control */
	WriteIndexUlong(BaseExpIOAddress, iNDEXeXP_PM_Port, PM_control_Port);
	if (AtaSoftReset_PM(bChannelNumber, 0) != camSUCCESS) {
		WriteIndexUlong(BaseExpIOAddress, iNDEXeXP_PM_Port, 0);
		return (camFAIL);
	}
	return (camSUCCESS);
}

U8 CAM_Init_SATA_Port_Multiplier(U8 bChannelNumber)
{
	U8	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;
	U8 	Status;
	U8 	deviceNumber,i;
	U16	wdata1,wdata2;
	U32	j,k;
	U8	deviceFound=FALSE;
	PTR_SIZE	BaseIOAddress = ATABaseAddr[bChannelNumber];
	PTR_SIZE	BaseExpIOAddress = ATA_ExpBaseAddr[bChannelNumber];

	/*
	1. SRST=1 then 0 and get PM's signature
	2. Read PM (PM Port=F, PortNum=F) and get the info of
	   "number of exposed device fan-out ports" from GSCR[2]
	3. Write PM (PM Port=F, PortNum=0) and RegNum=02h,
	   Value=00000001h to reset device port 0.
	4. Write PM (PM Port=F, PortNum=0) and RegNum=02h, 
	   Value=00000000h to make Phy Ready on device port 0.
	5. Read PM (PM Port=F, PortNum=0) and RegNum=00h,
	   if the response is Values=00000113h, 00000123h, or an
	   indication of PhyRdy, determines the active connection
	   with a drive, and move on to 6. Otherwise, goto 8 for
	   port 1.
	6. Write PM (PM Port=F, PortNum=0) and RegNum=01h
	   with Values=FFFFFFFFh to clear errors to all zeros,
	   especially the "X" bit. If X=1, any FIS is disallowed to the
	   device port. This completes the enabling of device port 0.
	7. SRST=1 then 0 (PM Port=0 in a RegHD FIS) and get the
	   drive's signature. Then you can do Identify Device and so
	   on.
	*/	
	
	/* check SATA connect well */
	/* 1. SRST=1 then 0 and get PM's signature */
	/* issue a software reset with the PM Port field set to the control port.*/
	/* PM port : Control */
	if (CAM_SATA_Port_Multiplier_Update_status(bChannelNumber) != camSUCCESS) {
		WriteIndexUlong(BaseExpIOAddress, iNDEXeXP_PM_Port, 0);
		return (camFAIL);
	}
	
	/* check Port Multiplier Signature */
	Status = ReadIndexUchar(BaseIOAddress, iNDEXIDEStatus);	
	if ( Status & (IDE_STATUS_BUSY|IDE_STATUS_ERROR) ) {
		WriteIndexUlong(BaseExpIOAddress, iNDEXeXP_PM_Port, 0);
		return(camFAIL);
	}
	/* Cylinder High : 0x96 */
	Status = ReadIndexUchar(BaseIOAddress, iNDEXIDECylinderHigh);	
	if ( Status != 0x96 ) {
		WriteIndexUlong(BaseExpIOAddress, iNDEXeXP_PM_Port, 0);
		return(camFAIL);
	}
	
	/* Cylinder Low : 0x69 */
	Status = ReadIndexUchar(BaseIOAddress, iNDEXIDECylinderLow);	
	if ( Status != 0x69 ) {
		WriteIndexUlong(BaseExpIOAddress, iNDEXeXP_PM_Port, 0);
		return(camFAIL);
	}
	/* Sector Number : 0x01 */
	Status = ReadIndexUchar(BaseIOAddress, iNDEXIDESectorNumber);	
	if ( Status != 0x01 ) {
		WriteIndexUlong(BaseExpIOAddress, iNDEXeXP_PM_Port, 0);
		return(camFAIL);
	}
	/* Sector Count : 0x01 */
	Status = ReadIndexUchar(BaseIOAddress, iNDEXIDESectorCount);	
	if ( Status != 0x01 ) {
		WriteIndexUlong(BaseExpIOAddress, iNDEXeXP_PM_Port, 0);
		return(camFAIL);
	}

	/*
	2. Read PM (PM Port=F, PortNum=F) and get the info of
	"number of exposed device fan-out ports" from GSCR[2]
	*/							
	/* disable INTA here, it will be re-enable when CAM use SEQ 0 for packets */
	WriteIndexUchar(BaseIOAddress, iNDEXIDECICR, 0x0);  /* that the drive INT pass to SEQ 0*/
	WritePortUchar( HostBaseAddr[Adapter_ID], 0, 0x20); /* but mask SEQ 0 INT */

	/* PM port : Control */
	WriteIndexUlong(BaseExpIOAddress, iNDEXeXP_PM_Port, PM_control_Port);	
	/* send command */
	WriteIndexUchar(BaseIOAddress, iNDEXIDEFeature , 0x02);	/* GSCR[2] */
	WriteIndexUchar(BaseIOAddress, iNDEXIDESectorCount , 0);
	WriteIndexUchar(BaseIOAddress, iNDEXIDESectorNumber , 0);
	WriteIndexUchar(BaseIOAddress, iNDEXIDECylinderLow , 0);
	WriteIndexUchar(BaseIOAddress, iNDEXIDECylinderHigh , 0);
	WriteIndexUchar(BaseIOAddress, iNDEXIDEDeviceHead, 0|PM_control_Port); /* control port */
	WriteIndexUchar(BaseIOAddress, iNDEXIDECommand, IDE_COMMAND_READ_PORT_MULTIPLIER);
	if ( WaitForINT(BaseIOAddress) != camSUCCESS ){
		return(camFAIL); /* bad device */
	}
	/* GSCR[2] bit 0~3 */
	deviceNumber = ReadIndexUchar(BaseIOAddress, iNDEXIDESectorCount);	
	deviceNumber &= 0x0F;
	deviceNumber = (deviceNumber>MAX_ID_PER_CHNL) ? MAX_ID_PER_CHNL:deviceNumber;	
	
	gChnlDrvControlConfig[bChannelNumber].Support_device = deviceNumber;
	gChnlDrvControlConfig[bChannelNumber].PM = TRUE;

	/* Device Enumeration */
	/* check device# exist? */
	for (i=0;i<deviceNumber;i++) {
		
		deviceFound=FALSE;
		/* The host will enable the device port. */
		for (j=0;j<PHY_Enable_RetryTime;j++) {
			/*
			3. Write PM (PM Port=F, PortNum=0) and RegNum=02h,
	   		Value=00000001h to reset device port #.
	  		 */			
			/* PM port : Control */
			WriteIndexUlong(BaseExpIOAddress, iNDEXeXP_PM_Port, PM_control_Port);
			WriteIndexUchar(BaseIOAddress, iNDEXIDEFeature , 0x02); /* PSCR[2] */
			WriteIndexUchar(BaseIOAddress, iNDEXIDESectorCount , 0x1);
			WriteIndexUchar(BaseIOAddress, iNDEXIDESectorNumber , 0x0);
			WriteIndexUchar(BaseIOAddress, iNDEXIDECylinderLow , 0);
			WriteIndexUchar(BaseIOAddress, iNDEXIDECylinderHigh , 0);
			
			WriteIndexUchar(BaseIOAddress, iNDEXIDEDeviceHead, (U8)(0|i)); /* device port */
			WriteIndexUchar(BaseIOAddress, iNDEXIDECommand, IDE_COMMAND_WRITE_PORT_MULTIPLIER);
			if ( WaitForINT(BaseIOAddress) != camSUCCESS ){
				continue; /* bad device */
			}
			camUStallExecution(100);
			/*
			4. Write PM (PM Port=F, PortNum=0) and RegNum=02h, 
   			Value=00000000h to make Phy Ready on device port #.
   			*/			
			/* PM port : Control */
			WriteIndexUlong(BaseExpIOAddress, iNDEXeXP_PM_Port, PM_control_Port);
			WriteIndexUchar(BaseIOAddress, iNDEXIDEFeature , 0x02); /* PSCR[2] */
			WriteIndexUchar(BaseIOAddress, iNDEXIDESectorCount , 0x0);
			WriteIndexUchar(BaseIOAddress, iNDEXIDESectorNumber , 0x0);
			WriteIndexUchar(BaseIOAddress, iNDEXIDECylinderLow , 0);
			WriteIndexUchar(BaseIOAddress, iNDEXIDECylinderHigh , 0);
			
			WriteIndexUchar(BaseIOAddress, iNDEXIDEDeviceHead, (U8)(0|i)); /* device port */
			WriteIndexUchar(BaseIOAddress, iNDEXIDECommand, IDE_COMMAND_WRITE_PORT_MULTIPLIER);
			if ( WaitForINT(BaseIOAddress) != camSUCCESS ){
				continue; /* bad device */
			}
			camUStallExecution(100);
			
			for (k=0;k<PHY_Wait_Ready_Time;k++) {
				/* from OOB to PhyRdy, it may require several miliseconds in PM FPGA. */

				/* 
				5. Read PM (PM Port=F, PortNum=0) and RegNum=00h,
	   			if the response is Values=00000113h, 00000123h, or an
	   			indication of PhyRdy, determines the active connection
	   			with a drive
				*/			
				/* PM port : Control */
				WriteIndexUlong(BaseExpIOAddress, iNDEXeXP_PM_Port, PM_control_Port);
				WriteIndexUchar(BaseIOAddress, iNDEXIDEFeature , 0x00);	/* PSCR[0] */
				WriteIndexUchar(BaseIOAddress, iNDEXIDESectorCount , 0);
				WriteIndexUchar(BaseIOAddress, iNDEXIDESectorNumber , 0);
				WriteIndexUchar(BaseIOAddress, iNDEXIDECylinderLow , 0);
				WriteIndexUchar(BaseIOAddress, iNDEXIDECylinderHigh , 0);
				WriteIndexUchar(BaseIOAddress, iNDEXIDEDeviceHead, (U8)(0|i)); /* device port */
				WriteIndexUchar(BaseIOAddress, iNDEXIDECommand, IDE_COMMAND_READ_PORT_MULTIPLIER);
				if ( WaitForINT(BaseIOAddress) != camSUCCESS ){
					continue; /* bad device */
				}
				/* check the response is Values=00000113h, 00000123h */
				ReadIndexWord(BaseIOAddress, iNDEXIDESectorCount,wdata1);
				ReadIndexWord(BaseIOAddress, iNDEXIDESectorNumber,wdata2);
				
				wdata2= ((wdata2&0x00FF) << 8) + (wdata1&0x00FF);
				if ( ((wdata2 & 0x123) != 0x123) && ((wdata2 & 0x113) != 0x113) ) {
					camStallExecution(1);
					continue;	
				}
				
				deviceFound=TRUE;			
				break; /* detect device */
			}
			
			if (deviceFound==TRUE)
				break; /* detect device */			
		}				
			
		if (deviceFound==TRUE) {	
			/* device# exist */
			/*
			6. Write PM (PM Port=F, PortNum=0) and RegNum=01h
			   with Values=FFFFFFFFh to clear errors to all zeros,
			   especially the "X" bit. If X=1, any FIS is disallowed to the
			   device port. This completes the enabling of device port 0.
			*/			
			/* PM port : Control */
			WriteIndexUlong(BaseExpIOAddress, iNDEXeXP_PM_Port, PM_control_Port);
			/* send command */
			WriteIndexUchar(BaseIOAddress, iNDEXIDEFeature , 0x01);	/* PSCR[1] */
			WriteIndexUchar(BaseIOAddress, iNDEXIDESectorCount , 0xFF);
			WriteIndexUchar(BaseIOAddress, iNDEXIDESectorNumber , 0xFF);
			WriteIndexUchar(BaseIOAddress, iNDEXIDECylinderLow , 0xFF);
			WriteIndexUchar(BaseIOAddress, iNDEXIDECylinderHigh , 0xFF);
			WriteIndexUchar(BaseIOAddress, iNDEXIDEDeviceHead, (U8)(0|i)); /* device port */
			WriteIndexUchar(BaseIOAddress, iNDEXIDECommand, IDE_COMMAND_WRITE_PORT_MULTIPLIER);
			if ( WaitForINT(BaseIOAddress) != camSUCCESS ){
				continue; /* bad device */
			}
			/* 
			7. SRST=1 then 0 (PM Port=0 in a RegHD FIS) and get the
			   drive's signature. Then you can do Identify Device and so
			   on.		
			*/									
			/*
			The signature generated by the device as a consequence of the initial COMRESET to the
			device port may be discarded if the host does not support context switching because the BSY
			bit may be clear when the Register FIS is received by the host. Therefore, to determine the
			signature of the attached device, the host should issue a software reset to the device port. If
			a valid signature is returned for a recognized device, the host may then proceed with normal
			initialization for that device type.
			*/	
			/* PM port : device port */
			/* will do AtaSoftReset at CAM_GetATADeviceConfig
			WriteIndexUlong(BaseExpIOAddress, iNDEXeXP_PM_Port, i);			
			AtaSoftReset(bChannelNumber,0);	
			*/
		}
		else {
			/* device# not exist */
			continue;
		}	
			
	}

	WriteIndexUlong(BaseExpIOAddress, iNDEXeXP_PM_Port, 0);
	return (camSUCCESS);
}

U8 CAM_GetATADeviceConfig(PDEV_CONFIG pDevConfig)
{
	U8			bChannelNumber=pDevConfig->bCh;
	U8			bID=pDevConfig->bID;
	PDEV_CONFIG	pCurDevConfig = &gDrvConfig[bChannelNumber][bID]; 
	PTR_SIZE			BaseIOAddress = ATABaseAddr[bChannelNumber];
	U8			statusByte;
	U8			signatureLow, signatureHigh;
	U8			i,bTmp;
	U8			Adapter_ID = bChannelNumber/MAX_ATA_CHNL;
	PCAM_IDENTIFY_DATA	pIdentifybuf;
	
	/* if master drive is in the swapbox, don't try to find slave drive */
	if(( bID != 0 ) && (gDrvConfig[bChannelNumber][0].pEnclosure)) {
		pDevConfig->DevFlag = DEV_NOTFOUND;
		return(camDEVICE_NOT_FOUND);
	}
	
	if ( (bChannelNumber >= MAX_ATA_MODULE )|| (bID > MAX_ID_PER_CHNL) ) {
		pDevConfig->DevFlag = DEV_NOTFOUND;
		return(camDEVICE_NOT_FOUND);
	}
	
	if ( gChnlConfig[bChannelNumber].bCh == camCHNL_NOT_FOUND )
		return (camDEVICE_NOT_FOUND);	
	
	if( ( pCurDevConfig->DevFlag & DEV_EXIST ) 
	 && (!(pCurDevConfig->pEnclosure) || ( gBOXStatus[bChannelNumber].NeedCheck == FALSE) ) ) 	
	{
		*pDevConfig = *pCurDevConfig;
		return(camSUCCESS);
	}

   	pIdentifybuf = (PCAM_IDENTIFY_DATA)( (PTR_SIZE)gAdapterConfig[Adapter_ID].MemVirAddr+ATA_512K_BUFFER_ADDR);
   	
	pCurDevConfig->DevFlag = DEV_NOTFOUND;
	
	/* first init device or just plug-in new device */
	/* Check if the device in swapbox */
	/* first init device or just plug-in new device */
	/* Check if the device in swapbox */
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_ATA ) {
		if (gBOXConfig[bChannelNumber].BOXType == camUNKNOWNBOX )
		{	
			bTmp = CAM_EnumEnclosure( CAM_ATA_Chnl_OutMAP(bChannelNumber), (PENCLOSURE_TYPE)&gBOXConfig[bChannelNumber]);
			if ( bTmp == 0xFF ) 
				pCurDevConfig->pEnclosure = 0;	
			else
				pCurDevConfig->pEnclosure = &gBOXConfig[bChannelNumber];
		}
	
		if ( pCurDevConfig->pEnclosure )
		{
			gBOXStatus[bChannelNumber].NeedCheck = FALSE;
			/* make sure the power is on */
			CAM_Enclosure_PowerOn(CAM_ATA_Chnl_OutMAP(bChannelNumber));
			camStallExecution(100);/* wait a while */
			if(gBOXStatus[bChannelNumber].PowerState == PowerOFF)
				return (camDEVICE_NOT_FOUND); /* no device */
		}
	}

	/* disable INTA here, it will be re-enable when CAM use SEQ 0 for packets */
	WriteIndexUchar(BaseIOAddress, iNDEXIDECICR, 0x0);  /* that the drive INT pass to SEQ 0*/
	WritePortUchar( HostBaseAddr[Adapter_ID], 0, 0x20); /* but mask SEQ 0 INT */
		
	/* set PM Port */
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) { /* SATA */
		if (gChnlDrvControlConfig[bChannelNumber].PM == TRUE) {
		
			U16	wdata1,wdata2;
			/* 
			5. Read PM (PM Port=F, PortNum=0) and RegNum=00h,
			if the response is Values=00000113h, 00000123h, or an
			indication of PhyRdy, determines the active connection
			with a drive
			*/			
			/* PM port : Control */
			WriteIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_PM_Port, PM_control_Port);
			WriteIndexUchar(BaseIOAddress, iNDEXIDEFeature , 0x00);	/* PSCR[0] */
			WriteIndexUchar(BaseIOAddress, iNDEXIDESectorCount , 0);
			WriteIndexUchar(BaseIOAddress, iNDEXIDESectorNumber , 0);
			WriteIndexUchar(BaseIOAddress, iNDEXIDECylinderLow , 0);
			WriteIndexUchar(BaseIOAddress, iNDEXIDECylinderHigh , 0);
			WriteIndexUchar(BaseIOAddress, iNDEXIDEDeviceHead, (U8)(0|bID)); /* device port */
			WriteIndexUchar(BaseIOAddress, iNDEXIDECommand, IDE_COMMAND_READ_PORT_MULTIPLIER);
			if ( WaitForINT(BaseIOAddress) != camSUCCESS ){
				return (camDEVICE_NOT_FOUND); /* no device */
			}
			/* check the response is Values=00000113h, 00000123h */
			ReadIndexWord(BaseIOAddress, iNDEXIDESectorCount,wdata1);
			ReadIndexWord(BaseIOAddress, iNDEXIDESectorNumber,wdata2);
			
			wdata2= ((wdata2&0x00FF) << 8) + (wdata1&0x00FF);
			if ( ((wdata2 & 0x123) != 0x123) && ((wdata2 & 0x113) != 0x113) ) {
				return (camDEVICE_NOT_FOUND); /* no device */
			}
						
			/*
			The signature generated by the device as a consequence of the initial COMRESET to the
			device port may be discarded if the host does not support context switching because the BSY
			bit may be clear when the Register FIS is received by the host. Therefore, to determine the
			signature of the attached device, the host should issue a software reset to the device port. If
			a valid signature is returned for a recognized device, the host may then proceed with normal
			initialization for that device type.
			*/	
			WriteIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_PM_Port, bID);		
			AtaSoftReset(bChannelNumber,0);		
		}
	}
	
	/* select drive */	
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */
		WriteIndexUchar(BaseIOAddress, iNDEXIDEDeviceHead, 0xE0);
	else /* PATA */
		WriteIndexUchar(BaseIOAddress, iNDEXIDEDeviceHead, (U8)(((bID & 0x01) << 4) | 0xA0));
		
	/* read drive status */
	for ( i = 0; i < 10 ; i++ ) {
		/* check not busy , MAX wait 30 seconds */
		if ( !(WaitOnBaseBusy(BaseIOAddress) & IDE_STATUS_BUSY) )
			break;
	}
	if(i >= 10)
		return (camDEVICE_NOT_FOUND); /* no device */
	
	if (!(gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA)) /* PATA */
	{
		WriteIndexUchar(BaseIOAddress, iNDEXIDEDeviceHead, (U8)(((bID & 0x01) << 4) | 0xA0));
		WriteIndexUchar(BaseIOAddress, iNDEXIDECylinderLow, 0x55 );
		WriteIndexUchar(BaseIOAddress, iNDEXIDECylinderHigh, 0xAA );
		statusByte = ReadIndexUchar( BaseIOAddress, iNDEXIDECylinderLow);
		if ( statusByte != 0x55 )
			return(camDEVICE_NOT_FOUND); /* no device */
	}

#ifdef	_ATAPI_				
	/* ATA or ATAPI device ? */
	AtapiSoftReset(bChannelNumber, bID);
	signatureLow = ReadIndexUchar(BaseIOAddress, iNDEXIDECylinderLow);
	signatureHigh = ReadIndexUchar(BaseIOAddress, iNDEXIDECylinderHigh);

	/* issue identify command */ 
	if (signatureLow == 0x14 && signatureHigh == 0xEB) {
		/*
		 * ATAPI signature found, Issue ATAPI packet identify command.
		 */ 
		if ( IssueIdentify( bChannelNumber,
							bID,
							(PU8)pIdentifybuf,
							IDE_COMMAND_IDENTIFY_PACKET_DEVICE) == camFAIL ) 								
			return(camDEVICE_NOT_FOUND);
		else {
			pCurDevConfig->DevFlag = DEV_EXIST+DEV_ATAPI_DEVICE;
			if ( pIdentifybuf->GeneralConfiguration & 0x20 )
				pCurDevConfig->DevFlag |= DEV_INT_DRQ;
		}

	}
	else 
#endif	
	{
			
		/*
		 * Issue IDE Identify. 
		 */
		if ( IssueIdentify( bChannelNumber, bID, (PU8)pIdentifybuf, IDE_COMMAND_IDENTIFY_DEVICE)  == camFAIL ) {
			if ( IssueIdentify( bChannelNumber, bID, (PU8)pIdentifybuf, IDE_COMMAND_IDENTIFY_DEVICE)  == camFAIL )			      			
				return(camDEVICE_NOT_FOUND);
		}
		pCurDevConfig->DevFlag = DEV_EXIST+DEV_ATA_DEVICE;
	}
	
	if (pCurDevConfig->DevFlag & DEV_ATA_DEVICE) { /* ATA */
		pCurDevConfig->CylNum = pIdentifybuf->NumberOfCylinders;
		pCurDevConfig->HeadNum = (U8)pIdentifybuf->NumberOfHeads;
		pCurDevConfig->SecPerTrk = (U8)pIdentifybuf->SectorsPerTrack;	
		
		if (pIdentifybuf->CommandSetSupport2 & bit(10)){
			pCurDevConfig->DevFlag |= DEV_48BIT;
	/*		pCurDevConfig->MAXLBAL = *((PU32)&pIdentifybuf->MaxLBA1);*/
	/*		pCurDevConfig->MAXLBAH = *((PU32)&pIdentifybuf->MaxLBA3);*/
			/* for Big-Endian  */
			pCurDevConfig->MAXLBAL = ( pIdentifybuf->MaxLBA2 * 0x10000UL ) + pIdentifybuf->MaxLBA1;
			pCurDevConfig->MAXLBAH = ( pIdentifybuf->MaxLBA4 * 0x10000UL ) + pIdentifybuf->MaxLBA3;
		}
		else {
	/*		pCurDevConfig->MAXLBAL = *((U32 FARPTR *)&pIdentifybuf->UserAddressableSectors1); */
			/* for Big-Endian  */
			pCurDevConfig->MAXLBAL = ( pIdentifybuf->UserAddressableSectors2 * 0x010000UL ) + pIdentifybuf->UserAddressableSectors1; 
			pCurDevConfig->MAXLBAH = (U32)0;
		}
	}
	else { /* ATAPI */
		pCurDevConfig->CylNum = 0;
		pCurDevConfig->HeadNum = 0;
		pCurDevConfig->SecPerTrk = 0;
		pCurDevConfig->MAXLBAL = 0;
		pCurDevConfig->MAXLBAH = 0;
	}

	/* put parameters to device info structure */
	for ( i = 0; i < 11; i++ ) {
		bTmp =(U8)pIdentifybuf->ModelNumber[i];
		pIdentifybuf->ModelNumber[i] >>= 8;
		pIdentifybuf->ModelNumber[i] |= (U16)bTmp << 8;
	/*	*(PU16)(&pCurDevConfig->Model[i*2]) = pIdentifybuf->ModelNumber[i];*/
		/* for Big-Endian  */
		pCurDevConfig->Model[i*2]   = (U8)pIdentifybuf->ModelNumber[i];
		pCurDevConfig->Model[i*2+1] = (U8)(pIdentifybuf->ModelNumber[i] >> 8 );
	}
	pCurDevConfig->Model[21] = (U8)0;		/* end of string */


	/* get serial no. and f/w version */
	for ( i = 0; i < 10; i++ ) {
		bTmp =(U8)pIdentifybuf->SerialNumber[i];
		pIdentifybuf->SerialNumber[i] >>= 8;
		pIdentifybuf->SerialNumber[i] |= (U16)bTmp << 8;
/*		*(PU16)(&pCurDevConfig->SerialNo[i*2]) = pIdentifybuf->SerialNumber[i];*/
		/* for Big-Endian */
		pCurDevConfig->SerialNo[i*2] = (U8)pIdentifybuf->SerialNumber[i];
		pCurDevConfig->SerialNo[i*2+1] = (U8)( pIdentifybuf->SerialNumber[i] >> 8 );
	}
	pCurDevConfig->SerialNo[20] =(U8)0;		/* end of string */
	
	for ( i = 0; i < 4; i++ ) {
		bTmp =(U8)pIdentifybuf->FirmwareRevision[i];
		pIdentifybuf->FirmwareRevision[i] >>= 8;
		pIdentifybuf->FirmwareRevision[i] |= (U16)bTmp << 8;
/*		*(PU16)(&pCurDevConfig->FWVersion[i*2]) = pIdentifybuf->FirmwareRevision[i];*/
		/* for Big-Endian */
		pCurDevConfig->FWVersion[i*2] = (U8)pIdentifybuf->FirmwareRevision[i];
		pCurDevConfig->FWVersion[i*2+1] = (U8)(pIdentifybuf->FirmwareRevision[i] >> 8 );
	}
	pCurDevConfig->FWVersion[8] =(U8)0;		/* end of string */

		
	if ( pIdentifybuf->MajorRevision != 0xFFFF && (pIdentifybuf->MajorRevision & bit(5)) )
		pCurDevConfig->DevFlag |= DEV_ATA5;

	if (pCurDevConfig->DevFlag & DEV_ATA_DEVICE) { /* ATA */
		if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */ {	
		#ifdef NCQ
			/* check NCQ */
			if ( pIdentifybuf->SerialATAcapabilities & bit(8)) {
				
				if (!(gChnlDrvControlConfig[bChannelNumber].TagCmdFlag & TAG_SEQALLOC)) {
					U8 FreeSEQID;
					/* get free seqid */
					FreeSEQID=CAMAllocSEQID(Adapter_ID);	
		
					if (FreeSEQID!=U8NULL){
						gChnlDrvControlConfig[bChannelNumber].TagCmdFlag |= TAG_SEQALLOC;
						gChnlDrvControlConfig[bChannelNumber].SEQID = FreeSEQID;
	
						pCurDevConfig->DevFlag |= (DEV_NCQ|DEV_FORCE_MEDIA);
						
						pCurDevConfig->QueueDepth = (U8)(pIdentifybuf->QueueDepth) + 1;
						pCurDevConfig->QueueDepth = (pCurDevConfig->QueueDepth>MAX_TAG) ? MAX_TAG:pCurDevConfig->QueueDepth;
						gChnlDrvControlConfig[bChannelNumber].FreeTAG = 0xFFFFFFFF;
						gChnlDrvControlConfig[bChannelNumber].UsedTAG = 0x0;
					}
					/* if no free deqID, can't issue NCQ */
				}
			}
		#endif
		
		#ifdef TCQ
			if (!(pCurDevConfig->DevFlag & DEV_NCQ) ) {
				/* check support QUEUED commands? */
				if ( pIdentifybuf->CommandSetSupport2 & bit(1) ) {
					
					/* check is IBM HD */
					U8 isWD=FALSE;
				#ifdef ALL_TCQ_HD_SUPPORT
					isWD = TRUE;				
				#endif

					gChnlDrvControlConfig[bChannelNumber].IsWDHardDrive = FALSE;
					
					if ( (pCurDevConfig->Model[0] == 0x57) && (pCurDevConfig->Model[1] == 0x44) ){/* not Hitachi */
						U8	*WD_Model;

						gChnlDrvControlConfig[bChannelNumber].IsWDHardDrive = TRUE;
						isWD = TRUE;
						
						/* check if model is WDC WD740G series, then check firmware version ( > 27) */
						WD_Model = pCurDevConfig->Model;
						if((*(WD_Model+4) == 'W') && (*(WD_Model+5) == 'D') && (*(WD_Model+6) == '7')
						&& (*(WD_Model+7) == '4') && (*(WD_Model+8) == '0') && (*(WD_Model+9) == 'G')){
							WD_Model = pCurDevConfig->FWVersion;
							if((*(WD_Model+1) == '.') || (*(WD_Model) < '2'))
								isWD=FALSE;
							else if((*(WD_Model) == '2') && (*(WD_Model+1) < '7'))	
								isWD=FALSE;
						}
					}

					if (isWD == TRUE) {
						if (!(gChnlDrvControlConfig[bChannelNumber].TagCmdFlag & TAG_SEQALLOC)) {
							U8 FreeSEQID;
							/* get free seqid */
							FreeSEQID=CAMAllocSEQID(Adapter_ID);	
				
							if (FreeSEQID!=U8NULL){
								gChnlDrvControlConfig[bChannelNumber].TagCmdFlag |= TAG_SEQALLOC;
								gChnlDrvControlConfig[bChannelNumber].SEQID = FreeSEQID;
								
								pCurDevConfig->QueueDepth = (U8)(pIdentifybuf->QueueDepth) + 1;
								pCurDevConfig->QueueDepth = (pCurDevConfig->QueueDepth>MAX_TAG) ? MAX_TAG:pCurDevConfig->QueueDepth;
								
								gChnlDrvControlConfig[bChannelNumber].FreeTAG = 0xFFFFFFFF;
								gChnlDrvControlConfig[bChannelNumber].UsedTAG = 0x0;
								gChnlDrvControlConfig[bChannelNumber].TCQTransferData = FALSE;
							#ifndef SINGLE_TAG_QUEUED
								if (pCurDevConfig->QueueDepth > 1) {
									pCurDevConfig->DevFlag |= DEV_QUEUED;
									/* check Write QUEUED FUA commands support? */
									if ( pIdentifybuf->CommandSetSupport3 & bit(7) ) {
										pCurDevConfig->DevFlag |= DEV_FORCE_MEDIA;
									}					
								}
							#else
								
								gChnlDrvControlConfig[bChannelNumber].FreeTAG = bit(0);
								
								if (pCurDevConfig->QueueDepth > 1) {
									pCurDevConfig->QueueDepth = 1;
									pCurDevConfig->DevFlag |= DEV_QUEUED;
									/* check Write QUEUED FUA commands support? */
									if ( pIdentifybuf->CommandSetSupport3 & bit(7) ) {
										pCurDevConfig->DevFlag |= DEV_FORCE_MEDIA;
									}					
								}
							#endif
											
							}
							/* if no free deqID, can't issue QUEUED */
						}
					}
				}
			}
		#endif /* TCQ */
		} /* end if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) */
		/* check Write FUA commands support? */
		if ( pIdentifybuf->CommandSetSupport3 & bit(6) )
			pCurDevConfig->DevFlag |= DEV_FORCE_MEDIA;			
	}
	
	/* Get Transfer mode */	
	if ( (pIdentifybuf->FieldsValid & 0x02) && (pIdentifybuf->AdvancedPIOModes & 0xf)
	  && (!pCurDevConfig->pEnclosure) )
	{
		U8 bPIO = (U8)pIdentifybuf->AdvancedPIOModes;
		if ( bPIO >= 0x02)
	  		pCurDevConfig->PIOMode = 0x04; 
		else
	  		pCurDevConfig->PIOMode = 0x03; 
	}
	else {
		pCurDevConfig->PIOMode = 0x02;
	}
    
	pCurDevConfig->MDMAMode = 0xFF;
	pCurDevConfig->UDMAMode = 0xFF;   
    
	if ( (pIdentifybuf->FieldsValid & 0x02) && (pIdentifybuf->MultiWordDMAModes & 0xf))
	{
		U8	bMultiDMA = (U8)pIdentifybuf->MultiWordDMAModes;
		while( bMultiDMA & 0x01 ) {
			pCurDevConfig->MDMAMode++;
			 bMultiDMA >>= 1;
	  	}
	}
	if ( (pIdentifybuf->FieldsValid & 0x04) &&
		(pIdentifybuf->CommandSetSupport7 & 0x7f))
	{
		U8	UMode = (U8)pIdentifybuf->CommandSetSupport7;
    	
		if( ( gChnlConfig[bChannelNumber].CableType < CABLE_80 )
		  ||(  gChnlConfig[bChannelNumber].ChnlSpeed < CHNL_U100 ) )
			UMode &= 0x07;
			
		if (  gChnlConfig[bChannelNumber].ChnlSpeed < CHNL_U133 )
			UMode &= 0x3F;
			
   		while (UMode & 0x01) {
   			pCurDevConfig->UDMAMode++;
   			UMode >>= 1;
		}
	}	
	
	
	/* get feature status */
	pCurDevConfig->Features = 0;
	if ( pIdentifybuf->CommandSetSupport1 & 0x01 ) {
		pCurDevConfig->Features |= DEV_SMART;
		if ( pIdentifybuf->CommandSetSupport4 & 0x01 )
			pCurDevConfig->Features |= DEV_EN_SMART;	
	}
	if ( pIdentifybuf->CommandSetSupport1 & 0x20 ) {
		pCurDevConfig->Features |= DEV_WRITECACHE;
		if ( pIdentifybuf->CommandSetSupport4 & 0x20 )
			pCurDevConfig->Features |= DEV_EN_WRITECACHE;
	}

	if (pCurDevConfig->DevFlag & DEV_NCQ) {
		/* check DMA Setup Auto-Activate optimization */
		if ( pIdentifybuf->SerialATAFeaturesSupported & bit(2)) {
			pCurDevConfig->Features |= DEV_NCQ_DMASetupAuto;
			if ( pIdentifybuf->SerialATAFeaturesEnabled & bit(2)) {
				pCurDevConfig->Features |= DEV_EN_NCQ_DMASetupAuto;
			}
		}
		/* check non-zero buffer offsets in DMA Setup FIS */
		if ( pIdentifybuf->SerialATAFeaturesSupported & bit(1)) {
			pCurDevConfig->Features |= DEV_NCQ_non_zero_offsets;
			if ( pIdentifybuf->SerialATAFeaturesEnabled & bit(1)) {
				pCurDevConfig->Features |= DEV_EN_NCQ_non_zero_offsets;
			}
		}		
	}

#ifdef PIORW	
	pCurDevConfig->MaximumBlockTransfer=pIdentifybuf->MaximumBlockTransfer;
	/* get PIO number of sectors per block */
	if (pIdentifybuf->CurrentMultiSectorSetting & bit(8) ) {
		pCurDevConfig->CurrentBlockTransfer=(U8)pIdentifybuf->CurrentMultiSectorSetting;
	}
	else {
		pCurDevConfig->CurrentBlockTransfer=0;
	}
		
#endif

		
	if((gChnlConfig[bChannelNumber].ChnlType != CHNL_SATA) && (bID)) {/* PATA slave */
		PDEV_CONFIG		pMasterDevConfig=&gDrvConfig[bChannelNumber][0];
		if(pMasterDevConfig->DevFlag & DEV_EXIST){
			if( pCurDevConfig->UDMAMode != 0xFF ){
				if( pMasterDevConfig->UDMAMode == 0xFF && pMasterDevConfig->MDMAMode != 0xFF ){
					pMasterDevConfig->MDMAMode = 0xFF;
					if (CAMInitDevice(pMasterDevConfig) != camSUCCESS)
						pMasterDevConfig->DevFlag = DEV_NOTFOUND;					
				}
			}
			else{
				if( pMasterDevConfig->UDMAMode != 0xFF && pCurDevConfig->MDMAMode != 0xFF )
					pCurDevConfig->MDMAMode = 0xFF;
			}
		}		
	}


	if ( CAMInitDevice(pCurDevConfig) != camSUCCESS ) {
		pCurDevConfig->DevFlag = DEV_NOTFOUND;
		return (camDEVICE_NOT_FOUND);			
	}
	*pDevConfig = *pCurDevConfig;

	/* first init device or just plug-in new device */
	/* Check if the device in swapbox */
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA ) {
		if (gBOXConfig[bChannelNumber].BOXType == camUNKNOWNBOX )
		{	
			bTmp = CAM_EnumEnclosure( CAM_ATA_Chnl_OutMAP(bChannelNumber), (PENCLOSURE_TYPE)&gBOXConfig[bChannelNumber]);
			if ( bTmp == 0xFF ) 
				pCurDevConfig->pEnclosure = 0;	
			else{
				pCurDevConfig->pEnclosure = &gBOXConfig[bChannelNumber];
			}
		}
	}
	
	return(camSUCCESS);

}

U8 CAM_EnumATADevice(PDEV_CONFIG pDevConfig)
{
	U8 			bChannelNumber=CAM_ATA_Chnl_InMAP(pDevConfig->bCh);
	U8			bID=pDevConfig->bID;
	PDEV_CONFIG	pCurDevConfig = &gDrvConfig[bChannelNumber][bID]; 
	PDEV_CONFIG pNxtDevConfig = &gDrvConfig[bChannelNumber][(bID+1)&0x01];
	
	/* if master drive is in the swapbox, don't try to find slave drive */
	if(( bID == 1 ) && (pNxtDevConfig->pEnclosure)) {
		pDevConfig->DevFlag = DEV_NOTFOUND;
		return(camDEVICE_NOT_FOUND);
	}
	
	if ( (bChannelNumber >= MAX_ATA_MODULE )|| (bID > MAX_ID_PER_CHNL) ) {
		pDevConfig->DevFlag = DEV_NOTFOUND;
		return(camDEVICE_NOT_FOUND);
	}
	
	if ( gChnlConfig[bChannelNumber].bCh == camCHNL_NOT_FOUND ){
		return (camDEVICE_NOT_FOUND);	
	}

	/* keep pDevConfig->bCh */
	bChannelNumber = pDevConfig->bCh;
	*pDevConfig = *pCurDevConfig;
	/* restore */
	pDevConfig->bCh=bChannelNumber;
		
	if ( pCurDevConfig->DevFlag & DEV_EXIST )
		return(camSUCCESS);
	else
		return (camDEVICE_NOT_FOUND);		
}

/* SATA chanel not support this function */
U8 CAM_ConfigATAChannel(PCHNL_CONFIG pChnlConfig)
{
	PCHNL_CONFIG	pChnl;
	DEV_CONFIG 		TmpDevConfig;
	U8 bChannelNumber=CAM_ATA_Chnl_InMAP(pChnlConfig->bCh);
	
	pChnl= &gChnlConfig[bChannelNumber];
	
	/* SATA chanel not support this function */
	if (pChnl->ChnlType == CHNL_SATA)
		return (camFAIL);
	
	if (pChnlConfig->bCh >= MAX_ATA_MODULE )
		return (camCHNL_NOT_FOUND);	
		
	if ( pChnl->bCh == camCHNL_NOT_FOUND )
		return (camCHNL_NOT_FOUND);	
		
	if ( pChnlConfig->ChnlSpeed > pChnl->ChnlSpeed )
		return (camFAIL);							/* can't set faster speed than max supported speed */
	
	if (pChnl->ChnlSpeed == pChnlConfig->ChnlSpeed)	/* no change */
		return(camSUCCESS);
	
	pChnl->ChnlSpeed = pChnlConfig->ChnlSpeed;		/*only ChnlSpeed can be changed*/

	/* master device */
	TmpDevConfig.bCh=pChnlConfig->bCh;
	TmpDevConfig.bID=0;
	if (CAM_EnumATADevice(&TmpDevConfig) == camSUCCESS)
		CAM_ConfigATADevice(&TmpDevConfig);

	/* slave device */
	TmpDevConfig.bCh=pChnlConfig->bCh;
	TmpDevConfig.bID=1;
	if (CAM_EnumATADevice(&TmpDevConfig) == camSUCCESS)
		CAM_ConfigATADevice(&TmpDevConfig);
				
	return (camSUCCESS);
}

U8 CAM_EnumATAChannel(PCHNL_CONFIG pChnlConfig)
{
	PCHNL_CONFIG	pChnl;
	U8 bChannelNumber=CAM_ATA_Chnl_InMAP(pChnlConfig->bCh);
	
	if (pChnlConfig->bCh >= MAX_ATA_MODULE )
		return (camCHNL_NOT_FOUND);	
	
	pChnl= &gChnlConfig[bChannelNumber];
		
	if ( pChnl->bCh == camCHNL_NOT_FOUND )
		return (camCHNL_NOT_FOUND);	
	
	/* keep pChnlConfig->bCh */
	bChannelNumber = pChnlConfig->bCh;
	*pChnlConfig = *pChnl;
	/* restore */
	pChnlConfig->bCh=bChannelNumber;
	
	return(camSUCCESS);
}

void CAM_Clear_UnmaskPlugUnplugFlags(U8 bChannelNumber)
{
	U8 bCh =  bChannelNumber%MAX_ATA_CHNL;
	U8 Adapter_ID = bChannelNumber/MAX_ATA_CHNL;
	U8 bData;
	
	/* clear plug/unplug flags */
	ReadPortUchar( HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG, bData );	
	bData &= (0x11 << bCh);
	bData |= (0x11 << bCh);		/* clear plug/unplug flags */
	WritePortUchar( HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG, bData );

	/* unmask plug/unplug INT */
	ReadPortUchar( HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG+2, bData );	
	bData &= ~(0x11 << bCh);   /* unmask plug/unplug INT */
	WritePortUchar( HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG+2, bData );
	
	return;	
}


U8 CAM_SATA_PHY_Enable(U8	bChannelNumber)
{
	U32 ldata;
	U8 i,j;

#ifndef	_MMIO_
	PTR_SIZE	BaseIOAddress = ATABaseAddr[bChannelNumber];
#endif
	
	/* change PHYMODE4 to 01 can improve the power-on performance */
	#ifdef	_MMIO_
	ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEPHYmode, ldata );
	ldata &= 0xFFFFFFFC;
	ldata |= 0x1;
	WriteIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEPHYmode, ldata );
	#else
	ReadIndexUlong(BaseIOAddress, iNDEXIDEPHYmode, ldata );
	ldata &= 0xFFFFFFFC;
	ldata |= 0x1;
	WriteIndexUlong(BaseIOAddress, iNDEXIDEPHYmode, ldata );
	#endif
	
	camUStallExecution(1);


	#ifdef _ERR_INJECT_
	{
		if(ErrInject_UnPlug_flag[bChannelNumber] == TRUE)		
			return 0;	
	}
	#endif

	

	#ifdef	_MMIO_
	ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATASControl, ldata );
	#else
	ReadIndexUlong(BaseIOAddress, iNDEXIDESATASControl, ldata );	
	#endif
	
	if ( (ldata&0x07) == 0){
		#ifdef	_MMIO_
		ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXSATAStatus, ldata );
		#else
		ReadIndexUlong(BaseIOAddress, iNDEXSATAStatus, ldata );	
		#endif
		
		if ((ldata & 0x123) == 0x123)
				return camSUCCESS;
		else if ((ldata & 0x113) == 0x113)
				return camSUCCESS;

	}
			
	for (i=0;i<PHY_Enable_RetryTime;i++)
	{
		#ifdef	_MMIO_
		for (j=0;j<10;j++) {
			WriteIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATASControl, 1 );
			camUStallExecution(1);
			ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATASControl, ldata );
			if ( (ldata&0x07) == 1)
				break;
		}
		camUStallExecution(100);
		for (j=0;j<10;j++) {
			WriteIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATASControl, 0 );
			camUStallExecution(1);
			ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATASControl, ldata );
			if ( (ldata&0x07) == 0)
				break;			
		}
		//camUStallExecution(100);
		camStallExecution(10);

		/* Missing TXCLK workaround */
		ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEPHYMode2Reg, ldata );
		ldata &= 0xFFFFFFFE;
		WriteIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEPHYMode2Reg, ldata );
		camUStallExecution(1);

		ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEPHYMode2Reg, ldata );
		ldata |= 0x40000000;
		WriteIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEPHYMode2Reg, ldata );
		camUStallExecution(20);

		ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEPHYMode2Reg, ldata );
		ldata &= 0xBFFFFFFF;
		WriteIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEPHYMode2Reg, ldata );
		camUStallExecution(1);

		ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEPHYMode2Reg, ldata );
		ldata |= 0x00000001;
		WriteIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEPHYMode2Reg, ldata );
		camUStallExecution(1);

		ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEPHYMode2Reg, ldata );
		ldata |= 0x80000000;
		WriteIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEPHYMode2Reg, ldata );
		camUStallExecution(20);

		ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEPHYMode2Reg, ldata );
		ldata &= 0x7FFFFFFF;
		WriteIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEPHYMode2Reg, ldata );
		//camUStallExecution(100);
		camStallExecution(10);

		ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXSATAStatus, ldata );
		#else
		for (j=0;j<10;j++) {
			WriteIndexUlong(BaseIOAddress, iNDEXIDESATASControl, 1 );
			camUStallExecution(1);
			ReadIndexUlong(BaseIOAddress, iNDEXIDESATASControl, ldata );
			if ( (ldata&0x07) == 1)
				break;
		}
		camUStallExecution(100);
		for (j=0;j<10;j++) {
			WriteIndexUlong(BaseIOAddress, iNDEXIDESATASControl, 0 );
			camUStallExecution(1);
			ReadIndexUlong(BaseIOAddress, iNDEXIDESATASControl, ldata );
			if ( (ldata&0x07) == 0)
				break;			
		}
		//camUStallExecution(100);
		camStallExecution(10);
		
		/* Missing TXCLK workaround */
		ReadIndexUlong(BaseIOAddress, iNDEXIDEPHYMode2Reg, ldata );
		ldata &= 0xFFFFFFFE;
		WriteIndexUlong(BaseIOAddress, iNDEXIDEPHYMode2Reg, ldata );
		camUStallExecution(1);

		ReadIndexUlong(BaseIOAddress, iNDEXIDEPHYMode2Reg, ldata );
		ldata |= 0x40000000;
		WriteIndexUlong(BaseIOAddress, iNDEXIDEPHYMode2Reg, ldata );
		camUStallExecution(20);

		ReadIndexUlong(BaseIOAddress, iNDEXIDEPHYMode2Reg, ldata );
		ldata &= 0xBFFFFFFF;
		WriteIndexUlong(BaseIOAddress, iNDEXIDEPHYMode2Reg, ldata );
		camUStallExecution(1);

		ReadIndexUlong(BaseIOAddress, iNDEXIDEPHYMode2Reg, ldata );
		ldata |= 0x00000001;
		WriteIndexUlong(BaseIOAddress, iNDEXIDEPHYMode2Reg, ldata );
		camUStallExecution(1);

		ReadIndexUlong(BaseIOAddress, iNDEXIDEPHYMode2Reg, ldata );
		ldata |= 0x80000000;
		WriteIndexUlong(BaseIOAddress, iNDEXIDEPHYMode2Reg, ldata );
		camUStallExecution(20);

		ReadIndexUlong(BaseIOAddress, iNDEXIDEPHYMode2Reg, ldata );
		ldata &= 0x7FFFFFFF;
		WriteIndexUlong(BaseIOAddress, iNDEXIDEPHYMode2Reg, ldata );
		//camUStallExecution(100);
		camStallExecution(10);

		ReadIndexUlong(BaseIOAddress, iNDEXSATAStatus, ldata );
		#endif
		
		if ((ldata & 0x123) == 0x123)
			break;
		if ((ldata & 0x113) == 0x113)
			break;
	}
	if(i >= PHY_Enable_RetryTime)
		return camFAIL;
	else{	
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEDeviceControl, 0x04 );
		ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEAlternateStatus);	
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEDeviceControl, 0x00 );
		camStallExecution(10);	/* delay 100 ms */	
	}		
	return camSUCCESS;
}

void CAM_NotAtCommandPacketPhase(U8 bChannelNumber)
{
	U8 i;
	U32 ldata;	

	for (i=0;i<100;i++) { /* check not at ASIC packet command phase */
		WriteIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_Internel_debug_1, 0x00);
		ReadIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_Internel_debug_2, ldata);
		if ( (ldata&0x0000000F) != 1)
			break;
		camUStallExecution(10);
	}
	
	return;
}

U8 CAM_HardResetSATAChannel(U8 bChannelNumber)
{
	U8 	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;
	U32	i;
	U8	bData;
	U8 	bCh =  bChannelNumber%MAX_ATA_CHNL;
	U32	ldata = 0;	

	CAM_NotAtCommandPacketPhase(bChannelNumber);

	/* set PM Port */
	WriteIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_PM_Port, PM_control_Port);
	
	/* mask plug/unplug INT */
	ReadPortUchar( HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG+2, bData );	
	bData |= (0x11 << bCh);		/* mask plug/unplug INT */
	WritePortUchar( HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG+2, bData );
			
	/* do hard reset */ 
	ReadPortUchar(HostBaseAddr[Adapter_ID],oFFSETHOSTPCICTL+1 , bData );
	bData &= ~ (0x10 << bCh);
	WritePortUchar(HostBaseAddr[Adapter_ID],oFFSETHOSTPCICTL+1 , bData );
	camStallExecution(1);			
	bData |= (0x10 << bCh);
	WritePortUchar(HostBaseAddr[Adapter_ID],oFFSETHOSTPCICTL+1 , bData );

	/* soft reset ATA Module. */
	ATASoftResetModule(bChannelNumber);	

	if((CAM_SATA_PHY_Enable(bChannelNumber)) != camSUCCESS){
		i = 100000;
	}
	else{
		/* wait SATA connect well in 1 second */
		for( i = 0; i < 100000; i++ )
		{
			ReadPortUlong( HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG, ldata );
			ldata >>= bCh;
			if ( ldata & 0x100 )
				break;
			camUStallExecution(1);
		}
	}

	/* clear plug/unplug flags */
	ReadPortUchar( HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG, bData );	
	bData &= (0x11 << bCh);
	bData |= (0x11 << bCh);		/* clear plug/unplug flags */
	WritePortUchar( HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG, bData );

	if( i < 100000 )
		CAM_SATA_Port_Multiplier_Update_status(bChannelNumber);
	/* set PM Port */
	WriteIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_PM_Port, 0);
				
	if( i < 100000 )
		return(camSUCCESS);
	else
		return(camDEVICE_NOT_FOUND);
}


void CAM_InitATAChnl_CallBack(U8 bChannelNumber, U8 status)
{
	U8	i;
	U8  found_disk=FALSE;
		
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA )
	{

		if(status != camSUCCESS){
			CAM_Clear_UnmaskPlugUnplugFlags(bChannelNumber);
			ATAModuleReceiveTask[bChannelNumber]=gChnlResetTable[bChannelNumber].receiveStatus;
			gChnlResetTable[bChannelNumber].notify(CAM_ATA_Chnl_OutMAP(bChannelNumber), camFAIL);
			return;						
		}
		
			/* check Port_Multiplier */
#ifdef Port_Multiplier
		if (CAM_Init_SATA_Port_Multiplier(bChannelNumber) != camFAIL) {
			/*ATA_Port_Multiplier_TimerID[bChannelNumber] = camTimer(((U8)(bChannelNumber/MAX_ATA_CHNL)),CAMTimeoutTimes[bChannelNumber/MAX_ATA_CHNL].Port_Multiplier_Check, ATA_Check_Port_Multiplier_Function[bChannelNumber]);*/
		}
		else {
			gChnlDrvControlConfig[bChannelNumber].PM = FALSE;
			gChnlDrvControlConfig[bChannelNumber].Support_device = 1;
		}
#else
			gChnlDrvControlConfig[bChannelNumber].PM = FALSE;
			gChnlDrvControlConfig[bChannelNumber].Support_device = 1;
#endif
				
		for (i=0;i<gChnlDrvControlConfig[bChannelNumber].Support_device;i++) {
			if ( CAM_GetATADeviceConfig(&gDrvConfig[bChannelNumber][i]) != camDEVICE_NOT_FOUND ) {
				found_disk=TRUE; /* device is found */
			}
		}

		if (found_disk==TRUE){
			CAM_Clear_UnmaskPlugUnplugFlags(bChannelNumber);
			ATAModuleReceiveTask[bChannelNumber]=gChnlResetTable[bChannelNumber].receiveStatus;
			gChnlResetTable[bChannelNumber].notify(CAM_ATA_Chnl_OutMAP(bChannelNumber), camSUCCESS);
		}
		else{
			if ( CAM_HardResetSATAChannel(bChannelNumber) == camDEVICE_NOT_FOUND ){
				CAM_Clear_UnmaskPlugUnplugFlags(bChannelNumber);
				ATAModuleReceiveTask[bChannelNumber]=gChnlResetTable[bChannelNumber].receiveStatus;
				gChnlResetTable[bChannelNumber].notify(CAM_ATA_Chnl_OutMAP(bChannelNumber), camFAIL);
			}
			else{
				ATA_Polling_Channel_TimerID[bChannelNumber] = camTimer((U8)(bChannelNumber/MAX_ATA_CHNL), 1, ATA_Polling_Channel_Function[bChannelNumber]);
			}	
		}
		return;
	}
	else
	{
		/* when master and slave drives attached at a parallel channel, but the master drive not power-on, as some special case, the STATUS will
			always BUSY when CAM_InitATAChannel, but if drive exist and select drive before read STATUS, the drive STATUS will not BUSY and
			can detect the slave drive only at this case. */
		/*
		if ( WaitOnBusy(ATABaseAddr[bChannelNumber]) & IDE_STATUS_BUSY )
		{
			ATAModuleReceiveTask[bChannelNumber]=receiveStatus;
			return(camFAIL);
		}
		*/

		/* not check this , because is we have swap box at this channel but the power is off
			, it will not be found. */
		/*
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECylinderLow, 0x55 );
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECylinderHigh, 0xAA );
		if ( ReadIndexUchar( ATABaseAddr[bChannelNumber], iNDEXIDECylinderLow) != 0x55 )
			return(camFAIL);
		*/
		
		if ( CAM_GetATADeviceConfig(&gDrvConfig[bChannelNumber][0]) != camDEVICE_NOT_FOUND ) {
		}
		if ( CAM_GetATADeviceConfig(&gDrvConfig[bChannelNumber][1]) != camDEVICE_NOT_FOUND ) {
		}
	

		ATAModuleReceiveTask[bChannelNumber]=gChnlResetTable[bChannelNumber].receiveStatus;
		if ( (gDrvConfig[bChannelNumber][0].DevFlag == DEV_NOTFOUND)
		  & ( gDrvConfig[bChannelNumber][1].DevFlag == DEV_NOTFOUND ) )
		{
			gChnlResetTable[bChannelNumber].notify(CAM_ATA_Chnl_OutMAP(bChannelNumber), camFAIL);
		}
		else
			gChnlResetTable[bChannelNumber].notify(CAM_ATA_Chnl_OutMAP(bChannelNumber), camSUCCESS);

		return;
	}	
}


U8 CAM_InitATAChannel_Previous(U8 bChannelNumber)
{
	U8	receiveStatus;
	U32	i;
	
	bChannelNumber=CAM_ATA_Chnl_InMAP(bChannelNumber);
	
	if ( gChnlConfig[bChannelNumber].bCh == camCHNL_NOT_FOUND )
		return (camCHNL_NOT_FOUND);	
		
	if (bChannelNumber >= MAX_ATA_MODULE)
		return (camCHNL_NOT_FOUND);

	for( i=0;i<MAX_ID_PER_CHNL;i++ ){
		ATASortHead[bChannelNumber][i] = (PCAM_ATA_PACKET)PTR_NULL;
		ATASortLBAl[bChannelNumber][i] = ZERO_NULL;
		ATASortLBAh[bChannelNumber][i] = ZERO_NULL;	
		gDrvConfig[bChannelNumber][i].DevFlag = DEV_NOTFOUND;		
	}
	
#ifdef Encryption
	if(gAdapterConfig[bChannelNumber/MAX_ATA_CHNL].Features & WITH_EncryKEY)
		WriteIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_Encry_SeedKey, gAdapterConfig[bChannelNumber/MAX_ATA_CHNL].EncryptionKey);

	/*WriteIndexUchar(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_Ext_Ctrl, 0x02 );*/
#endif	

	/* disable receive task.*/
	receiveStatus=ATAModuleReceiveTask[bChannelNumber];
	ATAModuleReceiveTask[bChannelNumber]=FALSE;
		
	/* soft reset ATA Module ANYWAY. */
	ATASoftResetModule(bChannelNumber);	
	
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA )
	{
		U8 	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;
		U8	bCh =  bChannelNumber%MAX_ATA_CHNL;
		U32	ldata = 0;
		U8	count=0;
		U8	status=0;
		U8  found_disk=FALSE;

		if((CAM_SATA_PHY_Enable(bChannelNumber)) != camSUCCESS){
			ATAModuleReceiveTask[bChannelNumber]=receiveStatus;	
			return(camFAIL);	/* no drive connected */							
		}
		
		CAM_SATA_Port_Multiplier_Update_status(bChannelNumber);
		
		for (count=0;count<DEF_SATA_INITCOUNT;count++) 
		{
			/* wait SATA connect well in 1 second */
			for ( i = 0; i < 100000; i++ )
			{
				ReadPortUlong( HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG, ldata );
				ldata >>= bCh;
				if ( ldata & 0x100 )
					break;
				camUStallExecution(1);
			}		
			if ( i >= 100000 ) 
			{
				ATAModuleReceiveTask[bChannelNumber]=receiveStatus;	
				return(camFAIL);	/* no drive connected */				
			}			
			
			/* wait 30 seconds, following the ATA spec. special for dongle+ ATA HDD combination */	
			for ( i = 0; i < 10 ; i++ ) {
				if ((count==0) && (i == 0)) { /* phy ready, but drive not update status */
					status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);
					if (status == 0xFF) 
						status=camWARNING;
				}
				else {
					if ( (status=WaitOnReady(bChannelNumber)) == camSUCCESS ) 
						break;
				}
				if (status==camWARNING) {
					if ( CAM_HardResetSATAChannel(bChannelNumber) == camDEVICE_NOT_FOUND ) {
						CAM_Clear_UnmaskPlugUnplugFlags(bChannelNumber);
						ATAModuleReceiveTask[bChannelNumber]=receiveStatus;
						return(camFAIL);	/* no drive connected */
					}
				}	
			}
			if( i >= 10 ) {
				if (status==camFAIL) {
					if ( CAM_HardResetSATAChannel(bChannelNumber) == camDEVICE_NOT_FOUND ) {
						CAM_Clear_UnmaskPlugUnplugFlags(bChannelNumber);
						ATAModuleReceiveTask[bChannelNumber]=receiveStatus;
						return(camFAIL);	/* no drive connected */
					}
				}				
				continue;	/* drive not ready after 30 seconds. */
			}

			/* check Port_Multiplier */
		#ifdef Port_Multiplier
			if (CAM_Init_SATA_Port_Multiplier(bChannelNumber) != camFAIL) {
				/*ATA_Port_Multiplier_TimerID[bChannelNumber] = camTimer(((U8)(bChannelNumber/MAX_ATA_CHNL)),CAMTimeoutTimes[bChannelNumber/MAX_ATA_CHNL].Port_Multiplier_Check, ATA_Check_Port_Multiplier_Function[bChannelNumber]);*/
			}
			else {
				gChnlDrvControlConfig[bChannelNumber].PM = FALSE;
				gChnlDrvControlConfig[bChannelNumber].Support_device = 1;
			}
		#else
			gChnlDrvControlConfig[bChannelNumber].PM = FALSE;
			gChnlDrvControlConfig[bChannelNumber].Support_device = 1;
		#endif
				
			for (i=0;i<gChnlDrvControlConfig[bChannelNumber].Support_device;i++) {
				if ( CAM_GetATADeviceConfig(&gDrvConfig[bChannelNumber][i]) != camDEVICE_NOT_FOUND ) {
					found_disk=TRUE; /* device is found */
				}
			}
			if (found_disk == TRUE)
				break;
			else{
				if ( CAM_HardResetSATAChannel(bChannelNumber) == camDEVICE_NOT_FOUND ) {
					CAM_Clear_UnmaskPlugUnplugFlags(bChannelNumber);
					ATAModuleReceiveTask[bChannelNumber]=receiveStatus;
					return(camFAIL);	/* no drive connected */
				}
			}
		}
		CAM_Clear_UnmaskPlugUnplugFlags(bChannelNumber);
		if ( count >= DEF_SATA_INITCOUNT ) {
			ATAModuleReceiveTask[bChannelNumber]=receiveStatus;
			return(camFAIL);	/* no drive connected */
		}
	}
	else
	{
		/* when master and slave drives attached at a parallel channel, but the master drive not power-on, as some special case, the STATUS will
			always BUSY when CAM_InitATAChannel, but if drive exist and select drive before read STATUS, the drive STATUS will not BUSY and
			can detect the slave drive only at this case. */
		/*
		if ( WaitOnBusy(ATABaseAddr[bChannelNumber]) & IDE_STATUS_BUSY )
		{
			ATAModuleReceiveTask[bChannelNumber]=receiveStatus;
			return(camFAIL);
		}
		*/

		/* not check this , because is we have swap box at this channel but the power is off
			, it will not be found. */
		/*
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECylinderLow, 0x55 );
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECylinderHigh, 0xAA );
		if ( ReadIndexUchar( ATABaseAddr[bChannelNumber], iNDEXIDECylinderLow) != 0x55 )
			return(camFAIL);
		*/
		
		if ( CAM_GetATADeviceConfig(&gDrvConfig[bChannelNumber][0]) != camDEVICE_NOT_FOUND ) {
		}
		if ( CAM_GetATADeviceConfig(&gDrvConfig[bChannelNumber][1]) != camDEVICE_NOT_FOUND ) {
		}
		
		if ( (gDrvConfig[bChannelNumber][0].DevFlag == DEV_NOTFOUND)
		  & ( gDrvConfig[bChannelNumber][1].DevFlag == DEV_NOTFOUND ) )
		{
			ATAModuleReceiveTask[bChannelNumber]=receiveStatus;
			return(camFAIL);	/* no drive connected */;
		}
	}	
	
	ATAModuleReceiveTask[bChannelNumber]=receiveStatus;
	return(camSUCCESS);
}


U8 CAM_InitATAChannel(U8 bChannelNumber, void (*notify)(U8 bCh, U8 status))
{
	U32	i;
	U8 	Adapter_ID;
	
	/* reserve older method for init Ch, for windows initialization */
	if(notify == 0)
		return (CAM_InitATAChannel_Previous(bChannelNumber));


	bChannelNumber=CAM_ATA_Chnl_InMAP(bChannelNumber);
	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;

	if(ATAModuleReceiveTask[bChannelNumber] != TRUE)
		return (camBUSY);
	
	if ( gChnlConfig[bChannelNumber].bCh == camCHNL_NOT_FOUND )
		return (camCHNL_NOT_FOUND);	
		
	if (bChannelNumber >= MAX_ATA_MODULE)
		return (camCHNL_NOT_FOUND);

	for( i=0;i<MAX_ID_PER_CHNL;i++ ){
		ATASortHead[bChannelNumber][i] = (PCAM_ATA_PACKET)PTR_NULL;
		ATASortLBAl[bChannelNumber][i] = ZERO_NULL;
		ATASortLBAh[bChannelNumber][i] = ZERO_NULL;	
		gDrvConfig[bChannelNumber][i].DevFlag = DEV_NOTFOUND;		
	}
	
#ifdef Encryption
	if(gAdapterConfig[Adapter_ID].Features & WITH_EncryKEY)
		WriteIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_Encry_SeedKey, gAdapterConfig[Adapter_ID].EncryptionKey);
	
	/*WriteIndexUchar(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_Ext_Ctrl, 0x02 );*/
#endif	

	/* disable receive task.*/
	gChnlResetTable[bChannelNumber].receiveStatus=ATAModuleReceiveTask[bChannelNumber];
	ATAModuleReceiveTask[bChannelNumber]=FALSE;
		
	/* soft reset ATA Module ANYWAY. */
	ATASoftResetModule(bChannelNumber);	
	NeedCheckFIFO[bChannelNumber] = TRUE;

	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA ){
		U8		bCh =  bChannelNumber%MAX_ATA_CHNL;
		U32		ldata = 0;
		U8		status=0;

		if((CAM_SATA_PHY_Enable(bChannelNumber)) != camSUCCESS){
			ATAModuleReceiveTask[bChannelNumber]=gChnlResetTable[bChannelNumber].receiveStatus;	
			return(camFAIL);	/* no drive connected */							
		}
		CAM_SATA_Port_Multiplier_Update_status(bChannelNumber);
		
		/* wait SATA connect well in 1 second */
		for ( i = 0; i < 100000; i++ )
		{
			ReadPortUlong( HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG, ldata );
			ldata >>= bCh;
			if ( ldata & 0x100 )
				break;
			camUStallExecution(1);
		}		
		if ( i >= 100000 ){
			ATAModuleReceiveTask[bChannelNumber]=gChnlResetTable[bChannelNumber].receiveStatus;	
			return(camFAIL);	/* no drive connected */				
		}

		/* phy ready, but drive not update status */
		status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);
		if (status == 0xFF){ 
			if ( CAM_HardResetSATAChannel(bChannelNumber) == camDEVICE_NOT_FOUND ) {
				CAM_Clear_UnmaskPlugUnplugFlags(bChannelNumber);
				ATAModuleReceiveTask[bChannelNumber]=gChnlResetTable[bChannelNumber].receiveStatus;	
				return(camFAIL);	/* no drive connected */				
			}
		}
	}
	gChnlResetTable[bChannelNumber].notify = notify;
	gChnlResetTable[bChannelNumber].OwnerType = WAIT_INIT_CH;
	gChnlResetTable[bChannelNumber].run_count = 0;
	/*camTimer(Adapter_ID, 1, ATA_Polling_Channel_Function[bChannelNumber]);*/
	ATA_Polling_Channel(bChannelNumber);
	return camSUCCESS;
}




void CAM_Reset_ReinitATA_CallBack(U8 bChannelNumber, U8 status)
{
	U8	i, j;
	U8  found_disk=FALSE;
		
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA )
	{

		if(status != camSUCCESS){
			for (j=0;j<gChnlDrvControlConfig[bChannelNumber].Support_device;j++) {
				if ( gDrvConfig[bChannelNumber][j].DevFlag & DEV_EXIST ){
					gDrvConfig[bChannelNumber][j].DevFlag = DEV_NOTFOUND;
					cam_SetStatusByPlug_unplug(bChannelNumber, j, NO_DRIVE);
				}
			}
			gChnlDrvControlConfig[bChannelNumber].PM = FALSE;
			gChnlDrvControlConfig[bChannelNumber].Support_device = 1;
			found_disk=FALSE;
			CAM_Clear_UnmaskPlugUnplugFlags(bChannelNumber);
			
			if(gChnlResetTable[bChannelNumber].OwnerType == WAIT_TIMEOUT)
				ATA_Timeout_CallBack(bChannelNumber, camFAIL);
			else if(gChnlResetTable[bChannelNumber].OwnerType == WAIT_RESET_CH)
				CAM_ResetATAChannel_CallBack(bChannelNumber, camFAIL);
			
			return;						
		}
		
			/* check Port_Multiplier */
#ifdef Port_Multiplier
		if (CAM_Init_SATA_Port_Multiplier(bChannelNumber) != camFAIL) {
			/*ATA_Port_Multiplier_TimerID[bChannelNumber] = camTimer(((U8)(bChannelNumber/MAX_ATA_CHNL)),CAMTimeoutTimes[bChannelNumber/MAX_ATA_CHNL].Port_Multiplier_Check, ATA_Check_Port_Multiplier_Function[bChannelNumber]);*/
		}
		else {
			gChnlDrvControlConfig[bChannelNumber].PM = FALSE;
			gChnlDrvControlConfig[bChannelNumber].Support_device = 1;
		}
#else
		gChnlDrvControlConfig[bChannelNumber].PM = FALSE;
		gChnlDrvControlConfig[bChannelNumber].Support_device = 1;
#endif
					
		/* init disks */
		for (i=0;i<gChnlDrvControlConfig[bChannelNumber].Support_device;i++) {
			if ( gDrvConfig[bChannelNumber][i].DevFlag & DEV_EXIST ) {
				gDrvConfig[bChannelNumber][i].Features &= ~DEV_EN_SMART;
				if (CAMInitDevice(&(gDrvConfig[bChannelNumber][i])) != camFAIL) 
					found_disk=TRUE;
			}
		}
		if (found_disk != TRUE){
			if ( CAM_HardResetSATAChannel(bChannelNumber) == camDEVICE_NOT_FOUND ){
				for (j=0;j<gChnlDrvControlConfig[bChannelNumber].Support_device;j++) {
					if ( gDrvConfig[bChannelNumber][j].DevFlag & DEV_EXIST ){
						gDrvConfig[bChannelNumber][j].DevFlag = DEV_NOTFOUND;
						cam_SetStatusByPlug_unplug(bChannelNumber, j, NO_DRIVE);
					}
				}
	
				gChnlDrvControlConfig[bChannelNumber].PM = FALSE;
				gChnlDrvControlConfig[bChannelNumber].Support_device = 1;
				CAM_Clear_UnmaskPlugUnplugFlags(bChannelNumber);

				if(gChnlResetTable[bChannelNumber].OwnerType == WAIT_TIMEOUT)
					ATA_Timeout_CallBack(bChannelNumber, camFAIL);
				else if(gChnlResetTable[bChannelNumber].OwnerType == WAIT_RESET_CH)
					CAM_ResetATAChannel_CallBack(bChannelNumber, camFAIL);
				return;
			}
			else{
				ATA_Polling_Channel_TimerID[bChannelNumber] = camTimer((U8)(bChannelNumber/MAX_ATA_CHNL), 1, ATA_Polling_Channel_Function[bChannelNumber]);
			}
		}
		else{
			CAM_Clear_UnmaskPlugUnplugFlags(bChannelNumber);

			if(gChnlResetTable[bChannelNumber].OwnerType == WAIT_TIMEOUT)
				ATA_Timeout_CallBack(bChannelNumber, camSUCCESS);
			else if(gChnlResetTable[bChannelNumber].OwnerType == WAIT_RESET_CH)
				CAM_ResetATAChannel_CallBack(bChannelNumber, camSUCCESS);
		}
	}
	else	/* parallel channel */
	{
		/* init disks */
		found_disk=FALSE;
		if ( gDrvConfig[bChannelNumber][0].DevFlag & DEV_EXIST ) {
			gDrvConfig[bChannelNumber][0].Features &= ~DEV_EN_SMART;
			if (CAMInitDevice(&(gDrvConfig[bChannelNumber][0])) == camFAIL) {
				gDrvConfig[bChannelNumber][0].DevFlag = DEV_NOTFOUND;
				cam_SetStatusByPlug_unplug(bChannelNumber, 0, NO_DRIVE);
			}
			else
				found_disk=TRUE;
		}	
							
		if ( gDrvConfig[bChannelNumber][1].DevFlag & DEV_EXIST ) {
			gDrvConfig[bChannelNumber][1].Features &= ~DEV_EN_SMART;
			if (CAMInitDevice(&(gDrvConfig[bChannelNumber][1])) == camFAIL) {
				gDrvConfig[bChannelNumber][1].DevFlag = DEV_NOTFOUND;
				cam_SetStatusByPlug_unplug(bChannelNumber, 1, NO_DRIVE);
			}
			else
				found_disk=TRUE;			
		}
		if(found_disk == TRUE){	
			if(gChnlResetTable[bChannelNumber].OwnerType == WAIT_TIMEOUT)
				ATA_Timeout_CallBack(bChannelNumber, camSUCCESS);
			else if(gChnlResetTable[bChannelNumber].OwnerType == WAIT_RESET_CH)
				CAM_ResetATAChannel_CallBack(bChannelNumber, camSUCCESS);
		}
		else{
			if(gChnlResetTable[bChannelNumber].OwnerType == WAIT_TIMEOUT)
				ATA_Timeout_CallBack(bChannelNumber, camFAIL);
			else if(gChnlResetTable[bChannelNumber].OwnerType == WAIT_RESET_CH)
				CAM_ResetATAChannel_CallBack(bChannelNumber, camFAIL);

		}
	}
	return;
}


U8 CAM_Reset_ReinitATA(U8 bChannelNumber)
{
	U8	Adapter_ID;
	U8	bData=0;
	U8	bCh = bChannelNumber%MAX_ATA_CHNL;
	U32	i;
	U8	j;
	U32 ldata;
	U8	disk_exist;

	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;

	/* Ask ATA engine don't execute the next command packet, so the ata module will stop after done current packet. */
	WritePortUchar(ATABaseAddr[bChannelNumber], ATA_STATUS_REGISTER_OFFSET, 1 );		
	for (i = 0; i < 300000; i++ ) {	
		ReadPortUchar(ATABaseAddr[bChannelNumber], (ATA_STATUS_REGISTER_OFFSET+3), bData);
		if ( !(bData & 0x20) )
			break;	/* not in packet command cycle. */
		if (i==0) {
			if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA ) {
				ReadPortUlong( HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG, ldata );
				ldata >>= (bChannelNumber%MAX_ATA_CHNL);
				if (!( ldata & 0x100 )) /* not well connect */
					break;
			}
		}
		camUStallExecution(1);	/* wait 3 secs */
	} ;	
	/* soft reset ATA Module. */
	ATASoftResetModule(bChannelNumber);
	NeedCheckFIFO[bChannelNumber] = TRUE;
		
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA ){
		if ( CAM_HardResetSATAChannel(bChannelNumber) == camDEVICE_NOT_FOUND ){
			/* no drive connected */
			for (j=0;j<gChnlDrvControlConfig[bChannelNumber].Support_device;j++) {
				if ( gDrvConfig[bChannelNumber][j].DevFlag & DEV_EXIST ){
					gDrvConfig[bChannelNumber][j].DevFlag = DEV_NOTFOUND;
					cam_SetStatusByPlug_unplug(bChannelNumber, j, NO_DRIVE);
				}
			}
			gChnlDrvControlConfig[bChannelNumber].PM = FALSE;
			gChnlDrvControlConfig[bChannelNumber].Support_device = 1;

			CAM_Clear_UnmaskPlugUnplugFlags(bChannelNumber);
			return(camDEVICE_NOT_FOUND);	/* no drive connected */	
		}
		else{ /* for Plug soon after UnPlug */
			disk_exist = FALSE;
			for (j=0;j<gChnlDrvControlConfig[bChannelNumber].Support_device;j++) {
				if ( gDrvConfig[bChannelNumber][j].DevFlag & DEV_EXIST )
					disk_exist = TRUE;
			}		
			
			if(disk_exist != TRUE){	/* device Plug */
				gChnlDrvControlConfig[bChannelNumber].PM = FALSE;
				gChnlDrvControlConfig[bChannelNumber].Support_device = 1;
	
				CAM_Clear_UnmaskPlugUnplugFlags(bChannelNumber);
				return(camWARNING);	/* no drive connected */	
			}			
		}
	}
	else	/* parallel channel */
	{
		CAM_NotAtCommandPacketPhase(bChannelNumber);

		/* When Drive DMARQ still exist, we can't do any PIO access to ATA bus. */
		/* But do hard reset, that will reset 199. It should be a problem for dual master mode!*/	
		/* do hard reset */ 
		ReadPortUchar(HostBaseAddr[Adapter_ID],oFFSETHOSTPCICTL+1 , bData );
		bData &= ~ (0x10 << bCh);
		WritePortUchar(HostBaseAddr[Adapter_ID],oFFSETHOSTPCICTL+1 , bData );
		camStallExecution(1);
		bData |= (0x10 << bCh);
		WritePortUchar(HostBaseAddr[Adapter_ID],oFFSETHOSTPCICTL+1 , bData );

		/* soft reset ATA Module. */
		ATASoftResetModule(bChannelNumber);
	}
	ATA_Polling_Channel_TimerID[bChannelNumber] = camTimer(Adapter_ID, 1, ATA_Polling_Channel_Function[bChannelNumber]);
	/*ATA_Polling_Channel(bChannelNumber);*/
	return(camSUCCESS);	
}


void CAM_ResetATAChannel_CallBack(U8 bChannelNumber, U8 status)
{
	U8	Adapter_ID;
	U8	j;
		
	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;
	
	ATAModuleReceiveTask[bChannelNumber]=gChnlResetTable[bChannelNumber].receiveStatus;
	
	if (status == camFAIL){
		gChnlResetTable[bChannelNumber].notify(CAM_ATA_Chnl_OutMAP(bChannelNumber), camFAIL);
		return;
	}
	
	/* must check module is need to push, By it maybe pushed when no SQEID. */
	for (j=(Adapter_ID*MAX_ATA_CHNL);j<((Adapter_ID+1)*MAX_ATA_CHNL);j++) {
		if (ATAModuleNeedPush[j]==TRUE){
			/* start ata module */
			StartATAOperation(j);
				break;
		}
	}
	
	gChnlResetTable[bChannelNumber].notify(CAM_ATA_Chnl_OutMAP(bChannelNumber), camSUCCESS);			
	return;
}


U8 CAM_ResetATAChannel(U8 bChannelNumber, void (*notify)(U8 bCh, U8 status))
{
	U8		Adapter_ID;
	U8		SEQID;
	U8		i;
	PSEQ_INFO			pCurSEQ_INFO=(PSEQ_INFO)PTR_NULL;
	PCAM_PACKET			ultmpATAPacketAddress;
	PCAM_PACKET			ultmpATAPacketAddress_next;
	PATA_TASK			pATATask;
	U8		TAG_Flag=FALSE;
	U8		SEQCntCtrl;
	U32		tmpData;
	U8		CRC_err=FALSE;	
	U8	 	STOPFlag=FALSE;
	U8		ResetStatus;
	

	bChannelNumber=CAM_ATA_Chnl_InMAP(bChannelNumber);
	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;
		
	if(ATAModuleReceiveTask[bChannelNumber] != TRUE){
		if(gChnlResetTable[bChannelNumber].OwnerType != WAIT_TIMEOUT)
			return (camBUSY);
		
		if(gChnlResetTable[bChannelNumber].receiveStatus != TRUE)
			return (camBUSY);
			
		ATAModuleReceiveTask[bChannelNumber] = TRUE;
		if(ATA_Polling_Channel_TimerID[bChannelNumber] != ZERO_NULL){
			camTimer(Adapter_ID,0,(void (*)(void))ATA_Polling_Channel_TimerID[bChannelNumber]);
			ATA_Polling_Channel_TimerID[bChannelNumber] = ZERO_NULL;
		}
		ATAModuleSEQID[bChannelNumber] = gChnlResetTable[bChannelNumber].SEQID;
	}
		
	if ( gChnlConfig[bChannelNumber].bCh  == camCHNL_NOT_FOUND )
		return (camCHNL_NOT_FOUND);	

	if (bChannelNumber >= MAX_ATA_MODULE )
		return (camCHNL_NOT_FOUND);	

	for( i=0;i<MAX_ID_PER_CHNL;i++ ){
		ATASortHead[bChannelNumber][i] = (PCAM_ATA_PACKET)PTR_NULL;
		ATASortLBAl[bChannelNumber][i] = ZERO_NULL;
		ATASortLBAh[bChannelNumber][i] = ZERO_NULL;			
	}
	
	/* disable receive task.*/
	gChnlResetTable[bChannelNumber].receiveStatus = ATAModuleReceiveTask[bChannelNumber];
	ATAModuleReceiveTask[bChannelNumber]=FALSE;
		
	/* turn off the LED */
	if ((gBOXConfig[bChannelNumber].BOXType != camNOBOX) && (gBOXConfig[bChannelNumber].BOXType & SATAEnclosure)) { 
		/* turn off LED */
		CAM_SBOX_HDAccess_LED(bChannelNumber, HDActiveOFF);
	}
		
	SEQID = ATAModuleSEQID[bChannelNumber];
	if (SEQID != U8NULL)
	{
		ATAModuleSEQID[bChannelNumber] = U8NULL;
		pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SEQID]);
		ultmpATAPacketAddress = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
		pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
		pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;
		
		if ((PTR_SIZE)ultmpATAPacketAddress != PTR_NULL) {
			if ( ((PPACKET_HEAD)ultmpATAPacketAddress)->Flag & (PACFLAG_NCQ|PACFLAG_QUEUED) )
				TAG_Flag=TRUE;
		}
		
		gChnlDrvControlConfig[bChannelNumber].TagCmdFlag &= ~(TAG_DATAIN|TAG_DATAOUT);
		gChnlDrvControlConfig[bChannelNumber].FreeTAG = 0xFFFFFFFF;
		gChnlDrvControlConfig[bChannelNumber].UsedTAG = 0x0;
	}
	else
		ultmpATAPacketAddress = (PCAM_PACKET)PTR_NULL;
	
	gChnlDrvControlConfig[bChannelNumber].TCQTransferData=FALSE;
	
	/* remove ata timeout timer */
	if (ATA_Timeout_TimerID[bChannelNumber]!=ZERO_NULL){
		camTimer(Adapter_ID,0,(void (*)(void))ATA_Timeout_TimerID[bChannelNumber]);	
		ATA_Timeout_TimerID[bChannelNumber]=ZERO_NULL;
	}


	/* check with executing packets in seqID */
	if((PTR_SIZE)ultmpATAPacketAddress != PTR_NULL) {
		
		/* read Sequence Counter Control Register # */
		ReadPortUchar(HostBaseAddr[Adapter_ID], SEQID*4, SEQCntCtrl);
		SEQCntCtrl &= 0x1F;
		
		if(gAdapterConfig[Adapter_ID].TapeOutVersion < REVISION_2){/* ASIC tapout prior revision 2 */	
			if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) {/* SATA */
				/* check is CRC_err */
				#ifdef	_MMIO_
					ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATAError, tmpData );
				#else
					ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDESATAError, tmpData );
				#endif
				if (tmpData & CRC_ERR) {	/* CRC error */
					CRC_err=TRUE;
				}
			}
		}
	
		/* if NCQ commands, abort all commands */
		if (TAG_Flag==TRUE)
			pCurSEQ_INFO->PACCNT=SEQCntCtrl;
		/* if CRC_err, abort all commands */
		else if (CRC_err==TRUE)
			pCurSEQ_INFO->PACCNT=SEQCntCtrl;			
		/* ATAPI PIO SEQCntCtrl will be zero or if error, current packet maybe discount or not */
		else if (SEQCntCtrl < pCurSEQ_INFO->PACCNT)
			SEQCntCtrl++;
					
		while ((PTR_SIZE)ultmpATAPacketAddress != PTR_NULL){						

			pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;
			
			/* check is error packet */
			if (pCurSEQ_INFO->PACCNT == SEQCntCtrl)
			{	/* current packet */

				/* abort packets queue at the rear of this packet and and wait queue packets. */
				Abort_All_Undone_ATAPackets(bChannelNumber,(PCAM_PACKET)ultmpATAPacketAddress->PacketHead.NPA);
				ultmpATAPacketAddress->PacketHead.NPA=(PVOID)PTR_NULL;			

				pATATask->ATAStatus = ATA_ABORT;
				STOPFlag = TRUE;
			}
			else
			{
				pCurSEQ_INFO->PACCNT--;
				pATATask->ATAStatus= 0;		/* no error */
				if (pATATask->ATACmdFlag & ATA_STATUS_RETURN) /* need read it's ata registers statses when command success completed */
					pATATask->ATAStatus = ATA_ABORT;
					/*ATAErrorStatus(pATATask);*/
			}
								
			pATATask->ATACmdFlag &= ~INTERNAL_SENSE;

			/* free sg resource */
			CAMFreeATASG(bChannelNumber, (PVOID)ultmpATAPacketAddress->PacketHead.PSG);	
			
			ultmpATAPacketAddress_next = (PCAM_PACKET)ultmpATAPacketAddress->PacketHead.NPA;
			
			/* free packet resource. */
			CAMFreeCamPacket(bChannelNumber, (PVOID)ultmpATAPacketAddress);		
			
			/* next packet */
			ultmpATAPacketAddress = ultmpATAPacketAddress_next;

			pATATask->callback(pATATask);
			
			if( STOPFlag == TRUE)
				break;
		}
		pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
		pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;
		ATAModuleBusy[bChannelNumber]=FALSE;
		if (TAG_Flag!=TRUE)
			CAMFreeSEQID(Adapter_ID,SEQID);
	}
	else{
		/* no executing packets,abort packets queue at wait queue packets. */
		Abort_All_Undone_ATAPackets(bChannelNumber,(PCAM_PACKET)PTR_NULL);
	}

	gChnlResetTable[bChannelNumber].notify = notify;
	gChnlResetTable[bChannelNumber].OwnerType = WAIT_RESET_CH;
	gChnlResetTable[bChannelNumber].run_count = 0;

	ResetStatus = CAM_Reset_ReinitATA(bChannelNumber);
	if(ResetStatus != camSUCCESS){
		ATAModuleReceiveTask[bChannelNumber] = gChnlResetTable[bChannelNumber].receiveStatus;
		if(ResetStatus == camWARNING)
			cam_SetStatusByPlug_unplug(bChannelNumber, 0, NEW_DRIVE);
		return(camFAIL);
	}
	return(camSUCCESS);
}


/*-------------------------------------------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------*/

U8 CAMInitDevice(PDEV_CONFIG pDevConfig)
{
	U8			bChannelNumber=pDevConfig->bCh;
	U8			bID=pDevConfig->bID;
	U8			statusByte= 0;
	U8			Adapter_ID = bChannelNumber/MAX_ATA_CHNL;
	PTR_SIZE		BaseAddress = ATABaseAddr[bChannelNumber];
	U8			i;

	
	/* disable INTA here, it will be re-enable when CAM use SEQ 0 for packets */
	WriteIndexUchar(BaseAddress, iNDEXIDECICR, 0x0);  /* that the drive INT pass to SEQ 0*/
	WritePortUchar( HostBaseAddr[Adapter_ID], 0, 0x20); /* but mask SEQ 0 INT */

	/* In parallel, prevent single slave condition */
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_ATA ){ /* PATA */
		/* wait for BUSY */
		for (i = 0; i < 10; i++) {
			statusByte = WaitOnBusy(ATABaseAddr[bChannelNumber]);
			if( !(statusByte & IDE_STATUS_BUSY))
				break;
		}		
		
		/* check if device exist */ 
		/* select drive */	
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEDeviceHead, (U8)(((bID & 0x01) << 4) | 0xA0));
		/* read drive status */
		if ( WaitOnBaseBusy(ATABaseAddr[bChannelNumber]) & 0x80) {
			return (camFAIL); /* no device */
		}
		
		/* select drive  again */
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEDeviceHead, (U8)(((bID & 0x01) << 4) | 0xA0));
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECylinderLow, 0x55 );
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECylinderHigh, 0xAA );
		statusByte = ReadIndexUchar( ATABaseAddr[bChannelNumber], iNDEXIDECylinderLow);
		if ( statusByte != 0x55 ){ /* no device */
			return(camFAIL);
		}
	}
	else { /* SATA */
		/* set PM Port */
		WriteIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_PM_Port, bID);
	}

#ifdef _ATAPI_
	/* ATA or ATAPI device ? */
	AtapiSoftReset(bChannelNumber, bID);
#endif
				
	/* set paramters */	
	if ( pDevConfig->DevFlag & DEV_ATA_DEVICE )
	{
		/* select drive */	
		if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */
			WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, 0xE0);
		else
			WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, (U8)(((bID & 0x01) << 4) | 0xA0));
			
 		statusByte = WaitOnBusy(BaseAddress);
	    	statusByte &= ~(IDE_STATUS_ERROR | IDE_STATUS_INDEX | IDE_STATUS_DRQ );
	    	if ( (statusByte != IDE_STATUS_IDLE) && (statusByte != IDE_STATUS_DRDY) )
				return(camFAIL); /* bad device */
			
		/* send command */
		WriteIndexUchar(BaseAddress, iNDEXIDESectorCount , pDevConfig->SecPerTrk);
		if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */
			WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, (U8)(pDevConfig->HeadNum - 1) );
		else
			WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, (U8)((pDevConfig->HeadNum - 1)|(bID << 4)) );
		WriteIndexUchar(BaseAddress, iNDEXIDECommand, IDE_COMMAND_SET_DRIVE_PARAMETERS);
		if ( WaitForINT(BaseAddress) != camSUCCESS ){
			return(camFAIL); /* bad device */
		}

	#ifdef PIORW
		if ((pDevConfig->MaximumBlockTransfer!=0) & (pDevConfig->CurrentBlockTransfer==0) ) {
			/* set to MAX sectors per block*/

			/* select drive */	
			if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */
				WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, 0xE0);
			else
				WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, (U8)(((bID & 0x01) << 4) | 0xA0));
				
	 		statusByte = WaitOnBusy(BaseAddress);
		    	statusByte &= ~(IDE_STATUS_ERROR | IDE_STATUS_INDEX | IDE_STATUS_DRQ );
		    	if ( (statusByte != IDE_STATUS_IDLE) && (statusByte != IDE_STATUS_DRDY) )
					return(camFAIL); /* bad device */
				
			/* send command */
			WriteIndexUchar(BaseAddress, iNDEXIDESectorCount , 1);

			WriteIndexUchar(BaseAddress, iNDEXIDECommand, IDE_COMMAND_SET_MULTIPLE_MODE);
			if ( WaitForINT(BaseAddress) != camSUCCESS ){
				return(camFAIL); /* bad device */
			}
			if (ReadIndexUchar(BaseAddress, iNDEXIDEStatus) & IDE_STATUS_ERROR) {
				pDevConfig->CurrentBlockTransfer=pDevConfig->MaximumBlockTransfer=0;
			}
			else
				pDevConfig->CurrentBlockTransfer=1;

		}
	#endif
			
	}
	
	/* set feature */
	WriteIndexUchar(BaseAddress, iNDEXIDEFeature , 0x03);	/* set transfer mode */
	WriteIndexUchar(BaseAddress, iNDEXIDESectorCount, (U8)(pDevConfig->PIOMode | 0x08));	/* PIO transfer */
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */
		WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, 0xE0);
	else	
		WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, (U8)(0xE0 |(bID << 4)) );
	WriteIndexUchar(BaseAddress, iNDEXIDECommand, IDE_COMMAND_SET_FEATURES);
	
	if ( pDevConfig->DevFlag & DEV_ATA_DEVICE ) { /* ATA */
		if ( WaitForINT(BaseAddress) != camSUCCESS ){
			return(camFAIL); /* bad device */
		}
	}
	else { /* ATAPI */
		if ( WaitForATAPIINT(bChannelNumber) != camSUCCESS ){
			return(camFAIL); /* bad device */
		}		
	}

	WriteIndexUchar(BaseAddress, iNDEXIDEFeature, 0x03);	/* set transfer mode */
	if ( pDevConfig->UDMAMode != 0xFF )	{	
		WriteIndexUchar(BaseAddress, iNDEXIDESectorCount, (U8)(pDevConfig->UDMAMode | 0x40));
	}
	else {
		WriteIndexUchar(BaseAddress, iNDEXIDESectorCount, (U8)(pDevConfig->MDMAMode | 0x20));
	}
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */
		WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, 0xE0);
	else	
		WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, (U8)(0xE0 |(bID << 4)) );
	WriteIndexUchar(BaseAddress, iNDEXIDECommand, IDE_COMMAND_SET_FEATURES);
	if ( pDevConfig->DevFlag & DEV_ATA_DEVICE ) { /* ATA */
		if ( WaitForINT(BaseAddress) != camSUCCESS ){
			return(camFAIL); /* bad device */
		}
	}
	else { /* ATAPI */
		if ( WaitForATAPIINT(bChannelNumber) != camSUCCESS ){
			return(camFAIL); /* bad device */
		}		
	}	
	
	if (( pDevConfig->DevFlag & DEV_ATA_DEVICE )
	 && ( pDevConfig->Features & DEV_SMART ) )
	{
		if (!(pDevConfig->Features & DEV_EN_SMART) ){
			/* enable smart */
			WriteIndexUchar(BaseAddress, iNDEXIDEFeature , 0xD8);	
			WriteIndexUchar(BaseAddress, iNDEXIDECylinderLow, 0x4F);	
			WriteIndexUchar(BaseAddress, iNDEXIDECylinderHigh, 0xC2);
			if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */
				WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, 0xE0);
			else		
				WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, (U8)(0xE0 |(bID << 4)) );
			WriteIndexUchar(BaseAddress, iNDEXIDECommand, IDE_COMMAND_SMART_CMD);
			if ( WaitForINT(BaseAddress) != camSUCCESS ){
				pDevConfig->Features &= ~DEV_EN_SMART;	
				return(camFAIL); /* bad device */
			}
			pDevConfig->Features |= DEV_EN_SMART;
		}
	}

	if ( (pDevConfig->Features & DEV_NCQ_DMASetupAuto)
	 && (!(pDevConfig->Features & DEV_EN_NCQ_DMASetupAuto)) )
	{
		/* enable DMA Setup Auto-Activate optimization support */
		WriteIndexUchar(BaseAddress, iNDEXIDEFeature, 0x10); /* Enable use of Serial ATA feature */
		WriteIndexUchar(BaseAddress, iNDEXIDESectorCount, 0x02);
		if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */
			WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, 0xE0);
		else		
			WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, (U8)(0xE0 |(bID << 4)) );
		WriteIndexUchar(BaseAddress, iNDEXIDECommand, IDE_COMMAND_SET_FEATURES);
		if ( pDevConfig->DevFlag & DEV_ATA_DEVICE ) { /* ATA */
			if ( WaitForINT(BaseAddress) != camSUCCESS ){
				return(camFAIL); /* bad device */
			}
		}
		else { /* ATAPI */
			if ( WaitForATAPIINT(bChannelNumber) != camSUCCESS ){
				return(camFAIL); /* bad device */
			}		
		}
		pDevConfig->Features |= DEV_EN_NCQ_DMASetupAuto;
	}	

	/* not support non-zero buffer offsets in DMA Setup FIS */
	if ( (pDevConfig->Features & DEV_NCQ_non_zero_offsets)
	 && (pDevConfig->Features & DEV_EN_NCQ_non_zero_offsets) )
	{
		/* disable non-zero buffer offsets in DMA Setup FIS support */
		WriteIndexUchar(BaseAddress, iNDEXIDEFeature, 0x90); /* Disable use of Serial ATA feature */
		WriteIndexUchar(BaseAddress, iNDEXIDESectorCount, 0x01);
		if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */
			WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, 0xE0);
		else		
			WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, (U8)(0xE0 |(bID << 4)) );
		WriteIndexUchar(BaseAddress, iNDEXIDECommand, IDE_COMMAND_SET_FEATURES);
		if ( pDevConfig->DevFlag & DEV_ATA_DEVICE ) { /* ATA */
			if ( WaitForINT(BaseAddress) != camSUCCESS ){
				return(camFAIL); /* bad device */
			}
		}
		else { /* ATAPI */
			if ( WaitForATAPIINT(bChannelNumber) != camSUCCESS ){
				return(camFAIL); /* bad device */
			}		
		}
		pDevConfig->Features &= ~DEV_EN_NCQ_non_zero_offsets;
	}
					
	return(camSUCCESS); /* OK */
}

U8 IssueIdentify(U8 bChannelNumber, U8 bID, PU8 pBuf, U8 command) 
{
	U8	statusByte;
	PTR_SIZE		BaseIOAddress = ATABaseAddr[bChannelNumber];

		
	/* select drive */
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */
		WriteIndexUchar(BaseIOAddress, iNDEXIDEDeviceHead, 0xE0);
	else		
		WriteIndexUchar(BaseIOAddress, iNDEXIDEDeviceHead, (U8)(((bID & 0x01) << 4) | 0xA0));
	GetBaseStatus(BaseIOAddress, statusByte);	
	if (command == IDE_COMMAND_IDENTIFY_DEVICE)
	{
    		statusByte &= ~(IDE_STATUS_ERROR | IDE_STATUS_INDEX | IDE_STATUS_DRQ );
	    	if ( (statusByte != IDE_STATUS_IDLE) && (statusByte != IDE_STATUS_DRDY) )
			return(camFAIL); /* bad device */	
	}
    	else
    	{
		/*
	    	* Load CylinderHigh and CylinderLow with number bytes to transfer.
	    	*/
	
	    	WriteIndexUchar(BaseIOAddress,iNDEXIDECylinderHigh, (0x200 >> 8));
	    	WriteIndexUchar(BaseIOAddress,iNDEXIDECylinderLow,  (0x200 & 0xFF));
	}
	
   	/* Send IDENTIFY command. */
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */
		WriteIndexUchar(BaseIOAddress, iNDEXIDEDeviceHead, 0xE0);
	else	   	
		WriteIndexUchar(BaseIOAddress, iNDEXIDEDeviceHead, (U8)(((bID & 0x01) << 4) | 0xA0));
   	WriteIndexUchar(BaseIOAddress, iNDEXIDECommand, command);
   	if (command == IDE_COMMAND_IDENTIFY_DEVICE) { /* ATA */
		if ( WaitForINT(BaseIOAddress) != camSUCCESS )
			return(camFAIL); /* bad device */   		
   	}
   	else { /* ATAPI */
		if ( WaitForATAPIINT(bChannelNumber) != camSUCCESS )
			return(camFAIL); /* bad device */
	}   	

    	/* Wait for DRQ */
   	statusByte = WaitForDrq(BaseIOAddress);
   	
   	if (!(command == IDE_COMMAND_IDENTIFY_DEVICE)) /* ATAPI */
   		ReadIndexUchar(BaseIOAddress, iNDEXIDEStatus); /*clear possible device INT */
   		
	if ( (!(statusByte & IDE_STATUS_DRQ)) | (statusByte& IDE_STATUS_BUSY)) {	
		return(camFAIL); /* bad device */	
	}

	/*
	* Check for error on really stupid master devices that assert random
	* patterns of bits in the status register at the slave address.
	*/

    	if ((command == IDE_COMMAND_IDENTIFY_DEVICE) && (statusByte & IDE_STATUS_ERROR)) {
		return(camFAIL); /* bad device */	
    	}

	/*
	 * Suck out 256 words. 
	 */
	 
    	if( (ReadBuffer(bChannelNumber,(PU16)pBuf, 256)) != camSUCCESS)
    		return(camFAIL);
    
	/* check status again, if DRQ still is 1. This command is failed. */	
	GetBaseStatus(BaseIOAddress, statusByte);
    	if ( statusByte & IDE_STATUS_DRQ )
			return(camFAIL); /* bad device */		    

	return(camSUCCESS); /* OK */	
}

U8 IssueService(U8 bChannelNumber, U8 bID, U8 command) 
{
		
	/* select drive */
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEDeviceHead, 0xE0);
	else		
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEDeviceHead, (U8)(((bID & 0x01) << 4) | 0xA0));
   	/* Send SERVICE command. */
   	WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECommand, command);

	return(camSUCCESS); /* OK */	
}

void cam_SetStatusByPlug_unplug(U8 bChannelNumber, U8 bID, U8 Status)
{
	U8 Adapter_ID=bChannelNumber/MAX_ATA_CHNL;
	PCAM_PACKET ultmpATAPacketAddress;
	PSEQ_INFO	pCurSEQ_INFO=(PSEQ_INFO)PTR_NULL;
	U8	SEQID;	
	U8	receiveStatus;
	
	if (Status == NO_DRIVE) {
		gDrvConfig[bChannelNumber][bID].DevFlag = DEV_NOTFOUND;
		
		if (bID == 0) {
			if (gChnlDrvControlConfig[bChannelNumber].TagCmdFlag & TAG_SEQALLOC) {
				CAMFreeSEQID(Adapter_ID,gChnlDrvControlConfig[bChannelNumber].SEQID);
			}
			gChnlDrvControlConfig[bChannelNumber].TagCmdFlag &= ~TAG_SEQALLOC;
		
			gChnlDrvControlConfig[bChannelNumber].SEQID = U8NULL;
		}
		
		SEQID = ATAModuleSEQID[bChannelNumber];
		if (SEQID != U8NULL)
		{
			ATAModuleSEQID[bChannelNumber] = U8NULL;
			pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SEQID]);
			ultmpATAPacketAddress = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
			pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
			pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;
			ATAModuleBusy[bChannelNumber]=FALSE;
		}
		else 
			ultmpATAPacketAddress = (PCAM_PACKET)PTR_NULL;	
		

		/* remove ata timeout timer */
		if (ATA_Timeout_TimerID[bChannelNumber] != ZERO_NULL) {
			camTimer(Adapter_ID,0,(void (*)(void))ATA_Timeout_TimerID[bChannelNumber]);	
			ATA_Timeout_TimerID[bChannelNumber]=ZERO_NULL;
		}
		/* disable receive task.*/
		receiveStatus = ATAModuleReceiveTask[bChannelNumber];
		ATAModuleReceiveTask[bChannelNumber] = FALSE;

		Abort_All_Undone_ATAPackets(bChannelNumber,ultmpATAPacketAddress);
		
		ATAModuleReceiveTask[bChannelNumber] = receiveStatus;
	}
	else { /* NEW_DRIVE */
		
	}

	bChannelNumber=CAM_ATA_Chnl_OutMAP(bChannelNumber);	
	camCheckDriveStatus(bChannelNumber, bID, Status);
	return;
}

void ATAFPDMAResetModule(U8 bChannelNumber)
{
	U8	bData=0;
	U32	ldata;
	
	/* do Reset FPDMA */ 
	ReadIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_Ctrl_Status, ldata);
	bData=(U8) ldata;
	bData &= 0x7F;
	bData |= bit(3);
	WriteIndexUchar(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_Ctrl_Status, bData );
	camStallExecution(1);
	bData &= ~bit(3);
	WriteIndexUchar(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_Ctrl_Status, bData );	
	
	FPDMAclearInterruptFlag(bChannelNumber);
}

void ATAClearInternalDebugRecordErrorRegister(U8 bChannelNumber)
{
#ifdef	_MMIO_
	WriteIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATAError, 0xFFFFFFFF );
	WriteIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEErrorsReportedfromLinkLayer, 0xFFFF0000 );
#else
	WriteIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDESATAError, 0xFFFFFFFF );
	WriteIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDEErrorsReportedfromLinkLayer, 0xFFFF0000 );
#endif
}

void ATASoftResetModule(U8 bChannelNumber)
{
	U32  tmpData;
	U32	i;

	CAM_NotAtCommandPacketPhase(bChannelNumber);

	ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDECICR, tmpData);
	tmpData |=0x00000800;	/* bit 11.*/
	WriteIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDECICR, tmpData);
	for( i = 0; i <= 10000; i++ )
	{
		ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDECICR, tmpData);
		if ( tmpData & 0x00000800 )
			break;
		camUStallExecution(1); 	
		tmpData |=0x00000800;	/* bit 11.*/
		WriteIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDECICR, tmpData);
	};
	tmpData &=0xfffff7ff;	/* bit 11.*/
	WriteIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDECICR, tmpData);
	
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA ) {
		ATAFPDMAResetModule(bChannelNumber);
		ATAClearInternalDebugRecordErrorRegister(bChannelNumber);
	}
}

void FPDMAclearInterruptFlag(U8 bChannelNumber)
{
	U32 			tmpData;
	
	/* when NCQ running, can't write FPDMA Control/Status Register whole DWORD */
	/* only clear the FPDMA Control/Status Register bit 10, 11*/
	ReadIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_Ctrl_Status, tmpData);
	tmpData |= FPDMA_Ctrl_Status_DMAsetup_INT_flag | FPDMA_Ctrl_Status_SetDb_INT_flag;
	WriteIndexUchar(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_Ctrl_Status+1, (U8) (tmpData>> 8));	
}

U8 checkLargeLBA(PATA_TASK pATATask)
{
	U8 LargeLBA=FALSE;
	/* ckeck is 48-bit LBA */
	switch(pATATask->Cmd.ATACmd.bCmd2Chip){
		case IDE_COMMAND_READ_SECTORS:
		case IDE_COMMAND_READ_DMA:
		case IDE_COMMAND_WRITE_SECTORS:
		case IDE_COMMAND_WRITE_DMA:	
		case IDE_COMMAND_READ_VERIFY_SECTORS:
		case IDE_COMMAND_FLUSH_CACHE:
			/*1.When R/W command start address is less than 128G but the end address across the 128G, cam still do 
				standard R/W command(no extended R/W command) will cause R/W datas error.
	     		  2.When R/W command, if the address less 128G, use standard R/W command, and if large than 128G, use extended R/W command,
	     			but this seem cause the drive confused.	*/
			/* check lba address large than 2^28*/
			/* if ( (pATATask->Cmd.ATACmd.LBAh) 
				|| (pATATask->Cmd.ATACmd.LBAl & 0xF0000000) ) */
			if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_48BIT)
				LargeLBA=TRUE;
			break;
		case IDE_COMMAND_READ_DMA_EXT:
		case IDE_COMMAND_READ_DMA_QUEUED_EXT:
		case IDE_COMMAND_READ_SECTORS_EXT:
		case IDE_COMMAND_READ_MULTIPLE_EXT:
		case IDE_COMMAND_WRITE_DMA_EXT:
		case IDE_COMMAND_WRITE_DMA_QUEUED_EXT:
		case IDE_COMMAND_WRITE_SECTORS_EXT:
		case IDE_COMMAND_WRITE_MULTIPLE_EXT:
		case IDE_COMMAND_READ_VERIFY_SECTORS_EXT:
		case IDE_COMMAND_FLUSH_CACHE_EXT:
				LargeLBA=TRUE;
			break;
	}
	return LargeLBA;
}

/* set disk status when error */
/* 1. return bStatus 
    2. return berror 
    3 return bCount
    4 return LBAl and LBAh */
void ATAErrorStatus(PATA_TASK pATATask)
{
	U32 Data;
	U8 bChannelNumber;
	U8 LargeLBA=FALSE;

	bChannelNumber=pATATask->bCh2Chip;

	/*    1. return bStatus , 3f6*/
	pATATask->Cmd.ATACmd.bStatus = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEAlternateStatus);
	/*    2. return berror , 1f1*/
	pATATask->Cmd.ATACmd.bError = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEError);

	/* ckeck is 48-bit LBA */
	LargeLBA=checkLargeLBA(pATATask);
	pATATask->Cmd.ATACmd.bCount = 0;
	pATATask->Cmd.ATACmd.LBAh= 0;
	pATATask->Cmd.ATACmd.LBAl = 0;		

	if (LargeLBA){
		/* most recently written */
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEDeviceControl, 0x0 ); 	/* set HOB=0 */	
		pATATask->Cmd.ATACmd.bCount = (U16)ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorCount);
		pATATask->Cmd.ATACmd.LBAl = (U32)ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorNumber);
		Data = (U32)ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECylinderLow);
		pATATask->Cmd.ATACmd.LBAl |= (Data << 8);
		Data = (U32)ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECylinderHigh);
		pATATask->Cmd.ATACmd.LBAl |= (Data << 16);
		
		/* previous content */
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEDeviceControl, 0x80 );	/* set HOB=1 */
		Data = (U32)ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorCount);
		pATATask->Cmd.ATACmd.bCount |= (U16)(Data  << 8);
		Data = (U32)ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorNumber);
		pATATask->Cmd.ATACmd.LBAl |= (Data << 24);
		pATATask->Cmd.ATACmd.LBAh = (U32)ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECylinderLow);
		Data = (U32)ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECylinderHigh);
		pATATask->Cmd.ATACmd.LBAh |= (Data << 8);			
		
	}
	else{
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEDeviceControl, 0x0 ); 	/* set HOB=0 */	
		
		/*    3 return bCount , 1f2 */
		pATATask->Cmd.ATACmd.bCount = (U16)ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorCount);

		/*    4 return LBAl and LBAh */
		Data = (U32)ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorNumber);
		pATATask->Cmd.ATACmd.LBAl |= Data;
		Data = (U32)ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECylinderLow);
		pATATask->Cmd.ATACmd.LBAl |= (Data << 8);
		Data = (U32)ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECylinderHigh);
		pATATask->Cmd.ATACmd.LBAl |= (Data << 16);
		Data = (U32)ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEDeviceHead);
		Data &= 0x0F;
		pATATask->Cmd.ATACmd.LBAl |= (Data << 24);

	}
	
	return;
}

/* set ATAPI error status */
void ATAPIErrorStatus(PATA_TASK pATATask)
{
	U8 bChannelNumber;

	bChannelNumber=pATATask->bCh2Chip;

	/* 1. ATAPI Error Register */
	pATATask->Cmd.ATACmd.bError = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEError);

	/* 2. ATAPI Status */
	pATATask->Cmd.ATACmd.bStatus = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEAlternateStatus);
	
	return;
}

void SubmitInternelSense(PATA_TASK pATATask,PCAM_PACKET CAM_ATAPacket)
{
	U8 Adapter_ID;
	PCAM_ATA_PACKET NEW_ATAPacket;

	Adapter_ID = pATATask->bCh2Chip/MAX_ATA_CHNL;
	
	/* parepare REQUEST SENSE task */
	pATATask->ATACmdFlag |= camDATA_XFER + PIO_XFER + camDATA_IN;
	pATATask->Cmd.Cdb[0] = REQUEST_SENSE_COMMAND;
	pATATask->Cmd.Cdb[1] = 0;
	pATATask->Cmd.Cdb[2] = 0;
	pATATask->Cmd.Cdb[3] = 0;
	pATATask->Cmd.Cdb[4] = (pATATask->SenseInfoLength) & (U8)~bit(0);	/* SenseInfoLength must be a multiple of 2 */
	pATATask->DataTransferLength=(U32)pATATask->Cmd.Cdb[4];
	pATATask->Cmd.Cdb[5] = 0;
	pATATask->Cmd.Cdb[6] = 0;
	pATATask->Cmd.Cdb[7] = 0;
	pATATask->Cmd.Cdb[8] = 0;
	pATATask->Cmd.Cdb[9] = 0;
	pATATask->Cmd.Cdb[10] = 0;
	pATATask->Cmd.Cdb[11] = 0;
	pATATask->SGCount = 0;	/* no SG */
	pATATask->DataBuffer = pATATask->SenseInfoBuffer;
	


	/* get free packet address */
	NEW_ATAPacket= (PCAM_ATA_PACKET)CAMAllocATAPacket(pATATask->bCh2Chip);
	if (NEW_ATAPacket==(PVOID)PTR_NULL){	/* no free ata packet. */
		/*Return_Execute_ATAPackets(bChannelNumber,CAM_ATAPacket,ATA_ABORT);*/
		return;		
	}	

	/* prepare ata packet */
	prepareATAPacket(NEW_ATAPacket, pATATask);

	/* free sg resource */
	CAMFreeATASG(pATATask->bCh2Chip, (PVOID)CAM_ATAPacket->PacketHead.PSG);

	CAMFreeCamPacket(pATATask->bCh2Chip, (PVOID)CAM_ATAPacket);
	
	/* put the atapacket to the haed of ata queue */
	Put2ATAQueue(NEW_ATAPacket,QUEUEHead);
	
	return;
}

PCAM_PACKET GetATA_PacketFromATAQueue(U8 bChannelNumber);

/* return all execute packets link after CAM_ATAPacket(include this). */  
void Return_Execute_ATAPackets(U8 bChannelNumber,PCAM_PACKET CAM_ATAPacket,U16 return_status) 
{
	PATA_TASK pATATask;
	PCAM_PACKET CAM_ATAPacket_next;

	
	/* return all packets at execute queue */
	while ((PTR_SIZE)CAM_ATAPacket != PTR_NULL){						

		pATATask=(PATA_TASK)CAM_ATAPacket->PacketHead.PTA;

		pATATask->ATAStatus = return_status;

		/* free sg resource */
		CAMFreeATASG(bChannelNumber, (PVOID)CAM_ATAPacket->PacketHead.PSG);
		
		CAM_ATAPacket_next = (PCAM_PACKET)CAM_ATAPacket->PacketHead.NPA;
		
		/* free packet resource. */
		CAMFreeCamPacket(bChannelNumber, (PVOID)CAM_ATAPacket);		

		/* next packet */
		CAM_ATAPacket = CAM_ATAPacket_next;

		pATATask->callback(pATATask);
				
	}
	return;
}

void Return_Waiting_ATAPackets(U8 bChannelNumber,U16 return_status) 
{
	PATA_TASK pATATask;
	PCAM_PACKET CAM_ATAPacket;

	/* return all packet at wait queue */
	while(1){
		CAM_ATAPacket = GetATA_PacketFromATAQueue(bChannelNumber);
		
		if ( CAM_ATAPacket == ZERO_NULL)	/* no packet */
			break;
			

		pATATask=(PATA_TASK)CAM_ATAPacket->PacketHead.PTA;

		pATATask->ATAStatus = return_status;

		
		/* free sg resource */
		CAMFreeATASG(bChannelNumber, (PVOID)CAM_ATAPacket->PacketHead.PSG);
		
		
		/* free packet resource. */
		CAMFreeCamPacket(bChannelNumber, (PVOID)CAM_ATAPacket);		

		pATATask->callback(pATATask);
				
	}
	return;
}

void Abort_All_Undone_ATAPackets(U8 bChannelNumber,PCAM_PACKET CAM_ATAPacket) 
{
	Return_Execute_ATAPackets(bChannelNumber,CAM_ATAPacket,ATA_ABORT);
	Return_Waiting_ATAPackets(bChannelNumber,ATA_ABORT);
	return;
}

/* back all execute packets to wait queue of this channel.*/
void Back_All_Undone_ATAPackets(U8 bChannelNumber,PCAM_PACKET CAM_ATAPacket) 
{
	PCAM_PACKET CAM_ATAPacket_tmp;
	PCAM_PACKET CAM_ATAPacket_tmp_pre=(PCAM_PACKET)PTR_NULL;
	PCAM_ATA_PACKET NEW_ATAPacket;
	

	if ( (PTR_SIZE)CAM_ATAPacket == PTR_NULL)
		return;
	
	CAM_ATAPacket_tmp_pre = CAM_ATAPacket;
	CAM_ATAPacket_tmp = (PCAM_PACKET)CAM_ATAPacket->PacketHead.NPA;
	
	/* form rear to head, put all undone execute packets into wait packet queue */
	while ( (PTR_SIZE)CAM_ATAPacket_tmp != PTR_NULL){						

		/* find the last packet */
		while ( (PTR_SIZE)CAM_ATAPacket_tmp->PacketHead.NPA != PTR_NULL){
			CAM_ATAPacket_tmp_pre = CAM_ATAPacket_tmp;
			CAM_ATAPacket_tmp = (PCAM_PACKET)CAM_ATAPacket_tmp->PacketHead.NPA;
		}

		CAM_ATAPacket_tmp_pre->PacketHead.NPA = (PCAM_PACKET)PTR_NULL;

		/* put to head of ata queue */

		/* get free packet address */
		NEW_ATAPacket= (PCAM_ATA_PACKET)CAMAllocATAPacket(bChannelNumber);
		if (NEW_ATAPacket==(PVOID)PTR_NULL){	/* no free ata packet. */
			Return_Execute_ATAPackets(bChannelNumber,CAM_ATAPacket_tmp,ATA_ABORT);
			return;		
		}	

		/* prepare ata packet */
		prepareATAPacket(NEW_ATAPacket, (PATA_TASK)CAM_ATAPacket_tmp->PacketHead.PTA);

		/* free resource */
		CAMFreeATASG(bChannelNumber, (PVOID)CAM_ATAPacket_tmp->PacketHead.PSG);
		CAMFreeCamPacket(bChannelNumber, (PVOID)CAM_ATAPacket_tmp);		
		Put2ATAQueue(NEW_ATAPacket ,QUEUEHead);
		
		CAM_ATAPacket_tmp_pre = CAM_ATAPacket;
		CAM_ATAPacket_tmp = (PCAM_PACKET)CAM_ATAPacket->PacketHead.NPA;
	}

	/* get free packet address */
	NEW_ATAPacket= (PCAM_ATA_PACKET)CAMAllocATAPacket(bChannelNumber);
	if (NEW_ATAPacket==(PVOID)PTR_NULL){	/* no free ata packet. */
		Return_Execute_ATAPackets(bChannelNumber,CAM_ATAPacket_tmp,ATA_ABORT);
		return;		
	}	

	/* prepare ata packet */
	prepareATAPacket(NEW_ATAPacket, (PATA_TASK)CAM_ATAPacket_tmp_pre->PacketHead.PTA);

	/* free resource */
	CAMFreeATASG(bChannelNumber, (PVOID)CAM_ATAPacket_tmp_pre->PacketHead.PSG);
	CAMFreeCamPacket(bChannelNumber, (PVOID)CAM_ATAPacket_tmp_pre);
	Put2ATAQueue(NEW_ATAPacket, QUEUEHead);
	return;
}

void stopATAModule(U8 bChannelNumber)
{
	U8	Adapter_ID;
	U8	SEQID;
	U8 	STOPFlag=FALSE;
	PSEQ_INFO	pCurSEQ_INFO=(PSEQ_INFO)PTR_NULL;
	U8	SEQCntCtrl;
	PATA_TASK pATATask;
	PCAM_PACKET ultmpATAPacketAddress;
	PCAM_PACKET ultmpATAPacketAddress_next;
	U8	receiveStatus;
	U32	i;
	U8	bData;
	U8	TAG_Flag=FALSE;
	U32 ldata;
	
	if ( gChnlConfig[bChannelNumber].bCh  == camCHNL_NOT_FOUND )
		return;

	if (bChannelNumber >= MAX_ATA_MODULE )
		return;
		
	/* disable receive task.*/
	receiveStatus=ATAModuleReceiveTask[bChannelNumber];
	ATAModuleReceiveTask[bChannelNumber]=FALSE;
				
	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;

	/* turn off the LED */
	if ((gBOXConfig[bChannelNumber].BOXType != camNOBOX) && (gBOXConfig[bChannelNumber].BOXType & SATAEnclosure)) {
		/* turn off LED */
		CAM_SBOX_HDAccess_LED(bChannelNumber, HDActiveOFF);
	}
		
	SEQID = ATAModuleSEQID[bChannelNumber];
	if (SEQID != U8NULL)
	{
		ATAModuleSEQID[bChannelNumber] = U8NULL;
		pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SEQID]);
		ultmpATAPacketAddress = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
		pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
		pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;

		if ((PTR_SIZE)ultmpATAPacketAddress != PTR_NULL) {
			if ( ((PPACKET_HEAD)ultmpATAPacketAddress)->Flag &(PACFLAG_NCQ|PACFLAG_QUEUED) )
				TAG_Flag=TRUE;
		}
		
		gChnlDrvControlConfig[bChannelNumber].TagCmdFlag &= ~(TAG_DATAIN|TAG_DATAOUT);
		gChnlDrvControlConfig[bChannelNumber].FreeTAG = 0xFFFFFFFF;
		gChnlDrvControlConfig[bChannelNumber].UsedTAG = 0x0;
				
	}
	else
		ultmpATAPacketAddress = (PCAM_PACKET)PTR_NULL;

	gChnlDrvControlConfig[bChannelNumber].TCQTransferData=FALSE;
	
	/* Ask ATA engine don't execute the next command packet, so the ata module will stop after done current packet. */
	WritePortUchar(ATABaseAddr[bChannelNumber], ATA_STATUS_REGISTER_OFFSET, 1 );		
	for (i = 0; i < 300000; i++ ) {	
		ReadPortUchar(ATABaseAddr[bChannelNumber], (ATA_STATUS_REGISTER_OFFSET+3), bData);
		if ( !(bData & 0x20) )
			break;	/* not in packet command cycle. */
		if (i==0) {
			if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA ) {
				ReadPortUlong( HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG, ldata );
				ldata >>= (bChannelNumber%MAX_ATA_CHNL);
				if (!( ldata & 0x100 )) /* not well connect */
					break;
			}
		}  
		camUStallExecution(1);	/* wait 3 secs */
	} ;	
	/* soft reset ATA Module. */
	ATASoftResetModule(bChannelNumber);		

	/* remove ata timeout timer */
	if (ATA_Timeout_TimerID[bChannelNumber]!=ZERO_NULL){
		camTimer(Adapter_ID,0,(void (*)(void))ATA_Timeout_TimerID[bChannelNumber]);	
		ATA_Timeout_TimerID[bChannelNumber]=ZERO_NULL;
	}
	
	/* check with executing packets in seqID */
	if((PTR_SIZE)ultmpATAPacketAddress != PTR_NULL) {
		
		/* read Sequence Counter Control Register # */
		ReadPortUchar(HostBaseAddr[Adapter_ID], SEQID*4, SEQCntCtrl);
		SEQCntCtrl &= 0x1F;		

		/* if NCQ commands, abort all commands */
		if (TAG_Flag==TRUE)
			pCurSEQ_INFO->PACCNT=SEQCntCtrl;
		/* ATAPI PIO SEQCntCtrl will be zero or if error, current packet maybe discount or not */
		else if (SEQCntCtrl < pCurSEQ_INFO->PACCNT)
			SEQCntCtrl++;
					
		while ((PTR_SIZE)ultmpATAPacketAddress != PTR_NULL){						

			pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;
			
			/* check is error packet */
			if (pCurSEQ_INFO->PACCNT == SEQCntCtrl)
			{ 	/* error packet */

				/* abort packets queue at the rear of this packet and and wait queue packets. */
				Abort_All_Undone_ATAPackets(bChannelNumber,(PCAM_PACKET)ultmpATAPacketAddress->PacketHead.NPA);
				ultmpATAPacketAddress->PacketHead.NPA=(PVOID)PTR_NULL;			

				pATATask->ATAStatus = ATA_ABORT;
				STOPFlag = TRUE;
			}
			else
			{
				pCurSEQ_INFO->PACCNT--;
				pATATask->ATAStatus= 0;		/* no error */
				if (pATATask->ATACmdFlag & ATA_STATUS_RETURN) /* need read it's ata registers statses when command success completed */
					ATAErrorStatus(pATATask);				
			}
								
			pATATask->ATACmdFlag &= ~INTERNAL_SENSE;

			/* free sg resource */
			CAMFreeATASG(bChannelNumber, (PVOID)ultmpATAPacketAddress->PacketHead.PSG);
			
			ultmpATAPacketAddress_next = (PCAM_PACKET)ultmpATAPacketAddress->PacketHead.NPA;
						
			/* free packet resource. */
			CAMFreeCamPacket(bChannelNumber, (PVOID)ultmpATAPacketAddress);		
			
			/* next packet */
			ultmpATAPacketAddress = ultmpATAPacketAddress_next;

			pATATask->callback(pATATask);
			
			if( STOPFlag == TRUE)
				break;
		}
		

		pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
		pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;
		ATAModuleBusy[bChannelNumber]=FALSE;
		if (TAG_Flag!=TRUE)
			CAMFreeSEQID(Adapter_ID,SEQID);
	}
	else{
		/* no executing packets,abort packets queue at wait queue packets. */
		Abort_All_Undone_ATAPackets(bChannelNumber,(PCAM_PACKET)PTR_NULL);
	}	
	
	ATAModuleReceiveTask[bChannelNumber]=receiveStatus;
	return;
}

void HandleInternalSwapbox_CMD(U8 bChannelNumber, U8 SEQID)
{
	PSEQ_INFO	pCurSEQ_INFO;
	PCAM_PACKET ultmpATAPacketAddress;
	PATA_TASK pATATask=ZERO_NULL;
	U8 i;
	U8	Adapter_ID;

	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;	
	
	pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SEQID]);

	/*If a packet is PIO_XFER or SWAPBOX_CMD, it will not queue with other packets when executing.*/
	ultmpATAPacketAddress = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;

	pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;

	pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
	pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;			
			
	pATATask->ATAStatus= 0;		/* no error */
								
	pATATask->ATACmdFlag &= ~INTERNAL_SENSE;
	
	/* free sg resource */
	CAMFreeATASG(bChannelNumber, (PVOID)ultmpATAPacketAddress->PacketHead.PSG);			
	/* free packet resource. */
	CAMFreeCamPacket(bChannelNumber, (PVOID)ultmpATAPacketAddress);		

	pATATask->callback(pATATask);

	ATAModuleBusy[bChannelNumber]=FALSE;
	ATAModuleSEQID[bChannelNumber] = U8NULL;
	CAMFreeSEQID(Adapter_ID,SEQID);

	/* must check module is need to push, By it maybe pushed when no SQEID. */
	for (i=(Adapter_ID*MAX_ATA_CHNL);i<((Adapter_ID+1)*MAX_ATA_CHNL);i++) {
		if ( (ATAModuleNeedPush[i]==TRUE) && (ATAModuleBusy[i]==FALSE) ){
			/* start ata module */
			StartATAOperation(i);
			break;
		}
	}
	
	return;
}

void IssueATAPIPktCmd(U8 bChannelNumber, U8 SEQID)
{
	U8	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;
	U8	bID;
	PTR_SIZE			BaseAddress = ATABaseAddr[bChannelNumber];
	PATA_TASK pATATask = (PATA_TASK)((PCAM_PACKET)gSEQ_INFO[Adapter_ID][SEQID].SEQID2PacketHead)->PacketHead.PTA;
	
	
	bID = pATATask->bID;

	/* disable INTA here, it will be re-enable when CAM use SEQ 0 for packets */
	WriteIndexUchar(BaseAddress, iNDEXIDECICR, 0x0);  /* that the drive INT pass to SEQ 0*/
	WritePortUchar( HostBaseAddr[Adapter_ID], 0, 0x20); /* but mask SEQ 0 INT */

	/* select drive */
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */
		WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, 0xE0);
	else		
		WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, (U8)(((bID & 0x01) << 4) | 0xA0));
 	WaitOnBusy(BaseAddress);

 	WriteIndexUchar(BaseAddress, iNDEXIDESectorCount, 0x00);
 	WriteIndexUchar(BaseAddress, iNDEXIDESectorNumber, 0x00);
 	
  	/* set feature register */   				
	if ( pATATask->ATACmdFlag & PIO_XFER ) {
 		WriteIndexUchar(BaseAddress, iNDEXIDEFeature, 0x00); 	/* data xfer by PIO ! */
		/* set real transfer byte count to byte counter register (cylinder low/high) */	
		if(pATATask->DataTransferLength < 0x10000){
			WriteIndexUchar(BaseAddress, iNDEXIDECylinderLow, (U8)pATATask->DataTransferLength);
 			WriteIndexUchar(BaseAddress, iNDEXIDECylinderHigh, (U8)(pATATask->DataTransferLength >> 8));
		}
		else{
			WriteIndexUchar(BaseAddress, iNDEXIDECylinderLow, (U8)0xFF);
 			WriteIndexUchar(BaseAddress, iNDEXIDECylinderHigh, (U8)0xFF);		
		}
	}
	else {
		WriteIndexUchar(BaseAddress, iNDEXIDEFeature, 0x01); 	/* set DMA bit */
		/* byte counter register (cylinder low/high) , set to 0 */
		WriteIndexUchar(BaseAddress, iNDEXIDECylinderLow, 0x00);
 		WriteIndexUchar(BaseAddress, iNDEXIDECylinderHigh, 0x00);
	}
	
	/* send ATAPI packet command 0xA0 */
	WriteIndexUchar(BaseAddress, iNDEXIDECommand, IDE_COMMAND_PACKET);

	/* Do we have to wait INT ? */	
	if ( gDrvConfig[bChannelNumber][bID].DevFlag & DEV_INT_DRQ )
		WaitForINT(BaseAddress);
	
	WaitForDrq(BaseAddress);
	
	/* now device wait CDB only */

	return;
}

#ifdef PIORW
void ReturnATA_PIOCmd(U8 bChannelNumber, U8 SEQID, U16 ATAStatus)
{
	PSEQ_INFO	pCurSEQ_INFO;
	PCAM_PACKET ultmpATAPacketAddress;
	PATA_TASK pATATask=ZERO_NULL;
	U8 i;
	U8	Adapter_ID;

	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;	
	
	pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SEQID]);

	/*If a packet is PIO_XFER or SWAPBOX_CMD, it will not queue with other packets when executing.*/
	ultmpATAPacketAddress = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;

	pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;

	pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
	pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;			
			
	pATATask->ATAStatus= ATAStatus;		/* no error */
								
	/* free sg resource */
	CAMFreeATASG(bChannelNumber, (PVOID)ultmpATAPacketAddress->PacketHead.PSG);			
	/* free packet resource. */
	CAMFreeCamPacket(bChannelNumber, (PVOID)ultmpATAPacketAddress);		

	pATATask->callback(pATATask);

	ATAModuleBusy[bChannelNumber]=FALSE;
	ATAModuleSEQID[bChannelNumber] = U8NULL;
	CAMFreeSEQID(Adapter_ID,SEQID);

	/* must check module is need to push, By it maybe pushed when no SQEID. */
	for (i=(Adapter_ID*MAX_ATA_CHNL);i<((Adapter_ID+1)*MAX_ATA_CHNL);i++) {
		if ( (ATAModuleNeedPush[i]==TRUE) && (ATAModuleBusy[i]==FALSE) ){
			/* start ata module */
			StartATAOperation(i);
			break;
		}
	}
	
	return;
}


U8 IssueATA_PIOCmd(U8 bChannelNumber, U8 SEQID)
{
	U8		Adapter_ID = bChannelNumber/MAX_ATA_CHNL;
	U8		bID;
	PTR_SIZE 	BaseAddress = ATABaseAddr[bChannelNumber];
	PATA_TASK 	pATATask = (PATA_TASK)((PCAM_PACKET)gSEQ_INFO[Adapter_ID][SEQID].SEQID2PacketHead)->PacketHead.PTA;
	U32 		startingSector;
	U8 		LargeLBA = checkLargeLBA(pATATask);
	U8 		Status;
	
	
	bID = pATATask->bID;

	/* disable INTA here, it will be re-enable when CAM use SEQ 0 for packets */
	WriteIndexUchar(BaseAddress, iNDEXIDECICR, SEQID);  /* that the drive INT pass to SEQ #*/
	WritePortUchar( HostBaseAddr[Adapter_ID], SEQID*4, 0x20); /* but mask SEQ # INT */

	/* select drive */
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */
		WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, 0xE0);
	else		
		WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, (U8)(((bID & 0x01) << 4) | 0xA0));
 	//WaitOnBusy(BaseAddress);

	if ( LargeLBA == TRUE )
	{
		/* previous content */
	 	WriteIndexUchar(BaseAddress, iNDEXIDESectorCount, (U8)(pATATask->Cmd.ATACmd.bCount >> 8));
		startingSector=pATATask->Cmd.ATACmd.LBAl;	
	 	WriteIndexUchar(BaseAddress, iNDEXIDESectorNumber, (U8)(startingSector >> 24));

		startingSector=pATATask->Cmd.ATACmd.LBAh;
	 	WriteIndexUchar(BaseAddress, iNDEXIDECylinderLow, (U8)(startingSector));
		WriteIndexUchar(BaseAddress, iNDEXIDECylinderHigh, (U8)(startingSector >> 8));
	
	}
 	/* most recently written */
 	

	startingSector=pATATask->Cmd.ATACmd.LBAl;
	
	WriteIndexUchar(BaseAddress, iNDEXIDEFeature, pATATask->Cmd.ATACmd.bFeature);
 	WriteIndexUchar(BaseAddress, iNDEXIDESectorCount, pATATask->Cmd.ATACmd.bCount);
 	WriteIndexUchar(BaseAddress, iNDEXIDESectorNumber, (U8)(startingSector));
 	WriteIndexUchar(BaseAddress, iNDEXIDECylinderLow, (U8)(startingSector >> 8));
	WriteIndexUchar(BaseAddress, iNDEXIDECylinderHigh, (U8)(startingSector >> 16));


	/* select drive */
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) {/* SATA */
		if ( LargeLBA != TRUE )
			WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, (((U8)(startingSector >> 24) & 0x0F) | 0xE0));
		else
			WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, 0xE0);
	}
	else {
		if ( LargeLBA != TRUE )
			WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, (U8) (((U8)(startingSector >> 24) & 0x0F) | 0xE0 | (bID << 4)));
		else
			WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, (U8)(((bID & 0x01) << 4) | 0xE0));
	}
	
	/* sendcommand */
	WriteIndexUchar(BaseAddress, iNDEXIDECommand, pATATask->Cmd.ATACmd.bCmd);

retry_ATA_PIO_WRITE:

	Status=WaitForDrq(BaseAddress);
	ReadIndexUchar(BaseAddress, iNDEXIDEStatus);

	if ( (!(Status & IDE_STATUS_BUSY)) && (!(Status & IDE_STATUS_ERROR)) && (Status & IDE_STATUS_DRQ) ) {

		U32	DataTransferLengthTotalTemp=0x100;
		U8 	bdata;
		
		ReadIndexUchar(BaseAddress, iNDEXIDEStatus); /* have data to be trasfer */ /*clear possible device INT */
		if ( pATATask->ATACmdFlag & camDATA_IN ) 
			Status = ReadBuffer(pATATask->bCh2Chip,(PU16)pATATask->DataBuffer, DataTransferLengthTotalTemp);
		else
			Status = WriteBuffer(pATATask->bCh2Chip,(PU16)pATATask->DataBuffer, DataTransferLengthTotalTemp );
		
		pATATask->DataBuffer = (PVOID)((PU16)pATATask->DataBuffer+DataTransferLengthTotalTemp);
		pATATask->DataTransferLength = pATATask->DataTransferLength - (DataTransferLengthTotalTemp*2);
	
		bdata = ReadIndexUchar(BaseAddress, iNDEXIDEAlternateStatus); /* have data to be trasfer */ /*clear possible device INT */

		if (Status != camFAIL) {
			if (!(bdata & IDE_STATUS_ERROR)) {
				if (pATATask->DataTransferLength) { /* next DRQ */
					goto retry_ATA_PIO_WRITE;
				}
			}
		}
	
		if ( (pATATask->DataTransferLength) && (Status != camFAIL) ) {
			return FALSE;
	
		}
		else if ( (!(pATATask->DataTransferLength)) && (bdata & IDE_STATUS_BUSY) )
			if ( WaitOnReady(bChannelNumber) == camSUCCESS ) 
				return TRUE;
			else
				return FALSE;
		else {
			// clear possible INT
			ReadIndexUchar(BaseAddress, iNDEXIDEStatus);
			return TRUE;
		}
	}


	return FALSE;
}
#endif

void ATA_Timeout_CallBack(U8 bChannelNumber, U8 status)
{
	PSEQ_INFO	pCurSEQ_INFO;
	U8	SEQCntCtrl;
	PCAM_PACKET		ultmpATAPacketAddress;
	PCAM_PACKET		ultmpATAPacketAddress_next;

	PATA_TASK pATATask=ZERO_NULL;
	U8 	SEQID;
	U8 	i;
	U8	Adapter_ID;
	U8	errorFlag=FALSE;

	U8	TAG_Flag = FALSE;
	U32 	tmpData;


	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;
	/*SEQID = ATAModuleSEQID[bChannelNumber];*/
	SEQID = gChnlResetTable[bChannelNumber].SEQID;
	SEQCntCtrl = gChnlResetTable[bChannelNumber].SEQCntCtrl;
	
	/* error timeout. If no SEQID assign to channel packets, timeout will not happen, so if timeout function be called at this case, it is a error timeout.*/
	/*if (SEQID == U8NULL)
		return;
	*/
		
	ATAModuleSEQID[bChannelNumber] = U8NULL;
		
	pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SEQID]);

	ultmpATAPacketAddress = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;

	if ((PTR_SIZE)ultmpATAPacketAddress != PTR_NULL) {
		if ( ((PPACKET_HEAD)ultmpATAPacketAddress)->Flag &(PACFLAG_NCQ|PACFLAG_QUEUED) )
			TAG_Flag=TRUE;
	}

	ATAModuleReceiveTask[bChannelNumber] = gChnlResetTable[bChannelNumber].receiveStatus;
	gChnlResetTable[bChannelNumber].OwnerType = 0;
				
	while ((PTR_SIZE)ultmpATAPacketAddress != PTR_NULL){						

		pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;

		/* check is error packet */
		if (pCurSEQ_INFO->PACCNT == SEQCntCtrl)
		{ 	/* error packet */

			errorFlag = TRUE;

			pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
			pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;	
			
			if (TAG_Flag != TRUE) {		
				/* back undo execute packets to wait queue. */
				Back_All_Undone_ATAPackets(bChannelNumber,(PCAM_PACKET)ultmpATAPacketAddress->PacketHead.NPA);
				ultmpATAPacketAddress->PacketHead.NPA=(PVOID)PTR_NULL;			
			}
			/* get ATA Global control & Status Register */
			ReadPortUlong(ATABaseAddr[bChannelNumber], ATA_STATUS_REGISTER_OFFSET, tmpData);
			pATATask->ATAStatus = (U16)(tmpData >> 8);
			
			pATATask->ATAStatus &= ATAERR_HWERROR;
			pATATask->ATAStatus &= ~ATAERR_DRV;	
			pATATask->ATAStatus |= (ATA_ERROR|ATA_TIMEOUT);	
			
			if (TAG_Flag == TRUE) {
				pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
				pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;
				ATAModuleBusy[bChannelNumber]=FALSE;

				Return_Execute_ATAPackets(bChannelNumber,(PCAM_PACKET)ultmpATAPacketAddress,pATATask->ATAStatus);
				ultmpATAPacketAddress=(PCAM_PACKET)PTR_NULL;				
				break;
			}																

			if (!(pATATask->ATACmdFlag & ATA_DEVICE)) {	/* ATAPI device */
				/* ATAPI error and not internel sense */
				if (!(pATATask->ATACmdFlag & INTERNAL_SENSE)){
					/* must have sense buffer so can do internel sense command */
					if ((pATATask->SenseInfoLength > 1)) {
						/* If ATAPI packet command error, than send a internel "REQUEST SENSE Command" 
						     to get the sense datas. 	*/
						pATATask->ATACmdFlag = INTERNAL_SENSE;
					}
				}
				else {
					pATATask->ATACmdFlag &= ~INTERNAL_SENSE;
					ReadIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDEStatus); /*clear possible device INT */
				}												
			}
		}
		else
		{
			pCurSEQ_INFO->PACCNT--;
			pATATask->ATAStatus= 0;		/* no error */
			if (pATATask->ATACmdFlag & ATA_STATUS_RETURN) /* need read it's ata registers statses when command success completed */
				ATAErrorStatus(pATATask);
		}
								
		ultmpATAPacketAddress_next = (PCAM_PACKET)ultmpATAPacketAddress->PacketHead.NPA;	

		if (pATATask->ATACmdFlag & INTERNAL_SENSE) {
			SubmitInternelSense(pATATask,ultmpATAPacketAddress);
			break;
		}

		/* free sg resource */
		CAMFreeATASG(bChannelNumber, (PVOID)ultmpATAPacketAddress->PacketHead.PSG);
						
		/* free packet resource. */
		CAMFreeCamPacket(bChannelNumber, (PVOID)ultmpATAPacketAddress);		
								
		if ( !(pATATask->ATACmdFlag & INTERNAL_SENSE) )
			pATATask->callback(pATATask);

		if (errorFlag == TRUE)
			break;
			
		/* next packet */
		ultmpATAPacketAddress = ultmpATAPacketAddress_next;
		
	}

	if (TAG_Flag != TRUE){
		pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
		pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;
		ATAModuleBusy[bChannelNumber]=FALSE;
	}

	if (TAG_Flag!=TRUE)
		CAMFreeSEQID(Adapter_ID,SEQID);

	/* Is caused by disk off ? */
	if (pATATask)
	{
		if( !pATATask->bID)
		{
			if ( (gDrvConfig[bChannelNumber][0].pEnclosure) && (gChnlConfig[bChannelNumber].ChnlType == CHNL_ATA)) {
				Check_Enclosure_PowerState(bChannelNumber);
				if ( gBOXStatus[bChannelNumber].PowerState == PowerOFF ) {
					if ( gDrvConfig[bChannelNumber][0].DevFlag & DEV_EXIST ){
						gDrvConfig[bChannelNumber][0].DevFlag = DEV_NOTFOUND;
						cam_SetStatusByPlug_unplug(bChannelNumber, 0, NO_DRIVE);
					}
				}					
			}	
		}		
	}

	/*ATAModuleReceiveTask[bChannelNumber] = gChnlResetTable[bChannelNumber].receiveStatus;*/
	/*gChnlResetTable[bChannelNumber].OwnerType = 0;*/
	
	/* must check module is need to push, By it maybe pushed when no SQEID. */
	for (i=(Adapter_ID*MAX_ATA_CHNL);i<((Adapter_ID+1)*MAX_ATA_CHNL);i++) {
		if ( (ATAModuleNeedPush[i]==TRUE) && (ATAModuleBusy[i]==FALSE) ){
			/* start ata module */
			StartATAOperation(i);
			break;
		}
	}
	return;
}



void ATA_Timeout(U8 bChannelNumber)
{
	PSEQ_INFO	pCurSEQ_INFO;
	U8	SEQCntCtrl;
	PCAM_PACKET		ultmpATAPacketAddress;

	U8 	SEQID;
	U8 	i;
	U8	Adapter_ID;
	U8	TAG_Flag = FALSE;
	U32 	tmpData;
	U8	CRC_err=FALSE;
	
	ATA_Timeout_TimerID[bChannelNumber]=ZERO_NULL;
	
	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;
	SEQID = ATAModuleSEQID[bChannelNumber];
	
 	/* error timeout. If no SEQID assign to channel packets, timeout will not happen, so if timeout function be called at this case, it is a error timeout.*/
	if (SEQID == U8NULL)
		return;

	if(ATAModuleReceiveTask[bChannelNumber] != TRUE)
		return;
	
	for( i=0;i<MAX_ID_PER_CHNL;i++ ){
		ATASortHead[bChannelNumber][i] = (PCAM_ATA_PACKET)PTR_NULL;
		ATASortLBAl[bChannelNumber][i] = ZERO_NULL;
		ATASortLBAh[bChannelNumber][i] = ZERO_NULL;			
	}
	
	/* disable receive task.*/
	gChnlResetTable[bChannelNumber].receiveStatus=ATAModuleReceiveTask[bChannelNumber];
	ATAModuleReceiveTask[bChannelNumber]=FALSE;


	/* turn off the LED */
	if ((gBOXConfig[bChannelNumber].BOXType != camNOBOX) && (gBOXConfig[bChannelNumber].BOXType & SATAEnclosure)) {
		/* turn off LED */
		CAM_SBOX_HDAccess_LED(bChannelNumber, HDActiveOFF);
	}
			
	ATAModuleSEQID[bChannelNumber] = U8NULL;
		
	pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SEQID]);

	ultmpATAPacketAddress = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;

	if ((PTR_SIZE)ultmpATAPacketAddress != PTR_NULL) {
		if ( ((PPACKET_HEAD)ultmpATAPacketAddress)->Flag &(PACFLAG_NCQ|PACFLAG_QUEUED) )
			TAG_Flag=TRUE;
	}
		
	gChnlDrvControlConfig[bChannelNumber].TagCmdFlag &= ~(TAG_DATAIN|TAG_DATAOUT);
	gChnlDrvControlConfig[bChannelNumber].FreeTAG = 0xFFFFFFFF;
	gChnlDrvControlConfig[bChannelNumber].UsedTAG = 0x0;
	gChnlDrvControlConfig[bChannelNumber].TCQTransferData=FALSE;
	
	/* read Sequence Counter Control Register # */
	ReadPortUchar(HostBaseAddr[Adapter_ID], SEQID*4, SEQCntCtrl);
	SEQCntCtrl &= 0x1F;
	if(gAdapterConfig[Adapter_ID].TapeOutVersion < REVISION_2){/* ASIC tapout prior revision 2 */
		if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) {/* SATA */
			/* check is CRC_err */
			#ifdef	_MMIO_
				ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATAError, tmpData );
			#else
				ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDESATAError, tmpData );
			#endif
			if (tmpData & CRC_ERR) {	/* CRC error */
				CRC_err=TRUE;
			}
		}
	}
	
	/* if NCQ commands, all commands error */
	if (TAG_Flag==TRUE)
		pCurSEQ_INFO->PACCNT=SEQCntCtrl;	
	/* if CRC_err */
	else if (CRC_err==TRUE)
		pCurSEQ_INFO->PACCNT=SEQCntCtrl;					
	/* ATAPI PIO SEQCntCtrl will be zero or if error, current packet maybe discount or not */
	else if (SEQCntCtrl < pCurSEQ_INFO->PACCNT)
		SEQCntCtrl++;

	gChnlResetTable[bChannelNumber].SEQCntCtrl = SEQCntCtrl;	/* keep SEQCntCtrl for Timeout_Callback( ) */
	gChnlResetTable[bChannelNumber].notify = (void (*)(U8, U8)) PTR_NULL;
	gChnlResetTable[bChannelNumber].OwnerType = WAIT_TIMEOUT;
	gChnlResetTable[bChannelNumber].run_count = 0;
	gChnlResetTable[bChannelNumber].SEQID = SEQID;
	
	if((CAM_Reset_ReinitATA(bChannelNumber)) != camSUCCESS){
		ATAModuleReceiveTask[bChannelNumber] = gChnlResetTable[bChannelNumber].receiveStatus;
		gChnlResetTable[bChannelNumber].OwnerType = 0;
	}
	return;
}


#if (MAX_ADAPTER_NUM>0)
void ATA_Timeout0(void)
{
	ATA_Timeout(0x00);
	return;
}

void ATA_Timeout1(void)
{
	ATA_Timeout(0x01);
	return;
}

void ATA_Timeout2(void)
{
	ATA_Timeout(0x02);
	return;
}

void ATA_Timeout3(void)
{
	ATA_Timeout(0x03);
	return;
}
#endif
#if (MAX_ADAPTER_NUM>1)
void ATA_Timeout4(void)
{
	ATA_Timeout(0x04);
	return;
}

void ATA_Timeout5(void)
{
	ATA_Timeout(0x05);
	return;
}

void ATA_Timeout6(void)
{
	ATA_Timeout(0x06);
	return;
}

void ATA_Timeout7(void)
{
	ATA_Timeout(0x07);
	return;
}
#endif
#if (MAX_ADAPTER_NUM>2)
void ATA_Timeout8(void)
{
	ATA_Timeout(0x08);
	return;
}

void ATA_Timeout9(void)
{
	ATA_Timeout(0x09);
	return;
}

void ATA_TimeoutA(void)
{
	ATA_Timeout(0x0A);
	return;
}

void ATA_TimeoutB(void)
{
	ATA_Timeout(0x0B);
	return;
}
#endif
#if (MAX_ADAPTER_NUM>3)
void ATA_TimeoutC(void)
{
	ATA_Timeout(0x0C);
	return;
}

void ATA_TimeoutD(void)
{
	ATA_Timeout(0x0D);
	return;
}

void ATA_TimeoutE(void)
{
	ATA_Timeout(0x0E);
	return;
}

void ATA_TimeoutF(void)
{
	ATA_Timeout(0x0F);
	return;
}
#endif

void (*ATA_Timeout_Function[]) (void) = {

#if (MAX_ADAPTER_NUM>0)
  ATA_Timeout0,
  ATA_Timeout1,
  ATA_Timeout2,
  ATA_Timeout3,
#endif
#if (MAX_ADAPTER_NUM>1)
  ATA_Timeout4,
  ATA_Timeout5,
  ATA_Timeout6,
  ATA_Timeout7,
#endif
#if (MAX_ADAPTER_NUM>2)
  ATA_Timeout8,
  ATA_Timeout9,
  ATA_TimeoutA,
  ATA_TimeoutB,
#endif  
#if (MAX_ADAPTER_NUM>3)
  ATA_TimeoutC,
  ATA_TimeoutD,
  ATA_TimeoutE,
  ATA_TimeoutF,  
#endif  
};

void ATA_Check_Port_Multiplier_Ports(PATA_TASK pATATask)
{
	U8 bChannelNumber;
	
	bChannelNumber = pATATask->bCh2Chip;
	/* check all port of Port_Multiplier been plug/unplug */
	
	/* restart check timer */	
	ATA_Port_Multiplier_TimerID[bChannelNumber] = camTimer(((U8)(bChannelNumber/MAX_ATA_CHNL)),CAMTimeoutTimes[bChannelNumber/MAX_ATA_CHNL].Port_Multiplier_Check, ATA_Check_Port_Multiplier_Function[bChannelNumber]);	
	
	return;
}


void ATA_Check_Port_Multiplier(U8 bChannelNumber)
{
	PATA_TASK pATATask = &InternalATATask[bChannelNumber];

	ATA_Port_Multiplier_TimerID[bChannelNumber]=ZERO_NULL;	
	
	pATATask->bCh = CAM_ATA_Chnl_OutMAP(bChannelNumber);
	pATATask->bID = 0;
	pATATask->ATACmdFlag = SWAPBOX_CMD;
	pATATask->callback = ATA_Check_Port_Multiplier_Ports;	
	
	CAM_SubmitATA (pATATask);
	return;
}

#if (MAX_ADAPTER_NUM>0)
void ATA_Check_Port_Multiplier0(void)
{
	ATA_Check_Port_Multiplier(0x00);
	return;
}

void ATA_Check_Port_Multiplier1(void)
{
	ATA_Check_Port_Multiplier(0x01);
	return;
}

void ATA_Check_Port_Multiplier2(void)
{
	ATA_Check_Port_Multiplier(0x02);
	return;
}

void ATA_Check_Port_Multiplier3(void)
{
	ATA_Check_Port_Multiplier(0x03);
	return;
}
#endif
#if (MAX_ADAPTER_NUM>1)
void ATA_Check_Port_Multiplier4(void)
{
	ATA_Check_Port_Multiplier(0x04);
	return;
}

void ATA_Check_Port_Multiplier5(void)
{
	ATA_Check_Port_Multiplier(0x05);
	return;
}

void ATA_Check_Port_Multiplier6(void)
{
	ATA_Check_Port_Multiplier(0x06);
	return;
}

void ATA_Check_Port_Multiplier7(void)
{
	ATA_Check_Port_Multiplier(0x07);
	return;
}
#endif
#if (MAX_ADAPTER_NUM>2)
void ATA_Check_Port_Multiplier8(void)
{
	ATA_Check_Port_Multiplier(0x08);
	return;
}

void ATA_Check_Port_Multiplier9(void)
{
	ATA_Check_Port_Multiplier(0x09);
	return;
}

void ATA_Check_Port_MultiplierA(void)
{
	ATA_Check_Port_Multiplier(0x0A);
	return;
}

void ATA_Check_Port_MultiplierB(void)
{
	ATA_Check_Port_Multiplier(0x0B);
	return;
}
#endif
#if (MAX_ADAPTER_NUM>3)
void ATA_Check_Port_MultiplierC(void)
{
	ATA_Check_Port_Multiplier(0x0C);
	return;
}

void ATA_Check_Port_MultiplierD(void)
{
	ATA_Check_Port_Multiplier(0x0D);
	return;
}

void ATA_Check_Port_MultiplierE(void)
{
	ATA_Check_Port_Multiplier(0x0E);
	return;
}

void ATA_Check_Port_MultiplierF(void)
{
	ATA_Check_Port_Multiplier(0x0F);
	return;
}
#endif

void (*ATA_Check_Port_Multiplier_Function[]) (void) = {

#if (MAX_ADAPTER_NUM>0)
  ATA_Check_Port_Multiplier0,
  ATA_Check_Port_Multiplier1,
  ATA_Check_Port_Multiplier2,
  ATA_Check_Port_Multiplier3,
#endif
#if (MAX_ADAPTER_NUM>1)
  ATA_Check_Port_Multiplier4,
  ATA_Check_Port_Multiplier5,
  ATA_Check_Port_Multiplier6,
  ATA_Check_Port_Multiplier7,
#endif
#if (MAX_ADAPTER_NUM>2)
  ATA_Check_Port_Multiplier8,
  ATA_Check_Port_Multiplier9,
  ATA_Check_Port_MultiplierA,
  ATA_Check_Port_MultiplierB,
#endif  
#if (MAX_ADAPTER_NUM>3)
  ATA_Check_Port_MultiplierC,
  ATA_Check_Port_MultiplierD,
  ATA_Check_Port_MultiplierE,
  ATA_Check_Port_MultiplierF,  
#endif  
};



/* for Polling SATA device ready */
void ATA_Polling_Channel(U8 bChannelNumber)
{
	if(ATA_Polling_Channel_TimerID[bChannelNumber] != ZERO_NULL)
		ATA_Polling_Channel_TimerID[bChannelNumber] = ZERO_NULL;

	if (gChnlConfig[bChannelNumber].ChnlType != CHNL_SATA ){/* CHNL_PATA */
		if(gChnlResetTable[bChannelNumber].OwnerType == WAIT_INIT_CH)
			CAM_InitATAChnl_CallBack(bChannelNumber, camSUCCESS);
		else
			CAM_Reset_ReinitATA_CallBack(bChannelNumber, camSUCCESS);
		return;
	}
	else	{/* SATA */
		U8 	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;
		U8	count;
		U8	status;
		U8	NeedReset=FALSE;
		U32 TimerDelay = 18*3;	/* 3 sec */

		
		count = gChnlResetTable[bChannelNumber].run_count;
		if(count >= 60)
			goto FailReturn;

		status = WaitDelayOnReady(bChannelNumber, 2);

		
		/* status == camFAIL */
		if(status != camSUCCESS){
			if(status == camWARNING){
				NeedReset = TRUE;
			}			
			else if(count <= 10){
				TimerDelay = 18/2;	/* 0.5 sec */
			}	
			else if((count > 10) && (count <= 20)){
				if((count % 2) == 0){
					NeedReset = TRUE;
				}
			}	
			else if((count > 20) && (count < 60)){
				if((count % 10) == 0){
					NeedReset = TRUE;
				}
			}


			count++;
			gChnlResetTable[bChannelNumber].run_count = count;

			if(NeedReset == TRUE){
				if ( CAM_HardResetSATAChannel(bChannelNumber) == camDEVICE_NOT_FOUND )
					goto FailReturn;
			}

			/*gChnlResetTable[bChannelNumber].notify(CAM_ATA_Chnl_OutMAP(bChannelNumber), camWARNING);*/
			ATA_Polling_Channel_TimerID[bChannelNumber] = camTimer(Adapter_ID, TimerDelay, ATA_Polling_Channel_Function[bChannelNumber]);	
			return;	/* no drive connected */
		}
		else{
			gChnlResetTable[bChannelNumber].run_count += 10;
			if(gChnlResetTable[bChannelNumber].OwnerType == WAIT_INIT_CH)
				CAM_InitATAChnl_CallBack(bChannelNumber, camSUCCESS);
			else
				CAM_Reset_ReinitATA_CallBack(bChannelNumber, camSUCCESS);
			return;		
		}	/* end of status == camSUCCESS */
	}
	
FailReturn:
	
	if(gChnlResetTable[bChannelNumber].OwnerType == WAIT_INIT_CH)
		CAM_InitATAChnl_CallBack(bChannelNumber, camDEVICE_NOT_FOUND);
	else
		CAM_Reset_ReinitATA_CallBack(bChannelNumber, camDEVICE_NOT_FOUND);
	return;
}


#if (MAX_ADAPTER_NUM>0)
void ATA_Polling_Channel0(void)
{
	ATA_Polling_Channel(0x00);
	return;
}

void ATA_Polling_Channel1(void)
{
	ATA_Polling_Channel(0x01);
	return;
}

void ATA_Polling_Channel2(void)
{
	ATA_Polling_Channel(0x02);
	return;
}

void ATA_Polling_Channel3(void)
{
	ATA_Polling_Channel(0x03);
	return;
}
#endif
#if (MAX_ADAPTER_NUM>1)
void ATA_Polling_Channel4(void)
{
	ATA_Polling_Channel(0x04);
	return;
}

void ATA_Polling_Channel5(void)
{
	ATA_Polling_Channel(0x05);
	return;
}

void ATA_Polling_Channel6(void)
{
	ATA_Polling_Channel(0x06);
	return;
}

void ATA_Polling_Channel7(void)
{
	ATA_Polling_Channel(0x07);
	return;
}
#endif
#if (MAX_ADAPTER_NUM>2)
void ATA_Polling_Channel8(void)
{
	ATA_Polling_Channel(0x08);
	return;
}

void ATA_Polling_Channel9(void)
{
	ATA_Polling_Channel(0x09);
	return;
}

void ATA_Polling_ChannelA(void)
{
	ATA_Polling_Channel(0x0A);
	return;
}

void ATA_Polling_ChannelB(void)
{
	ATA_Polling_Channel(0x0B);
	return;
}
#endif
#if (MAX_ADAPTER_NUM>3)
void ATA_Polling_ChannelC(void)
{
	ATA_Polling_Channel(0x0C);
	return;
}

void ATA_Polling_ChannelD(void)
{
	ATA_Polling_Channel(0x0D);
	return;
}

void ATA_Polling_ChannelE(void)
{
	ATA_Polling_Channel(0x0E);
	return;
}

void ATA_Polling_ChannelF(void)
{
	ATA_Polling_Channel(0x0F);
	return;
}
#endif

void (*ATA_Polling_Channel_Function[]) (void) = {

#if (MAX_ADAPTER_NUM>0)
  ATA_Polling_Channel0,
  ATA_Polling_Channel1,
  ATA_Polling_Channel2,
  ATA_Polling_Channel3,
#endif
#if (MAX_ADAPTER_NUM>1)
  ATA_Polling_Channel4,
  ATA_Polling_Channel5,
  ATA_Polling_Channel6,
  ATA_Polling_Channel7,
#endif
#if (MAX_ADAPTER_NUM>2)
  ATA_Polling_Channel8,
  ATA_Polling_Channel9,
  ATA_Polling_ChannelA,
  ATA_Polling_ChannelB,
#endif  
#if (MAX_ADAPTER_NUM>3)
  ATA_Polling_ChannelC,
  ATA_Polling_ChannelD,
  ATA_Polling_ChannelE,
  ATA_Polling_ChannelF,  
#endif  
};


/* if command of ATA device without use of addressing the data on the device, 
	it means the head of device will not move, 
	so this type of command will needn't sort */
	
U8 Check_Command_Sort(PCAM_ATA_PACKET CAM_ATAPacket)
{
	PATA_TASK pATATask;

	pATATask = (PATA_TASK)CAM_ATAPacket->PacketHead.PTA;	
	
	if (!(pATATask->ATACmdFlag & ATA_DEVICE)) /* ATAPI device */
		return FALSE;
		
	/* check is a data transfer */
	if (pATATask->ATACmdFlag & camDATA_XFER){ /* with data */
		switch(pATATask->Cmd.ATACmd.bCmd2Chip){
			/* default data commands need sort, but these case will not */
			case IDE_COMMAND_DOWNLOAD_MICROCODE:
			case IDE_COMMAND_IDENTIFY_DEVICE:
			case IDE_COMMAND_IDENTIFY_PACKET_DEVICE:
			case IDE_COMMAND_READ_BUFFER:
			case IDE_COMMAND_READ_LOG_EXT:
			case IDE_COMMAND_SECURITY_DISABLE_PASSWORD:
			case IDE_COMMAND_SECURITY_ERASE_UNIT:
			case IDE_COMMAND_SECURITY_SET_PASSWORD:
			case IDE_COMMAND_SECURITY_UNLOCK:
			case IDE_COMMAND_WRITE_BUFFER:
			case IDE_COMMAND_WRITE_LOG_EXT:
				return(FALSE);
				break;
			
			/* default data commands need sort */
			default:
				return(TRUE);
				break;			
		}	
	}
	else { /* non-data */
		switch(pATATask->Cmd.ATACmd.bCmd2Chip){
			/* default non data commands needn't sort, BUT these case will */
			case IDE_COMMAND_CFA_ERASE_SECTORS:
			case IDE_COMMAND_CONFIGURE_STREAM:
			case IDE_COMMAND_READ_VERIFY_SECTORS:
			case IDE_COMMAND_READ_VERIFY_SECTORS_EXT:
			case IDE_COMMAND_SEEK:
				return(TRUE);
				break;
				
			/* default non data commands needn't sort */
			default:
				return(FALSE);
				break;				
				
		}	
	}
	
}

/* return value
	1. TRUE : first > second
	2. FALSE : first < second
	3. camWARNING : first = second
*/
U8 cpmpare_64bit_LBA(U32 firstLBAh,U32 firstLBAl,U32 secondLBAh,U32 secondLBAl)
{
	if (firstLBAh > secondLBAh)
		return TRUE;
	else if (firstLBAh < secondLBAh)
		return FALSE;
	else {
		if (firstLBAl > secondLBAl)
			return TRUE;
		else if (firstLBAl < secondLBAl)
			return FALSE;	
		else
			return camWARNING;
	}
}


void SortDiskQueue(PCAM_ATA_PACKET CAM_ATAPacket)
{
	U8 direction=0;
	PCAM_ATA_PACKET CAM_ATAPacket_prev;
	PCAM_ATA_PACKET CAM_ATAPacket_tmp;
	PATA_TASK pATATask;
	U8 bChannelNumber;
	U8 bID;
	U8 status;
	U32 PrevLBAh,PrevLBAl;
	U32 PacketLBAh,PacketLBAl;
	U32 NextLBAh,NextLBAl;

CAM_ATAPacket->PacketHead.CountStamp = 0;

	pATATask = (PATA_TASK)CAM_ATAPacket->PacketHead.PTA;	
	bChannelNumber=pATATask->bCh2Chip;
	bID=pATATask->bID;

	PacketLBAh = pATATask->Cmd.ATACmd.LBAh;
	PacketLBAl = pATATask->Cmd.ATACmd.LBAl;
	
	CAM_ATAPacket_prev = (PCAM_ATA_PACKET)PTR_NULL;
	
	CAM_ATAPacket_tmp = ATASortHead[bChannelNumber][bID];

	NextLBAh = ((PATA_TASK)CAM_ATAPacket_tmp->PacketHead.PTA)->Cmd.ATACmd.LBAh;
	NextLBAl = ((PATA_TASK)CAM_ATAPacket_tmp->PacketHead.PTA)->Cmd.ATACmd.LBAl;

	PrevLBAh=ATASortLBAh[bChannelNumber][bID];
	PrevLBAl=ATASortLBAl[bChannelNumber][bID];
		
	if ( CAM_ATAPacket_tmp != (PCAM_ATA_PACKET)PTR_NULL ) {
		status = cpmpare_64bit_LBA(PrevLBAh,PrevLBAl,NextLBAh,NextLBAl);
    	
	    	if (status == TRUE)
	    		direction=0;
	    	else
	    		direction=1;
	}

	
	while ((PTR_SIZE)CAM_ATAPacket_tmp != PTR_NULL) {
		
	if(!(CAM_ATAPacket_tmp->PacketHead.Flag & PACFLAG_SORT_PASS)){
		/* sort tasks of the same bID */
		if (((PATA_TASK)CAM_ATAPacket_tmp->PacketHead.PTA)->bID == bID) {			
			NextLBAh = ((PATA_TASK)CAM_ATAPacket_tmp->PacketHead.PTA)->Cmd.ATACmd.LBAh;
			NextLBAl = ((PATA_TASK)CAM_ATAPacket_tmp->PacketHead.PTA)->Cmd.ATACmd.LBAl;	
					
			if(direction) {
				status=cpmpare_64bit_LBA(PrevLBAh,PrevLBAl,NextLBAh,NextLBAl);
				
				if (status != TRUE) {
					status=cpmpare_64bit_LBA(PrevLBAh,PrevLBAl,PacketLBAh,PacketLBAl);
					if (status != TRUE) {
						status=cpmpare_64bit_LBA(PacketLBAh,PacketLBAl,NextLBAh,NextLBAl);
						if (status == FALSE) 
							break;
					}
				}
				else {
					status=cpmpare_64bit_LBA(PrevLBAh,PrevLBAl,PacketLBAh,PacketLBAl);
					if (status != TRUE) 
						break;
					else 
						direction=0;

					status=cpmpare_64bit_LBA(PacketLBAh,PacketLBAl,NextLBAh,NextLBAl);
					if (status == TRUE)
						break;
				}
			}
			else {
				status=cpmpare_64bit_LBA(PrevLBAh,PrevLBAl,NextLBAh,NextLBAl);
				if (status==TRUE) {
					status=cpmpare_64bit_LBA(PrevLBAh,PrevLBAl,PacketLBAh,PacketLBAl);
					if (status != FALSE) {
						status=cpmpare_64bit_LBA(PacketLBAh,PacketLBAl,NextLBAh,NextLBAl);
						if (status == TRUE)
							break;
					}
				}
				else {
					status=cpmpare_64bit_LBA(PrevLBAh,PrevLBAl,PacketLBAh,PacketLBAl);
					if (status ==TRUE) 
						break;
					else {
						direction=1;
					}
					status=cpmpare_64bit_LBA(PacketLBAh,PacketLBAl,NextLBAh,NextLBAl);
					if (status !=TRUE)
						break;
				}
			}
			
			PrevLBAh=NextLBAh;
			PrevLBAl=NextLBAl;
				
		}
	}
		CAM_ATAPacket_prev = CAM_ATAPacket_tmp;
		CAM_ATAPacket_tmp = (PCAM_ATA_PACKET)CAM_ATAPacket_tmp->PacketHead.NPA;
	}
	
	if (CAM_ATAPacket_prev == (PCAM_ATA_PACKET)PTR_NULL) {
		
		ATASortHead[bChannelNumber][bID] = CAM_ATAPacket;
		
		if (ATAQueueHead[bChannelNumber]==CAM_ATAPacket_tmp) {
			ATAQueueHead[bChannelNumber] = CAM_ATAPacket;
			CAM_ATAPacket->PacketHead.NPA = CAM_ATAPacket_tmp;
		}
		else {
			CAM_ATAPacket_prev=ATAQueueHead[bChannelNumber];
			while((PTR_SIZE)CAM_ATAPacket_prev != PTR_NULL) {
				if (CAM_ATAPacket_prev->PacketHead.NPA==CAM_ATAPacket_tmp) {
					CAM_ATAPacket_prev->PacketHead.NPA=CAM_ATAPacket;
					break;
				}
				
				CAM_ATAPacket_prev = (PCAM_ATA_PACKET)CAM_ATAPacket_prev->PacketHead.NPA;
			}
			CAM_ATAPacket->PacketHead.NPA = CAM_ATAPacket_tmp;
		}			
	}
	else {
		CAM_ATAPacket->PacketHead.NPA = CAM_ATAPacket_prev->PacketHead.NPA;
		CAM_ATAPacket_prev->PacketHead.NPA = CAM_ATAPacket;
	}



	CAM_ATAPacket_tmp = ATASortHead[bChannelNumber][bID];
	while ((PTR_SIZE)CAM_ATAPacket_tmp != PTR_NULL) {
		if (((PATA_TASK)CAM_ATAPacket_tmp->PacketHead.PTA)->bID == bID) {
			CAM_ATAPacket_tmp->PacketHead.CountStamp += 1;

			if((CAM_ATAPacket_tmp->PacketHead.CountStamp > PACKET_COUNTDOWN) && (CAM_ATAPacket_tmp != ATASortHead[bChannelNumber][bID])){
/*			if((CAM_ATAPacket_tmp->PacketHead.CountStamp > XXXXXX) && (CAM_ATAPacket_tmp != ATASortHead[bChannelNumber][bID])){
				CAM_ATAPacket_prev->PacketHead.NPA = CAM_ATAPacket_tmp->PacketHead.NPA;
				CAM_ATAPacket_tmp->PacketHead.NPA = ATASortHead[bChannelNumber][bID];
				ATASortHead[bChannelNumber][bID] = CAM_ATAPacket_tmp;

				CAM_ATAPacket_tmp = CAM_ATAPacket_prev->PacketHead.NPA;
				continue;
*/
				CAM_ATAPacket_tmp->PacketHead.CountStamp = 0;
				CAM_ATAPacket_tmp->PacketHead.Flag |= PACFLAG_SORT_PASS;

				if(CAM_ATAPacket_prev != ATASortHead[bChannelNumber][bID]){
					CAM_ATAPacket_prev->PacketHead.NPA = CAM_ATAPacket_tmp->PacketHead.NPA;
					CAM_ATAPacket_tmp->PacketHead.NPA = (ATASortHead[bChannelNumber][bID])->PacketHead.NPA;
					(ATASortHead[bChannelNumber][bID])->PacketHead.NPA = CAM_ATAPacket_tmp;
	
					CAM_ATAPacket_tmp = (PCAM_ATA_PACKET)CAM_ATAPacket_prev->PacketHead.NPA;
					continue;
				}
			}
		}
		CAM_ATAPacket_prev = CAM_ATAPacket_tmp;
		CAM_ATAPacket_tmp = (PCAM_ATA_PACKET)CAM_ATAPacket_tmp->PacketHead.NPA;
	}

	return;
}

/* 	1. put at head of ATAQueue not sort.
	2. device without packet in not sort.
*/
/* if can allocate packet, it can insert to ataqueue. */
U8 Put2ATAQueue(PCAM_ATA_PACKET CAM_ATAPacket,U8 direction)
{
	PATA_TASK pATATask;
	U8 bChannelNumber;
	PCAM_ATA_PACKET CAM_ATAPacket_tmp;

	pATATask = (PATA_TASK)CAM_ATAPacket->PacketHead.PTA;	
	bChannelNumber = pATATask->bCh2Chip;
	CAM_ATAPacket->PacketHead.NPA=(PCAM_ATA_PACKET)PTR_NULL;
	
	if (direction == QUEUETail){	/* put packet at the tail of packet queue */
		if (ATAQueueHead[bChannelNumber] != (PCAM_ATA_PACKET)PTR_NULL) {
			if (Check_Command_Sort(CAM_ATAPacket) == TRUE) { /* need sort */
				if (ATASortHead[bChannelNumber][pATATask->bID]==(PCAM_ATA_PACKET)PTR_NULL) {
					ATAQueueTail[bChannelNumber]->PacketHead.NPA=CAM_ATAPacket;
					ATAQueueTail[bChannelNumber]=CAM_ATAPacket;					
					ATASortHead[bChannelNumber][pATATask->bID]=CAM_ATAPacket;
				}
				else {
					SortDiskQueue(CAM_ATAPacket);
					
					CAM_ATAPacket_tmp = ATAQueueHead[bChannelNumber];
					while(CAM_ATAPacket_tmp->PacketHead.NPA!=(PCAM_ATA_PACKET)PTR_NULL) {
						CAM_ATAPacket_tmp = (PCAM_ATA_PACKET)CAM_ATAPacket_tmp->PacketHead.NPA;
					}
						
					ATAQueueTail[bChannelNumber]=CAM_ATAPacket_tmp;
				}
				ATAModuleNeedPush[bChannelNumber]=TRUE;
			
				return TRUE;					
			}
		}		
	}

	/* needn't sort */
	ATASortHead[bChannelNumber][pATATask->bID]=CAM_ATAPacket;
		
	if (direction == QUEUETail){	/* put packet at the tail of packet queue */	
		
		if (ATAQueueHead[bChannelNumber] == (PCAM_ATA_PACKET)PTR_NULL) {
			ATAQueueHead[bChannelNumber]=CAM_ATAPacket;
			ATAQueueTail[bChannelNumber]=CAM_ATAPacket;
			CAM_ATAPacket->PacketHead.NPA=(PCAM_ATA_PACKET)PTR_NULL;
		}
		else {
			ATAQueueTail[bChannelNumber]->PacketHead.NPA=CAM_ATAPacket;
			ATAQueueTail[bChannelNumber]=CAM_ATAPacket;
			CAM_ATAPacket->PacketHead.NPA=(PCAM_ATA_PACKET)PTR_NULL;			
		}
	}
	else{	/* direction == QUEUEHead  */	 /* put packet at the head of packet queue */

		if (ATAQueueHead[bChannelNumber] == (PCAM_ATA_PACKET)PTR_NULL) {
			ATAQueueHead[bChannelNumber]=CAM_ATAPacket;
			ATAQueueTail[bChannelNumber]=CAM_ATAPacket;
			CAM_ATAPacket->PacketHead.NPA=(PCAM_ATA_PACKET)PTR_NULL;			
		}
		else {
			CAM_ATAPacket->PacketHead.NPA=ATAQueueHead[bChannelNumber];
			ATAQueueHead[bChannelNumber]=CAM_ATAPacket;
		}
	}
	
	ATAModuleNeedPush[bChannelNumber]=TRUE;

	return TRUE;
}

U8 NCQ_Read_Log_Ext_Page10(U8 bChannelNumber)
{
	U8 		Adapter_ID=bChannelNumber/MAX_ATA_CHNL;
	U8		statusByte= 0;
	PTR_SIZE	BaseAddress = ATABaseAddr[bChannelNumber];
	PU8		base_addr;
	U8 		TAG;
	
    	if (Adapter_ID > (MAX_ADAPTER_NUM -1))
        	return U8NULL;
        
   	base_addr = (PU8)(gAdapterConfig[Adapter_ID].MemVirAddr);
	
   	base_addr += ATA_512K_BUFFER_ADDR;
	
	/* disable INTA here, it will be re-enable when CAM use SEQ 0 for packets */
	WriteIndexUchar(BaseAddress, iNDEXIDECICR, 0x0);  /* that the drive INT pass to SEQ 0*/
	WritePortUchar( HostBaseAddr[Adapter_ID], 0, 0x20); /* but mask SEQ 0 INT */
	
	/* select drive */	
	WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, 0xA0);
 	statusByte = WaitOnBusy(BaseAddress);
    	statusByte &= ~(IDE_STATUS_ERROR | IDE_STATUS_INDEX | IDE_STATUS_DRQ );
    	if ( (statusByte != IDE_STATUS_IDLE) && (statusByte != IDE_STATUS_DRDY) ) 
		return(U8NULL);
			
	/* send command */
	/* previous content */
	WriteIndexUchar(BaseAddress, iNDEXIDEFeature , 0);
	WriteIndexUchar(BaseAddress, iNDEXIDESectorCount , 0);
	WriteIndexUchar(BaseAddress, iNDEXIDESectorNumber , 0);
	WriteIndexUchar(BaseAddress, iNDEXIDECylinderLow , 0);
	WriteIndexUchar(BaseAddress, iNDEXIDECylinderHigh , 0);
	
	/* current content */
	WriteIndexUchar(BaseAddress, iNDEXIDEFeature , 0);
	WriteIndexUchar(BaseAddress, iNDEXIDESectorCount , 1);
	WriteIndexUchar(BaseAddress, iNDEXIDESectorNumber , 0x10);
	WriteIndexUchar(BaseAddress, iNDEXIDECylinderLow , 0);
	WriteIndexUchar(BaseAddress, iNDEXIDECylinderHigh , 0);
	
	WriteIndexUchar(BaseAddress, iNDEXIDEDeviceHead, 0xA0);
	WriteIndexUchar(BaseAddress, iNDEXIDECommand, IDE_COMMAND_READ_LOG_EXT);
	
	if ( gDrvConfig[bChannelNumber][0].DevFlag & DEV_ATA_DEVICE ) { /* ATA */
		if ( WaitForINT(BaseAddress) != camSUCCESS ){
			return(U8NULL);
		}
	}
	else { /* ATAPI */
		if ( WaitForATAPIINT(bChannelNumber) != camSUCCESS ){
			return(U8NULL);
		}		
	}
		
    /* Wait for DRQ */
   	statusByte = WaitForDrq(BaseAddress);
   	
   	ReadIndexUchar(BaseAddress, iNDEXIDEStatus); /*clear possible device INT */
   		
	if ( (!(statusByte & IDE_STATUS_DRQ)) | (statusByte& IDE_STATUS_BUSY) | (statusByte & IDE_STATUS_ERROR)) {	
		return(U8NULL);
	}

	/*
	 * Suck out 256 words. 
	 */
	 
    if( (ReadBuffer(bChannelNumber,(PU16)base_addr, 256)) != camSUCCESS)
		return(camFAIL);

	/* check status again, if DRQ still is 1. This command is failed. */	
	GetBaseStatus(BaseAddress, statusByte);
    if ( statusByte & IDE_STATUS_DRQ )
		return(camFAIL);
	if ((U8)*base_addr & bit(7))
		return(camFAIL);
	TAG= (U8)*base_addr & 0x1F;    		
	return (TAG);
}

/* 
GetATA_PacketFromSEQ_INFO : get the correspond packet from gSEQ_INFO 
ReturnATA_PacketFromSEQ_INFO : return the correspond packet from gSEQ_INFO, gSEQ_INFO not beed changed
*/
PCAM_PACKET ReturnATA_PacketFromSEQ_INFO(U8 bChannelNumber,U8 SeqID,U8 TAG)
{
	PSEQ_INFO		pCurSEQ_INFO;
	PCAM_PACKET		ultmpATAPacketAddress;
	U8			Adapter_ID = bChannelNumber/MAX_ATA_CHNL;	
		
	pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SeqID]);

	ultmpATAPacketAddress = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
	while ((PTR_SIZE)ultmpATAPacketAddress != PTR_NULL){
		
		if (((PCAM_PACKET)ultmpATAPacketAddress)->ATAPacket.bDelaySeqID==TAG){
			break;
		}
		/* next packet */
		ultmpATAPacketAddress = (PCAM_PACKET)ultmpATAPacketAddress->PacketHead.NPA;
	}

	return ultmpATAPacketAddress;
}

PCAM_PACKET GetATA_PacketFromSEQ_INFO(U8 bChannelNumber,U8 SeqID,U8 TAG)
{
	PSEQ_INFO		pCurSEQ_INFO;
	PCAM_PACKET		ultmpATAPacketAddress;
	PCAM_PACKET		ultmpATAPacketAddress_next;
	PCAM_PACKET		ultmpATAPacketAddress_prev=(PCAM_PACKET)PTR_NULL;
	U8			Adapter_ID = bChannelNumber/MAX_ATA_CHNL;	
		
	pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SeqID]);


	ultmpATAPacketAddress = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
	while ((PTR_SIZE)ultmpATAPacketAddress != PTR_NULL){
		if (((PCAM_PACKET)ultmpATAPacketAddress)->ATAPacket.bDelaySeqID==TAG){
			if (ultmpATAPacketAddress_prev==(PVOID)PTR_NULL) {
				pCurSEQ_INFO->SEQID2PacketHead = (PVOID)ultmpATAPacketAddress->PacketHead.NPA;
			}
			else {
				ultmpATAPacketAddress_prev->PacketHead.NPA=ultmpATAPacketAddress->PacketHead.NPA;
			}
				
			ultmpATAPacketAddress->PacketHead.NPA = (PVOID)PTR_NULL;
			break;
		}
		ultmpATAPacketAddress_next = (PCAM_PACKET)ultmpATAPacketAddress->PacketHead.NPA;
		
		ultmpATAPacketAddress_prev = ultmpATAPacketAddress;
		/* next packet */
		ultmpATAPacketAddress = ultmpATAPacketAddress_next;	
			
	}

	if (pCurSEQ_INFO->SEQID2PacketHead!=(PVOID)PTR_NULL) {
		/* re-set pCurSEQ_INFO->SEQID2PacketTail */
		ultmpATAPacketAddress_prev = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
		while ((PTR_SIZE)ultmpATAPacketAddress_prev != PTR_NULL){
		
			ultmpATAPacketAddress_next = (PCAM_PACKET)ultmpATAPacketAddress_prev->PacketHead.NPA;
		
			pCurSEQ_INFO->SEQID2PacketTail=(PVOID)ultmpATAPacketAddress_prev;
			/* next packet */
			ultmpATAPacketAddress_prev = ultmpATAPacketAddress_next;	
		}
	}
	else {
		pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;
		
		/* remove ata timeout timer */
		camTimer(Adapter_ID,0,(void (*)(void))ATA_Timeout_TimerID[bChannelNumber]);	
		ATA_Timeout_TimerID[bChannelNumber]=ZERO_NULL;
		ATAModuleSEQID[bChannelNumber] = U8NULL;
		ATAModuleBusy[bChannelNumber]=FALSE;
	}	

	return ultmpATAPacketAddress;
}

PCAM_PACKET GetATA_PacketFromATAQueue(U8 bChannelNumber)
{
	PCAM_ATA_PACKET	CAM_ATAPacket;
	PATA_TASK			pATATask;

	PCAM_PACKET			CAM_Packet;
	PCAM_ATA_SG		CAM_ATASG;

	if(ATAQueueHead[bChannelNumber] == (PCAM_ATA_PACKET)PTR_NULL)	/* no queue packet */
	{
		ATAModuleNeedPush[bChannelNumber]=FALSE;
		return ZERO_NULL;
	}

	/* get free packet address */
	CAM_Packet = (PCAM_PACKET)CAMAllocCamPacket(bChannelNumber);

	if (CAM_Packet == (PVOID)PTR_NULL){	/* no free ata packet. */
		return ZERO_NULL;		
	}	

	/* get free SG address */
	CAM_ATASG = (PCAM_ATA_SG)CAMAllocATASG(bChannelNumber);
	
	if (CAM_ATASG == (PVOID)PTR_NULL){	/* no free ata sg. */
		CAMFreeCamPacket(bChannelNumber, (PVOID)CAM_Packet);
		return ZERO_NULL;		
	}

	
	CAM_ATAPacket=ATAQueueHead[bChannelNumber];
	ATAQueueHead[bChannelNumber] = (PCAM_ATA_PACKET)CAM_ATAPacket->PacketHead.NPA;

	pATATask = (PATA_TASK)CAM_ATAPacket->PacketHead.PTA;
	
	prepareCamPacket(CAM_Packet, CAM_ATAPacket, CAM_ATASG, pATATask);	
	
	CAMFreeATAPacket(bChannelNumber , (PVOID)CAM_ATAPacket);

#ifdef _ERR_INJECT_
	if(ErrInj_Buffer_Table[bChannelNumber].count)
		CheckErrInjection(CAM_Packet, pATATask);
#endif



	/* still have packet */
	if(ATAQueueHead[bChannelNumber] != (PCAM_ATA_PACKET)PTR_NULL)	{
		if (ATASortHead[bChannelNumber][pATATask->bID]==CAM_ATAPacket) {
			PCAM_ATA_PACKET CAM_ATAPacket_tmp;
			/* find the next */ 
			ATASortHead[bChannelNumber][pATATask->bID]=(PCAM_ATA_PACKET)PTR_NULL;
			CAM_ATAPacket_tmp=ATAQueueHead[bChannelNumber];
			while ((PTR_SIZE)CAM_ATAPacket_tmp != PTR_NULL){
				if ( ((PATA_TASK)CAM_ATAPacket_tmp->PacketHead.PTA)->bID == pATATask->bID) {
					ATASortHead[bChannelNumber][pATATask->bID]=CAM_ATAPacket_tmp;
					break;
				}
				/* next packet */
				CAM_ATAPacket_tmp = (PCAM_ATA_PACKET)CAM_ATAPacket_tmp->PacketHead.NPA;
			}
		}
		ATAModuleNeedPush[bChannelNumber]=TRUE;

	}else {
		ATASortHead[bChannelNumber][pATATask->bID]=(PCAM_ATA_PACKET)PTR_NULL;
		ATAModuleNeedPush[bChannelNumber]=FALSE;
	}

	ATASortLBAl[bChannelNumber][pATATask->bID]=pATATask->Cmd.ATACmd.LBAl;
	ATASortLBAh[bChannelNumber][pATATask->bID]=pATATask->Cmd.ATACmd.LBAh;

	return CAM_Packet;
}

/* if packet at seq queue, return the first packet ATACmdFlag, else return ZERO_NULL*/
U16 SeqIDQueueFirstATATaskType(U8 Adapter_ID, U8 SeqID)
{
	PSEQ_INFO	pCurSEQ_INFO;
	PCAM_PACKET ultmpATAPacketAddress;
	PATA_TASK pATATask;
	
	pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SeqID]);
	
	if(pCurSEQ_INFO->SEQID2PacketHead==(PVOID)PTR_NULL)
		return ZERO_NULL;
	else{
		
		ultmpATAPacketAddress = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
					
		pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;

		switch(pATATask->Cmd.ATACmd.bCmd2Chip){
			/* NCQ commands */
			case IDE_COMMAND_READ_FPDMA_QUEUED:
			case IDE_COMMAND_WRITE_FPDMA_QUEUED:
			/* QUEUED commands */
			case IDE_COMMAND_READ_DMA_QUEUED:
			case IDE_COMMAND_READ_DMA_QUEUED_EXT:
			case IDE_COMMAND_WRITE_DMA_QUEUED:
			case IDE_COMMAND_WRITE_DMA_QUEUED_EXT:
			case IDE_COMMAND_WRITE_DMA_QUEUED_FUA_EXT:
				return (pATATask->ATACmdFlag | PSEUDO_TAG_CMD);
		}
								
		/* return (pATATask->ATACmdFlag & ~PSEUDO_TAG_CMD); */
	}
	return ((pATATask->ATACmdFlag & ~PSEUDO_TAG_CMD) | INTERNAL_EXIST);
}

/* if packet at seq queue, return the first PCAM_ATA_PACKET, else return ZERO_NULL*/
PCAM_PACKET SeqIDQueueFirstPACKET(U8 Adapter_ID, U8 SeqID)
{
	PSEQ_INFO	pCurSEQ_INFO;
	PCAM_PACKET ultmpATAPacketAddress;
	
	pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SeqID]);
	
	if(pCurSEQ_INFO->SEQID2PacketHead==(PVOID)PTR_NULL)
		return ZERO_NULL;
		
	ultmpATAPacketAddress = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
					
	return (ultmpATAPacketAddress);
}

U16 ATAQueueType(U8 bChannelNumber)
{
	PCAM_ATA_PACKET CAM_ATAPacket;
	PATA_TASK pATATask;

	if(ATAQueueHead[bChannelNumber] == (PCAM_ATA_PACKET)PTR_NULL)	/* no queue packet */
		return (ZERO_NULL);
	
	
	CAM_ATAPacket=ATAQueueHead[bChannelNumber];
	pATATask=(PATA_TASK)CAM_ATAPacket->PacketHead.PTA;

	switch(pATATask->Cmd.ATACmd.bCmd2Chip){
		/* NCQ commands */
		case IDE_COMMAND_READ_FPDMA_QUEUED:
		case IDE_COMMAND_WRITE_FPDMA_QUEUED:
		/* QUEUED commands */
		case IDE_COMMAND_READ_DMA_QUEUED:
		case IDE_COMMAND_READ_DMA_QUEUED_EXT:
		case IDE_COMMAND_WRITE_DMA_QUEUED:
		case IDE_COMMAND_WRITE_DMA_QUEUED_EXT:
		case IDE_COMMAND_WRITE_DMA_QUEUED_FUA_EXT:
			return (pATATask->ATACmdFlag | PSEUDO_TAG_CMD);
	}
	return ((pATATask->ATACmdFlag & ~PSEUDO_TAG_CMD) | INTERNAL_EXIST);							
}

U8 prepareATANCQTAG(PCAM_PACKET pCAMPacket, U8 TAG)
{
	/* set FPDMA TAG */
	pCAMPacket->ATAPacket.bDelaySeqID=TAG; 	
	/* set TAG of command packet bATRL */	
	pCAMPacket->ATAPacket.bATRL[17]=(TAG << 3);/* calculate by prepareATANCQCommand*/
			
	return TRUE;
}

/* 
prepareSATAPacketSEQID(): PATA can't run different device ID at the same command packet.
preparePATAPacketSEQID(): PATA can run master/slave device at the same command packet.
*/
U8 prepareSATAPacketSEQID(U8 bChannelNumber, U8 FreeSEQID)
{
	U8 PacketCount=ZERO_NULL;
	PCAM_ATA_PACKET CAM_ATAPacket;
	PCAM_PACKET CAM_Packet;
	
	U8 Adapter_ID;
	U16 SeqIDType;
	U16 packetTaskType;

	U8 bCh;
	U8 putAtLastPacket=FALSE;
	PSEQ_INFO	pCurSEQ_INFO;
	
	bCh=bChannelNumber%MAX_ATA_CHNL;	

	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;	
	
	pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][FreeSEQID]);	
	pCurSEQ_INFO->GO=TRUE;

	/* get ata packet. */
	while(1)
	{
		SeqIDType = SeqIDQueueFirstATATaskType(Adapter_ID, FreeSEQID);

		/* If a packet is PIO_XFER or SWAPBOX_CMD, it will not queue with other packets when executing.
			(Only DMA XFER and non-data packet can queued at executing packet queue.)		
		*/

		/* If the packet is ATAPI_DEVICE , it only can queued at the tail of execute packets queue(or only itself at the queue),
		    Because when errors occurred at ATAPI packet command, it must run "Request Sense Buffer" to get the sense datas before other
		    command had be sent.
		*/
		
		/* If the packet's ATACmdFlag be set with ATA_STATUS_RETURN , it only can queued at the tail of execute packets queue(or only itself at the queue),
		    Because when command success completed, also need read it's ata registers statses.
		*/
				
		if (SeqIDType == ZERO_NULL){	/* no packet at SeqIDQueue */

			if ((packetTaskType=ATAQueueType(bChannelNumber)) == ZERO_NULL)
				return PacketCount;
					
			if (packetTaskType & PIO_XFER)
				pCurSEQ_INFO->GO=FALSE;
			if (packetTaskType & SWAPBOX_CMD)
				pCurSEQ_INFO->GO=FALSE;
			else if (!(packetTaskType & ATA_DEVICE))	/* ATAPI Packet */
				pCurSEQ_INFO->GO=FALSE;
			else if (packetTaskType & ATA_STATUS_RETURN)	/* need read it's ata registers statses */
				pCurSEQ_INFO->GO=FALSE;
			CAM_Packet = GetATA_PacketFromATAQueue(bChannelNumber);
			if(CAM_Packet != ZERO_NULL)
				gChnlDrvControlConfig[bChannelNumber].Current_bID=((PATA_TASK)CAM_Packet->PacketHead.PTA)->bID;
		}
		else{	/* have packet at SeqIDQueue */
			if (SeqIDType & PIO_XFER){ /* Is PIO transfer command at SeqID queue */
				return PacketCount;
			}
			else if (SeqIDType & SWAPBOX_CMD){ /* Is SWAPBOX_CMD at SeqID queue */
				return PacketCount;
			}
			else if (!(SeqIDType & ATA_DEVICE))	/* ATAPI Packet */
				return PacketCount;
			else if (SeqIDType & ATA_STATUS_RETURN)	/* need read it's ata registers statses */
				return PacketCount;				
			else{
				/* get next packet type */
				if ((packetTaskType=ATAQueueType(bChannelNumber)) == ZERO_NULL)
					return PacketCount;
								
				if (packetTaskType & PIO_XFER){
					pCurSEQ_INFO->GO=FALSE;
					return PacketCount;
				}
 				if (packetTaskType & SWAPBOX_CMD){ /* SWAPBOX_CMD packet*/
 					pCurSEQ_INFO->GO=FALSE;
					return PacketCount;
				}				
 				else if (!(packetTaskType & ATA_DEVICE))	/* ATAPI Packet */
					putAtLastPacket=TRUE;			
				else if (SeqIDType & ATA_STATUS_RETURN)	/* need read it's ata registers statses */
					putAtLastPacket=TRUE;	
				/* check bID of first packet at ATAQueue is the same as Current_bID */
				CAM_ATAPacket=ATAQueueHead[bChannelNumber];
				if (((PATA_TASK)CAM_ATAPacket->PacketHead.PTA)->bID != gChnlDrvControlConfig[bChannelNumber].Current_bID)
					return PacketCount;
				else			
					CAM_Packet = GetATA_PacketFromATAQueue(bChannelNumber);
			}
		}
		
		if ( CAM_Packet == ZERO_NULL)	/* no packet */
			break;
			
		PacketCount++;
		
		/* set SEQID */
		CAM_Packet->ATAPacket.bSynSeqID=FreeSEQID; 		

		/* put ata packets to seqid2packet queue,so in isr, */
		/* driver can release the relevant resource and  */
		/* can point to the right PTA pointer */
		Put2SeqIDQueue(Adapter_ID,(PVOID)CAM_Packet,FreeSEQID,TRUE);

		if (putAtLastPacket)
		{
			pCurSEQ_INFO->GO=FALSE;
			break;
		}
			
		if (PacketCount>=ATAPacketQueueCount[Adapter_ID].ATAModule[bCh])
			break;		
	}
	return PacketCount;

}

U8 preparePATAPacketSEQID(U8 bChannelNumber, U8 FreeSEQID)
{
	U8 PacketCount=ZERO_NULL;
	/*PCAM_ATA_PACKET	CAM_ATAPacket;*/
	PCAM_PACKET			CAM_Packet;

	
	U8 Adapter_ID;
	U16 SeqIDType;
	U16 packetTaskType;

	U8 bCh;
	U8 putAtLastPacket=FALSE;
	PSEQ_INFO	pCurSEQ_INFO;
	
	bCh=bChannelNumber%MAX_ATA_CHNL;	

	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;	
	
	pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][FreeSEQID]);	
	pCurSEQ_INFO->GO=TRUE;

	/* get ata packet. */
	while(1)
	{
		SeqIDType = SeqIDQueueFirstATATaskType(Adapter_ID, FreeSEQID);

		/* If a packet is PIO_XFER or SWAPBOX_CMD, it will not queue with other packets when executing.
			(Only DMA XFER and non-data packet can queued at executing packet queue.)		
		*/

		/* If the packet is ATAPI_DEVICE , it only can queued at the tail of execute packets queue(or only itself at the queue),
		    Because when errors occurred at ATAPI packet command, it must run "Request Sense Buffer" to get the sense datas before other
		    command had be sent.
		*/
		
		/* If the packet's ATACmdFlag be set with ATA_STATUS_RETURN , it only can queued at the tail of execute packets queue(or only itself at the queue),
		    Because when command success completed, also need read it's ata registers statses.
		*/
				
		if (SeqIDType == ZERO_NULL){	/* no packet at SeqIDQueue */

			packetTaskType = ATAQueueType(bChannelNumber);
			if (packetTaskType & PIO_XFER)
				pCurSEQ_INFO->GO=FALSE;
			if (packetTaskType & SWAPBOX_CMD)
				pCurSEQ_INFO->GO=FALSE;
			else if (!(packetTaskType & ATA_DEVICE))	/* ATAPI Packet */
				pCurSEQ_INFO->GO=FALSE;
			else if (packetTaskType & ATA_STATUS_RETURN)	/* need read it's ata registers statses */
				pCurSEQ_INFO->GO=FALSE;
			CAM_Packet = GetATA_PacketFromATAQueue(bChannelNumber);
		}
		else{	/* have packet at SeqIDQueue */
			if (SeqIDType & PIO_XFER){ /* Is PIO transfer command at SeqID queue */
				return PacketCount;
			}
			else if (SeqIDType & SWAPBOX_CMD){ /* Is SWAPBOX_CMD at SeqID queue */
				return PacketCount;
			}
			else if (!(SeqIDType & ATA_DEVICE))	/* ATAPI Packet */
				return PacketCount;
			else if (SeqIDType & ATA_STATUS_RETURN)	/* need read it's ata registers statses */
				return PacketCount;				
			else{
				/* get next packet type */
				packetTaskType = ATAQueueType(bChannelNumber);
				if (packetTaskType & PIO_XFER){
					pCurSEQ_INFO->GO=FALSE;
					return PacketCount;
				}
 				if (packetTaskType & SWAPBOX_CMD){ /* SWAPBOX_CMD packet*/
 					pCurSEQ_INFO->GO=FALSE;
					return PacketCount;
				}				
 				else if (!(packetTaskType & ATA_DEVICE))	/* ATAPI Packet */
					putAtLastPacket=TRUE;			
				else if (SeqIDType & ATA_STATUS_RETURN)	/* need read it's ata registers statses */
					putAtLastPacket=TRUE;				
				CAM_Packet = GetATA_PacketFromATAQueue(bChannelNumber);
			}
		}
		
		if ( CAM_Packet == ZERO_NULL)	/* no packet */
			break;
			
		PacketCount++;
		
		/* set SEQID */
		CAM_Packet->ATAPacket.bSynSeqID=FreeSEQID; 		

		/* put ata packets to seqid2packet queue,so in isr, */
		/* driver can release the relevant resource and  */
		/* can point to the right PTA pointer */
		Put2SeqIDQueue(Adapter_ID,(PVOID)CAM_Packet,FreeSEQID,TRUE);

		if (putAtLastPacket)
		{
			pCurSEQ_INFO->GO=FALSE;
			break;
		}
			
		if (PacketCount>=ATAPacketQueueCount[Adapter_ID].ATAModule[bCh])
			break;		
	}
	return PacketCount;

}


void ReSortNCQDiskQueue(U8 bChannelNumber)
{
	PCAM_ATA_PACKET CAM_ATAPacket_prev;
	PCAM_ATA_PACKET CAM_ATAPacket_tmp;
	PCAM_ATA_PACKET CAM_ATAPacket_insert;
	PATA_TASK pATATask;
	U8 bID=0;
	U8	HitFlag;
	U8	InsertCount = 32 - ATAPacketCurrent[bChannelNumber];
	U8	DataIn;	
	
	CAM_ATAPacket_prev = (PCAM_ATA_PACKET)PTR_NULL;
	CAM_ATAPacket_insert = (PCAM_ATA_PACKET)PTR_NULL;
	CAM_ATAPacket_tmp = ATAQueueHead[bChannelNumber];

	pATATask = (PATA_TASK)CAM_ATAPacket_tmp->PacketHead.PTA;
	if(pATATask->ATACmdFlag & camDATA_IN)
		DataIn = TRUE;
	else				
		DataIn = FALSE;

	/* get insert point */
	HitFlag = FALSE;
	while ((PTR_SIZE)CAM_ATAPacket_tmp != PTR_NULL) {
		pATATask = (PATA_TASK)CAM_ATAPacket_tmp->PacketHead.PTA;
		if (pATATask->bID == bID) {
			
			if( ! (CAM_ATAPacket_tmp->PacketHead.Flag & PACFLAG_NCQ))
				break;
			if(CAM_ATAPacket_tmp == ATAQueueTail[bChannelNumber])
				break;
				
			if(DataIn == TRUE){
				if(!(pATATask->ATACmdFlag & camDATA_IN)){
					HitFlag = TRUE;
				}
			}
			else{
				if(pATATask->ATACmdFlag & camDATA_IN){
					HitFlag = TRUE;
				}
			}
			
			if(HitFlag == TRUE){
				CAM_ATAPacket_insert = CAM_ATAPacket_prev;
				break;
			}
		}		
		CAM_ATAPacket_prev = CAM_ATAPacket_tmp;
		CAM_ATAPacket_tmp = (PCAM_ATA_PACKET)CAM_ATAPacket_tmp->PacketHead.NPA;
	}		



	if(HitFlag == TRUE){
		/* resort all R/W task */
		//CAM_ATAPacket_prev = (PCAM_ATA_PACKET)PTR_NULL;
		CAM_ATAPacket_prev = CAM_ATAPacket_insert;
		
		if(CAM_ATAPacket_insert == (PCAM_ATA_PACKET)PTR_NULL)
			CAM_ATAPacket_tmp = ATAQueueHead[bChannelNumber];
		else
			CAM_ATAPacket_tmp = (PCAM_ATA_PACKET)CAM_ATAPacket_insert->PacketHead.NPA;
		
		HitFlag = FALSE;
		while ((PTR_SIZE)CAM_ATAPacket_tmp != PTR_NULL) {
			pATATask = (PATA_TASK)CAM_ATAPacket_tmp->PacketHead.PTA;
			if (pATATask->bID == bID) {
				
				if( ! (CAM_ATAPacket_tmp->PacketHead.Flag & PACFLAG_NCQ))
					break;
				if(CAM_ATAPacket_tmp == ATAQueueTail[bChannelNumber])
					break;
				if(!InsertCount)
					break;
	
	
				if(DataIn == TRUE){
					if(pATATask->ATACmdFlag & camDATA_IN){
						HitFlag = TRUE;
					}
				}
				else{
					if(!(pATATask->ATACmdFlag & camDATA_IN)){
						HitFlag = TRUE;
					}
				}
				if(HitFlag == TRUE){
					
					if(CAM_ATAPacket_insert == (PCAM_ATA_PACKET)PTR_NULL){
						Put2ATAQueue(CAM_ATAPacket_tmp, QUEUEHead);
					}
					else{	
						CAM_ATAPacket_prev->PacketHead.NPA = CAM_ATAPacket_tmp->PacketHead.NPA;
						CAM_ATAPacket_tmp->PacketHead.NPA = CAM_ATAPacket_insert->PacketHead.NPA;				
						CAM_ATAPacket_insert->PacketHead.NPA = CAM_ATAPacket_tmp;
					}
					CAM_ATAPacket_insert = CAM_ATAPacket_tmp;
					CAM_ATAPacket_tmp = CAM_ATAPacket_prev;
					HitFlag = FALSE;
				}
			}		
			CAM_ATAPacket_prev = CAM_ATAPacket_tmp;
			CAM_ATAPacket_tmp = (PCAM_ATA_PACKET)CAM_ATAPacket_tmp->PacketHead.NPA;
			InsertCount--;
		}		
	}
	return;
}


void StartATANCQOperation(U8 bChannelNumber)
{
	U8 				SEQID;
	U8 				TAG;
	U8 				Adapter_ID;
	U32  			tmpData, j;
	PCAM_ATA_PACKET CAM_ATAPacket;	
	PATA_TASK 		pATATask;
	U8 				i;
	U8				sendCommand=FALSE;	
	U8				ASIC_NCQ_BusyFlag=FALSE;

	PCAM_PACKET		CAM_Packet;	

	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;

	if (ATAQueueType(bChannelNumber) == ZERO_NULL)
		return;
		
	SEQID = gChnlDrvControlConfig[bChannelNumber].SEQID;
	ATAModuleSEQID[bChannelNumber]=SEQID;
	
	if (ATAModuleBusy[bChannelNumber]==TRUE) {
		if (!(SeqIDQueueFirstATATaskType(Adapter_ID, gChnlDrvControlConfig[bChannelNumber].SEQID) & PSEUDO_TAG_CMD)) /* Not NCQ command running */
			return;
		
		/* check is the same bID */
		CAM_ATAPacket=ATAQueueHead[bChannelNumber];
		pATATask=(PATA_TASK)CAM_ATAPacket->PacketHead.PTA;
				
		if (pATATask->bID != gChnlDrvControlConfig[bChannelNumber].Current_bID)
			return;
	}

		
	/* when NCQ running, can't write FPDMA Control/Status Register whole DWORD */
	if (gChnlDrvControlConfig[bChannelNumber].UsedTAG == 0x0) {
		/* set Sequence Counter Control Register */
		WritePortUlong(HostBaseAddr[Adapter_ID], SEQID*4, SEQ_FPDMA|0);
		/* write FPDMA Control/Status Register */
		tmpData = FPDMA_Ctrl_Status_FPDMA_auto | FPDMA_Ctrl_Status_DMAsetup_INT_mask;	
		/*tmpData = FPDMA_Ctrl_Status_FPDMA_auto | FPDMA_Ctrl_Status_Packet_PRD | FPDMA_Ctrl_Status_DMAsetup_INT_mask;*/
		WriteIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_Ctrl_Status, tmpData);
		gChnlDrvControlConfig[bChannelNumber].TagCmdFlag &= ~(TAG_DATAOUT|TAG_DATAIN);
	}

	while (ATAQueueType(bChannelNumber) & PSEUDO_TAG_CMD) {

		CAM_ATAPacket=ATAQueueHead[bChannelNumber];
		pATATask=(PATA_TASK)CAM_ATAPacket->PacketHead.PTA;
				
		/* check is the same bID */
		if (ATAModuleBusy[bChannelNumber]==TRUE) {
			/* check is the same bID */
			if (pATATask->bID != gChnlDrvControlConfig[bChannelNumber].Current_bID)
				break;
		}
			
		if(pATATask->ATACmdFlag & camDATA_IN){
			if(gChnlDrvControlConfig[bChannelNumber].TagCmdFlag & TAG_DATAOUT)
				break;
		}
		else{
			if(gChnlDrvControlConfig[bChannelNumber].TagCmdFlag & TAG_DATAIN)
				break;
		}

		/* get free tag */
		for (i=0;i<gDrvConfig[pATATask->bCh2Chip][pATATask->bID].QueueDepth;i++) {
			if (gChnlDrvControlConfig[bChannelNumber].FreeTAG & (1 << i) )
				break;
		}
		if (i >= gDrvConfig[pATATask->bCh2Chip][pATATask->bID].QueueDepth)	/* no free tag */	
			break;
		
		ATAModuleBusy[bChannelNumber]=TRUE;	
		gSEQ_INFO[Adapter_ID][SEQID].GO=TRUE;
		
		TAG=i;

		gChnlDrvControlConfig[bChannelNumber].UsedTAG |= (1 << TAG);
		gChnlDrvControlConfig[bChannelNumber].FreeTAG &= ~(1 << TAG);
		

		/* tune performance for ASIC bug */
		if(!(gChnlDrvControlConfig[bChannelNumber].TagCmdFlag & (TAG_DATAOUT|TAG_DATAIN)))
			ReSortNCQDiskQueue(bChannelNumber);

		CAM_Packet = GetATA_PacketFromATAQueue(bChannelNumber);
		if (CAM_Packet == ZERO_NULL) {
			gChnlDrvControlConfig[bChannelNumber].UsedTAG &= ~(1 << TAG);
			gChnlDrvControlConfig[bChannelNumber].FreeTAG |= (1 << TAG);
			break;
		}

		/* polling ASIC status for submit NCQ command */
		ASIC_NCQ_BusyFlag = FALSE;
		for (j=0; j<100000; j++) {
			ReadPortUlong(ATABaseAddr[bChannelNumber], ATA_STATUS_REGISTER_OFFSET, tmpData);
			if (!(tmpData & bit(29))){ 
				ReadIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_Ctrl_Status, tmpData);
				if (!(tmpData & bit(14))) 
					break;
			}
			camUStallExecution(1);
		}
		if ( j >=100000 ){
			ASIC_NCQ_BusyFlag = TRUE;
			//ATA_Timeout(bChannelNumber); /* time-out */
			//return;
		}

		if(!(gChnlDrvControlConfig[bChannelNumber].TagCmdFlag & (TAG_DATAOUT|TAG_DATAIN))){
			if(pATATask->ATACmdFlag & camDATA_IN)
				gChnlDrvControlConfig[bChannelNumber].TagCmdFlag |= TAG_DATAIN;
			else
				gChnlDrvControlConfig[bChannelNumber].TagCmdFlag |= TAG_DATAOUT;
		}
		

		prepareATANCQTAG(CAM_Packet, TAG);
		/* set SEQID */
		CAM_Packet->ATAPacket.bSynSeqID=SEQID; 		
		Put2SeqIDQueue(Adapter_ID,(PVOID)CAM_Packet,SEQID,FALSE);

		sendCommand = TRUE;
		gChnlDrvControlConfig[bChannelNumber].Current_bID=((PATA_TASK)CAM_Packet->PacketHead.PTA)->bID;
		
		SYNC_CPU_CACHE();

		if(ASIC_NCQ_BusyFlag == TRUE){
			break;
		}
		else{
			/* set TAG position */
			WriteIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_SActive, (U32)bit(TAG));
			/* Write FPDMA PRD High base address */
			WriteIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_PRD_High_Addr, (CAM_Packet->ATAPacket.ulPSG & 0xFFFF0000));
	
			WritePortUlong(ATABaseAddr[bChannelNumber], ATA_PACSTART_REGISTER_OFFSET, (U32)((PPACKET_HEAD)CAM_Packet)->Address);
	
	//		/* Write FPDMA PRD High base address */
	//		WriteIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_PRD_High_Addr, (CAM_Packet->ATAPacket.ulPSG & 0xFFFF0000));
	//		/* Write FPDMA Packet High base address */
	//		WriteIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_Packet_High_Addr, ((U32)((PPACKET_HEAD)CAM_Packet)->Address & 0xFFFF0000));				
	//		/* Write FPDMA Vector table register */
	//		WriteIndexWord(ATA_ExpBaseAddr[bChannelNumber], (U8)(iNDEXeXP_FPDMA_Vector_Table_Base+(TAG*2)), (U16)((((PPACKET_HEAD)CAM_Packet)->Address & 0x0000FFFF)));
		}
	}/* end of while () */

	if (sendCommand == TRUE) {
	
		/* set ATA Module timeout timer. */
		if (ATA_Timeout_TimerID[bChannelNumber]!=ZERO_NULL) {
			/* remove ata timeout timer */
			camTimer(Adapter_ID,0,(void (*)(void))ATA_Timeout_TimerID[bChannelNumber]);	
			ATA_Timeout_TimerID[bChannelNumber]=ZERO_NULL;	
		}
		
		if(ASIC_NCQ_BusyFlag == TRUE){
			ATA_Timeout(bChannelNumber); /* time-out */
			return;
		}
		
		ATA_Timeout_TimerID[bChannelNumber]=camTimer(Adapter_ID,CAMTimeoutTimes[Adapter_ID].ATACommand,ATA_Timeout_Function[bChannelNumber]);

		if ((gBOXConfig[bChannelNumber].BOXType != camNOBOX) && (gBOXConfig[bChannelNumber].BOXType & SATAEnclosure)) {
			/* turn on LED */
			CAM_SBOX_HDAccess_LED(bChannelNumber , HDActiveON);
		}			
	}
	return;
}


void StartATAQUEUEDOperation(U8 bChannelNumber)
{
	U8 				SEQID;
	U8 				TAG=0;
	U8 				Adapter_ID;
	U32  			tmpData;
	PCAM_ATA_PACKET CAM_ATAPacket;	
	PATA_TASK 		pATATask=0;
	U32 			startingSector;
	U8 				i;
	U8 				Status;
	U16				tempdata16=0;
	U32				StatusPollingCount;
	U8				ServiceFlag=FALSE, ErrorOccurFlag=FALSE, sendCommand=FALSE, transferReady=FALSE;				

	PCAM_PACKET		CAM_Packet;	


	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;

	if (ATAQueueType(bChannelNumber) == ZERO_NULL)
		return;
		
	if (ATAModuleBusy[bChannelNumber]==TRUE) {
		if (!(SeqIDQueueFirstATATaskType(Adapter_ID, gChnlDrvControlConfig[bChannelNumber].SEQID) & PSEUDO_TAG_CMD)) /* Not QUEUED command running */
			return;
			
		/* check is the same bID */
		CAM_ATAPacket = ATAQueueHead[bChannelNumber];
		pATATask=(PATA_TASK)CAM_ATAPacket->PacketHead.PTA;
		
		if (pATATask->bID != gChnlDrvControlConfig[bChannelNumber].Current_bID)
			return;			
	}

	SEQID = gChnlDrvControlConfig[bChannelNumber].SEQID;
	ATAModuleSEQID[bChannelNumber]=SEQID;

	/* set Sequence Counter Control Register */
	WritePortUlong(HostBaseAddr[Adapter_ID], SEQID*4, SEQ_FPDMA|0);
	
	gSEQ_INFO[Adapter_ID][SEQID].GO=TRUE;

	while (ATAQueueType(bChannelNumber) & PSEUDO_TAG_CMD) {

		CAM_ATAPacket=ATAQueueHead[bChannelNumber];
		pATATask=(PATA_TASK)CAM_ATAPacket->PacketHead.PTA;

		/* check is the same bID */
		if (ATAModuleBusy[bChannelNumber]==TRUE) {
			/* check is the same bID */
			if (pATATask->bID != gChnlDrvControlConfig[bChannelNumber].Current_bID) {
				if (sendCommand == TRUE)
					break;
				else
					return;				
			}
		}
						
		/* read alt-status*/
		Status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEAlternateStatus);
		if ( (Status & IDE_STATUS_BUSY) || (Status & IDE_STATUS_DRQ) ) {
			if (sendCommand == TRUE)
				break;
			else
				return;
		}
		
		if (gSEQ_INFO[Adapter_ID][SEQID].SEQID2PacketHead != ((PVOID)PTR_NULL) )	{
			if (gChnlDrvControlConfig[bChannelNumber].TCQTransferData!=FALSE) {
				if (sendCommand == TRUE)
					break;
				else
					return;
			}
		}
						
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECICR, SEQID);  /* pass the drive INT pass to SEQ # */
	
		/* get free tag */
		for (i=0;i<gDrvConfig[pATATask->bCh2Chip][pATATask->bID].QueueDepth;i++) {
			if ( (gChnlDrvControlConfig[bChannelNumber].FreeTAG & (1 << i))
					&& (!(gChnlDrvControlConfig[bChannelNumber].UsedTAG & (1 << i))) )
				break;
		}
		if (i >= gDrvConfig[pATATask->bCh2Chip][pATATask->bID].QueueDepth) {	/* no free tag */	
			if (sendCommand == TRUE)
				break;
			else
				return;
		}

		ATAModuleBusy[bChannelNumber]=TRUE;	
		
		
		TAG=i;

		gChnlDrvControlConfig[bChannelNumber].UsedTAG |= (1 << TAG);
		gChnlDrvControlConfig[bChannelNumber].FreeTAG &= ~(1 << TAG);
		
		CAM_Packet = GetATA_PacketFromATAQueue(bChannelNumber);
		if (CAM_Packet == ZERO_NULL) {
			gChnlDrvControlConfig[bChannelNumber].UsedTAG &= ~(1 << TAG);
			gChnlDrvControlConfig[bChannelNumber].FreeTAG |= (1 << TAG);
			if (sendCommand == TRUE)
				break;
			else
				return;
		}

		
		CAM_Packet->ATAPacket.bDelaySeqID=TAG; 
		/* set SEQID */
		CAM_Packet->ATAPacket.bSynSeqID=SEQID; 		
		Put2SeqIDQueue(Adapter_ID,(PVOID)CAM_Packet,SEQID,FALSE);
		
		pATATask=(PATA_TASK)CAM_Packet->PacketHead.PTA;

		gChnlDrvControlConfig[bChannelNumber].Current_bID=pATATask->bID;
		sendCommand = TRUE;

		
		if(gChnlDrvControlConfig[bChannelNumber].IsWDHardDrive != TRUE) /* Hitachi */
			/* Mask SEQId INT */
			WritePortUlong(HostBaseAddr[Adapter_ID], SEQID*4, SEQ_FPDMA|SEQ_INT_MASK|0);

				
		if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_48BIT) {
			/* previous content */
			WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEFeature , ((U8)(pATATask->Cmd.ATACmd.bCount >> 8)) ); 	/* SecCnt 15:8 */
			WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorCount , 0); 	/* Reserved */
			startingSector=pATATask->Cmd.ATACmd.LBAl;
			WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorNumber, ((U8)( startingSector >> 24 )) );	/* LBA 31:24 */
			
			startingSector=pATATask->Cmd.ATACmd.LBAh;
			WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECylinderLow, ((U8)( startingSector)) );	/* LBA 39:32 */
			WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECylinderHigh, ((U8)( startingSector >> 8 )) );	/* LBA 47:40 */
		}

		/* most recently written */
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEFeature , ((U8)pATATask->Cmd.ATACmd.bCount) ); 	/* SecCnt 7:0 */
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorCount , (U8)(TAG << 3) ); 	/* TAG */
		
		startingSector=pATATask->Cmd.ATACmd.LBAl;
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorNumber, ((U8)( startingSector)) );	/* LBA 7:0 */
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECylinderLow, ((U8)( startingSector >> 8 )) );	/* LBA 15:8 */
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECylinderHigh, ((U8)( startingSector >> 16 )) );	/* LBA 23:16 */

		/* select drive */
		if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) {/* SATA */
			if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_48BIT) 					
				WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEDeviceHead, 0xE0);
			else 
				/* LBA 27:24 */
				WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEDeviceHead, ((U8) (( ((U8)( startingSector >> 24 )) & 0x0F) | 0xE0  )) );
		}
		else	{
			if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_48BIT) 					
				WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEDeviceHead, ((U8) (0xE0 | (pATATask->bID << 4))) );
			else 
				/* LBA 27:24 */
				WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEDeviceHead, ((U8) (( ((U8)( startingSector >> 24 )) & 0x0F) | 0xE0 | (pATATask->bID << 4))) );
		}
		
		/* set Sequence Counter Control Register */
		/* not support service interrupt */
		
		WriteIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDECommand, pATATask->Cmd.ATACmd.bCmd2Chip );
		
		/* wait at least 400 ns */
		/* read 2 port will more than 400 ns */
		ReadPortUlong( HostBaseAddr[Adapter_ID],oFFSETHOSTEXTRA0 ,tmpData);
		ReadPortUlong( HostBaseAddr[Adapter_ID],oFFSETHOSTEXTRA0 ,tmpData);

		/* check servuce, DRQ & not busy */
		Status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);

		StatusPollingCount = 0;
		ServiceFlag = FALSE;
		transferReady = FALSE;
		ErrorOccurFlag = FALSE;
		
		while(1) {
			if(StatusPollingCount >= 100000){	/* polling above 1 sec */
				ErrorOccurFlag = TRUE;
				break;
			}

			if (!(Status & IDE_STATUS_BUSY)) {
				if (Status & IDE_STATUS_SERVICE) {
					/*ServiceFlag = TRUE;*/
					/*transferReady = TRUE;*/
					transferReady = FALSE;
					break;
				}
				else if (Status & IDE_STATUS_DRQ) {
					ServiceFlag = FALSE;
					transferReady = TRUE;
					break;
				}
				else {
					Status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorCount);
					if (Status & IDE_SectorCount_RELEASE) {
						transferReady = FALSE;
						break;					
					}
				}
			}

			/* This is dongle bug, when device DRQ assert, status 0x48 willnot update to host.*/
			if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA )
			{
				#ifdef _MMIO_
				ReadIndexWord(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDELatestFISD2H, tempdata16 );
				#else
				ReadIndexWord(ATABaseAddr[bChannelNumber], iNDEXIDELatestFISD2H,tempdata16);
				#endif					
				if (pATATask->ATACmdFlag & camDATA_IN) /* read */
				{
					if ( ((U8)(tempdata16&0x00FF)) == 0x46) {
						U32 tempdata32;
						
						#ifdef _MMIO_
						ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEMiscInternalControlState, tempdata32 );
						#else
						ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDEMiscInternalControlState,tempdata32);
						#endif
						
						if (!(tempdata32 & bit(24))) {
							ServiceFlag = FALSE;
							transferReady = TRUE;
							break;
						}
					}
				}
				else { /* write  */
					if ( ((U8)(tempdata16&0x00FF)) == 0x39) {
						ServiceFlag = FALSE;
						transferReady = TRUE;
						break;
					}	
				}
			}
			/* check servuce, DRQ & not busy */
			camUStallExecution(1);			
			Status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);
			StatusPollingCount++;
		}

		if ((gBOXConfig[bChannelNumber].BOXType != camNOBOX) && (gBOXConfig[bChannelNumber].BOXType & SATAEnclosure)) {
			/* turn on LED */
			CAM_SBOX_HDAccess_LED(bChannelNumber , HDActiveON);
		}
				
		if ((transferReady == TRUE) || (ErrorOccurFlag == TRUE))
			break;
	} /* end while */

	/* set ATA Module timeout timer. */
	if (ATA_Timeout_TimerID[bChannelNumber]!=ZERO_NULL) {
		/* remove ata timeout timer */
		camTimer(Adapter_ID,0,(void (*)(void))ATA_Timeout_TimerID[bChannelNumber]);	
		ATA_Timeout_TimerID[bChannelNumber]=ZERO_NULL;	
	}
	ATA_Timeout_TimerID[bChannelNumber]=camTimer(Adapter_ID,CAMTimeoutTimes[Adapter_ID].ATACommand,ATA_Timeout_Function[bChannelNumber]);


	if (gSEQ_INFO[Adapter_ID][SEQID].SEQID2PacketHead == ((PVOID)PTR_NULL) )	
		return;

	if ( (transferReady == FALSE) && (ErrorOccurFlag != TRUE) ){
		if(gChnlDrvControlConfig[bChannelNumber].IsWDHardDrive == TRUE) /* not Hitachi */
			return;
			
		/* check servuce, DRQ & not busy */
		Status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);

		StatusPollingCount = 0;
		ServiceFlag = FALSE;
		transferReady = FALSE;
		ErrorOccurFlag = FALSE;

		while(1) {
			if(StatusPollingCount >= 100000){	/* polling above 1 sec */
				ErrorOccurFlag = TRUE;
				break;
			}

			if (!(Status & IDE_STATUS_BUSY)) {
				if (Status & IDE_STATUS_SERVICE) {
					ServiceFlag = TRUE;
					transferReady = TRUE;
					break;
				}
				else if (Status & IDE_STATUS_DRQ) {
					ServiceFlag = FALSE;
					transferReady = TRUE;
					break;
				}
			}
			/* This is dongle bug, when device DRQ assert, status 0x48 willnot update to host.*/
			if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA )
			{
				#ifdef _MMIO_
				ReadIndexWord(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDELatestFISD2H, tempdata16 );
				#else
				ReadIndexWord(ATABaseAddr[bChannelNumber], iNDEXIDELatestFISD2H,tempdata16);
				#endif					
				if (pATATask->ATACmdFlag & camDATA_IN) /* read */
				{
					if ( ((U8)(tempdata16&0x00FF)) == 0x46) {
						U32 tempdata32;
						
						#ifdef _MMIO_
						ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEMiscInternalControlState, tempdata32 );
						#else
						ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDEMiscInternalControlState,tempdata32);
						#endif
						
						if (!(tempdata32 & bit(24))) {
							ServiceFlag = FALSE;
							transferReady = TRUE;
							break;
						}
					}
				}
				else { /* write  */
					if ( ((U8)(tempdata16&0x00FF)) == 0x39) {
						ServiceFlag = FALSE;
						transferReady = TRUE;
						break;
					}	
				}
			}
			/* check servuce, DRQ & not busy */
			camUStallExecution(1);
			Status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);
			StatusPollingCount++;					
		}
	}
				
	if (ServiceFlag == TRUE) {
				
		gChnlDrvControlConfig[bChannelNumber].TCQTransferData=TRUE;
		/* send service command */
		IssueService(bChannelNumber,0,IDE_COMMAND_SERVICE);

		Status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);
		StatusPollingCount = 0;
		ErrorOccurFlag = FALSE;
		
		/* wait DRQ */	
		while(1) {
			if(StatusPollingCount >= 100000){	/* polling above 1 sec */
				ErrorOccurFlag = TRUE;
				break;
			}

			if (Status & IDE_STATUS_DRQ) {
				break;
			}
			camUStallExecution(1);
			Status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);
			StatusPollingCount++;
		}
				
		if(ErrorOccurFlag != TRUE){
			/* get the TAG */
			TAG = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorCount);
			TAG = (TAG >> 3);
			gChnlDrvControlConfig[bChannelNumber].CurrentXferTAG = TAG;
			/* prepare SG */
		
			if(gChnlDrvControlConfig[bChannelNumber].IsWDHardDrive != TRUE){ /* Hitachi */
				if(sendCommand == TRUE)	/*clear possible device INT */	
					ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);
			}
	
			/* return the correspond packet from gSEQ_INFO */
			CAM_Packet=ReturnATA_PacketFromSEQ_INFO(bChannelNumber,SEQID,TAG);
			
			if (CAM_Packet == (PCAM_PACKET)PTR_NULL)
				return;
						
			pATATask=(PATA_TASK)CAM_Packet->PacketHead.PTA;
				
			/* write SG table */
			WriteIndexUlong(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDEDestriptorPointer, CAM_Packet->ATAPacket.ulPSG);
			
			if(gChnlDrvControlConfig[bChannelNumber].IsWDHardDrive == TRUE){ /* not Hitachi */
				/* set Sequence Counter Control Register */
				WritePortUlong(HostBaseAddr[Adapter_ID], SEQID*4, SEQ_FPDMA|0);
			}
				
			/* start DMA transfer */
			if (pATATask->ATACmdFlag & camDATA_IN) /* read */
				WriteIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDECICR, (U8)(IDECICR_DMA_Start|SEQID) );		
			else /* write */
				WriteIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDECICR, (U8)(IDECICR_DMA_Start|IDECICR_DMA_RW_Direction|SEQID) );				
	
			if(gChnlDrvControlConfig[bChannelNumber].IsWDHardDrive != TRUE){ /* Hitachi */
				if(sendCommand == TRUE)	/* UnMask SEQId INT */
					WritePortUlong(HostBaseAddr[Adapter_ID], SEQID*4, SEQ_FPDMA|0);		
			}
		}
	}
	else if(ErrorOccurFlag != TRUE){ /* ServiceFlag != TRUE */
	
		gChnlDrvControlConfig[bChannelNumber].TCQTransferData=TRUE;
		gChnlDrvControlConfig[bChannelNumber].CurrentXferTAG = TAG;

		if(gChnlDrvControlConfig[bChannelNumber].IsWDHardDrive != TRUE){ /* Hitachi */
			if(sendCommand == TRUE)	/*clear possible device INT */	
				ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);
		}

		/* return the correspond packet from gSEQ_INFO */
		CAM_Packet=ReturnATA_PacketFromSEQ_INFO(bChannelNumber,SEQID,TAG);
		pATATask=(PATA_TASK)CAM_Packet->PacketHead.PTA;
		/* write SG table */
		WriteIndexUlong(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDEDestriptorPointer, CAM_Packet->ATAPacket.ulPSG);
		
		if(gChnlDrvControlConfig[bChannelNumber].IsWDHardDrive == TRUE){ /* not Hitachi */
			/* set Sequence Counter Control Register */
			WritePortUlong(HostBaseAddr[Adapter_ID], SEQID*4, SEQ_FPDMA|0);
		}
			
		/* start DMA transfer */
		if (pATATask->ATACmdFlag & camDATA_IN) /* read */
			WriteIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDECICR, (U8)(IDECICR_DMA_Start|SEQID) );		
		else /* write */
			WriteIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDECICR, (U8)(IDECICR_DMA_Start|IDECICR_DMA_RW_Direction|SEQID) );		

		if(gChnlDrvControlConfig[bChannelNumber].IsWDHardDrive != TRUE){ /* Hitachi */	
			if(sendCommand == TRUE)	/* UnMask SEQId INT */
				WritePortUlong(HostBaseAddr[Adapter_ID], SEQID*4, SEQ_FPDMA|0);
		}
	
	}

	if(ErrorOccurFlag == TRUE){
		if (ATA_Timeout_TimerID[bChannelNumber] != ZERO_NULL)
		/* remove ata timeout timer */
			camTimer(Adapter_ID,0,(void (*)(void))ATA_Timeout_TimerID[bChannelNumber]);	

		ATA_Timeout(bChannelNumber);
	}

	return;
}

	
void StartATAOperation(U8 bChannelNumber)
{
	U8 PacketCount=ZERO_NULL;
	U8 FreeSEQID;
	U8 Adapter_ID;
	U16 SeqIDType;
	PCAM_ATA_PACKET CAM_ATAPacket;

	Adapter_ID = bChannelNumber/MAX_ATA_CHNL;


	if((ATAModuleBusy[bChannelNumber] != TRUE) && (NeedCheckFIFO[bChannelNumber] == TRUE)){
		NeedCheckFIFO[bChannelNumber] = FALSE;
		/* check FIFO not empty */
		if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) {/* SATA */
	 
			U32		templdata1,templdata2;
			U8		i;
			U32		ldata;
	  
			#ifdef _MMIO_
			ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEMiscInternalControlState, templdata1 );
			#else
			ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDEMiscInternalControlState,templdata1);
			#endif
	  
			ReadIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_Ctrl_Status, templdata2); 
	  
			if ( (( templdata1&0x0F000000) != 0x0F000000) | (( templdata2&0x05000000) != 0x05000000) ) {
				/* do IDE_COMMAND_ATAPI_RESET to clear FIFO.*/
				for (i=0;i<gChnlDrvControlConfig[bChannelNumber].Support_device;i++) {
					if ( gDrvConfig[bChannelNumber][i].DevFlag & DEV_EXIST ) {
						ReadPortUlong( HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG, ldata );
						ldata >>= (bChannelNumber%MAX_ATA_CHNL);
						if(ldata & 0x100){
							if(gAdapterConfig[Adapter_ID].TapeOutVersion < REVISION_2){/* ASIC tapout prior revision 2 */
								if (CAMInitDevice(&(gDrvConfig[bChannelNumber][i])) == camFAIL) {
									gDrvConfig[bChannelNumber][i].DevFlag = DEV_NOTFOUND;
									cam_SetStatusByPlug_unplug(bChannelNumber, i, NO_DRIVE);
								}
							}
							else{
								#ifdef _MMIO_
								ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEMiscControlReg, templdata1 );
								#else
								ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDEMiscControlReg,templdata1);
								#endif
							
								templdata1 |= 0x80000000; /* bit 31, reset */
							
								#ifdef _MMIO_
								WriteIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEMiscControlReg, templdata1 );
								#else
								WriteIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDEMiscControlReg,templdata1);
								#endif

								#ifdef _MMIO_
								ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEMiscControlReg, templdata1 );
								#else
								ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDEMiscControlReg,templdata1);
								#endif
							
								templdata1 &= ~0x80000000; /* bit 31, reset */
							
								#ifdef _MMIO_
								WriteIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEMiscControlReg, templdata1 );
								#else
								WriteIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDEMiscControlReg,templdata1);
								#endif
							}
						}
						else{
							gDrvConfig[bChannelNumber][i].DevFlag = DEV_NOTFOUND;
							cam_SetStatusByPlug_unplug(bChannelNumber, i, NO_DRIVE);						
						}
						break;
					}
				}
			}
		}/* end of Ch_type == SATA */		
	}/* if (ATAModuleBusy[bChannelNumber]==TRUE) */

	if (ATAQueueType(bChannelNumber) & PSEUDO_TAG_CMD) {
		PATA_TASK pATATask;
		
#ifdef CHNL_BUSY_RECEIVE_TASK
		/* can Receive task? */
		if (ATAModuleReceiveTask[bChannelNumber] != TRUE)
			return;
#endif		


		CAM_ATAPacket=ATAQueueHead[bChannelNumber];
		pATATask=(PATA_TASK)CAM_ATAPacket->PacketHead.PTA;

		/* NCQ */	
		if ( gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_NCQ)
			StartATANCQOperation(bChannelNumber);
		else 
			StartATAQUEUEDOperation(bChannelNumber);
		return;
	}

	if (ATAModuleBusy[bChannelNumber]==TRUE)
		return;
			
	/* get free seqid */
	FreeSEQID=CAMAllocSEQID(Adapter_ID);	

	if (FreeSEQID==U8NULL){
		return;
	}

	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA)
		PacketCount = prepareSATAPacketSEQID(bChannelNumber,FreeSEQID);
	else
		PacketCount = preparePATAPacketSEQID(bChannelNumber,FreeSEQID);
		
	if ( PacketCount== ZERO_NULL)	/* no packet */
	{
		CAMFreeSEQID(Adapter_ID,FreeSEQID);	
		return;		
	}

	ATAModuleBusy[bChannelNumber]=TRUE;
	ATAModuleSEQID[bChannelNumber]=FreeSEQID;

	/* if swap box command, not send command to device , only get the priority*/
	SeqIDType = SeqIDQueueFirstATATaskType(Adapter_ID, FreeSEQID);
	
	if (SeqIDType & SWAPBOX_CMD){ /* Is SWAPBOX_CMD at SeqID queue */
		HandleInternalSwapbox_CMD(bChannelNumber, FreeSEQID);
	}
	else{		

#ifdef PIORW			
		/* set ATA Module timeout timer. */
		if ( !(((PPACKET_HEAD)gSEQ_INFO[Adapter_ID][FreeSEQID].SEQID2PacketHead)->Flag & PACFLAG_ATAPI) ) {
			if (!(((PPACKET_HEAD)gSEQ_INFO[Adapter_ID][FreeSEQID].SEQID2PacketHead)->Flag & PACFLAG_ATA_PIO)) {/* non ata PIO data transfer */
				if ( !(((PPACKET_HEAD)gSEQ_INFO[Adapter_ID][FreeSEQID].SEQID2PacketHead)->Flag & PACFLAG_FLUSH) )
					ATA_Timeout_TimerID[bChannelNumber]=camTimer(Adapter_ID,CAMTimeoutTimes[Adapter_ID].ATACommand,ATA_Timeout_Function[bChannelNumber]);
				else {
					((PPACKET_HEAD)gSEQ_INFO[Adapter_ID][FreeSEQID].SEQID2PacketHead)->Flag &= ~PACFLAG_FLUSH;
					ATA_Timeout_TimerID[bChannelNumber]=camTimer(Adapter_ID,(U32)(18*30),ATA_Timeout_Function[bChannelNumber]);
				}
			}

			if (((PPACKET_HEAD)gSEQ_INFO[Adapter_ID][FreeSEQID].SEQID2PacketHead)->Flag & PACFLAG_ATA_PIO) /* PIO data transfer */ {
				/* set Sequence Counter Control Register */
				WritePortUlong(HostBaseAddr[Adapter_ID], FreeSEQID*4, SEQ_FPDMA|0);				
			}
			else {
				/* set Sequence Counter Control Register */
				WritePortUlong(HostBaseAddr[Adapter_ID], FreeSEQID*4, PacketCount);
			}
			
			/* check not busy */
			if (!((ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus)) & IDE_STATUS_BUSY)) {
				SYNC_CPU_CACHE();

				if (((PPACKET_HEAD)gSEQ_INFO[Adapter_ID][FreeSEQID].SEQID2PacketHead)->Flag & PACFLAG_ATA_PIO) /* PIO data transfer */
				{

					/* When issue ATA PIO Write to HD, HD will update FIS status without INT,
					, the package can't close without INT, so must use direct command. */

					if (IssueATA_PIOCmd(bChannelNumber, FreeSEQID) == TRUE) { /* command done */
						ReturnATA_PIOCmd(bChannelNumber, FreeSEQID,0);
					}
					else {/* command doing or fail */
						U8 Status=ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);;
						
						if ( (!(Status & IDE_STATUS_BUSY)) && (Status & IDE_STATUS_ERROR) ) {
							ReturnATA_PIOCmd(bChannelNumber, FreeSEQID, ATA_ERROR);
						}
						else
							ATA_Timeout_TimerID[bChannelNumber]=camTimer(Adapter_ID,CAMTimeoutTimes[Adapter_ID].ATACommand,ATA_Timeout_Function[bChannelNumber]);						
					}
				
				}
				else {
					
					/* start ata module */
					WritePortUlong(ATABaseAddr[bChannelNumber], ATA_PACSTART_REGISTER_OFFSET, (U32)((PPACKET_HEAD)gSEQ_INFO[Adapter_ID][FreeSEQID].SEQID2PacketHead)->Address);

				}
					
			}
#else
		/* set ATA Module timeout timer. */
		if ( !(((PPACKET_HEAD)gSEQ_INFO[Adapter_ID][FreeSEQID].SEQID2PacketHead)->Flag & PACFLAG_ATAPI) ) {
			if ( !(((PPACKET_HEAD)gSEQ_INFO[Adapter_ID][FreeSEQID].SEQID2PacketHead)->Flag & PACFLAG_FLUSH) )
				ATA_Timeout_TimerID[bChannelNumber]=camTimer(Adapter_ID,CAMTimeoutTimes[Adapter_ID].ATACommand,ATA_Timeout_Function[bChannelNumber]);
			else {
				((PPACKET_HEAD)gSEQ_INFO[Adapter_ID][FreeSEQID].SEQID2PacketHead)->Flag &= ~PACFLAG_FLUSH;
				ATA_Timeout_TimerID[bChannelNumber]=camTimer(Adapter_ID,(U32)(18*30),ATA_Timeout_Function[bChannelNumber]);
			}
			
			/* set Sequence Counter Control Register */
			WritePortUlong(HostBaseAddr[Adapter_ID], FreeSEQID*4, PacketCount);

			/* check not busy */
			if (!((ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus)) & IDE_STATUS_BUSY)) {
				SYNC_CPU_CACHE();
				
				/* start ata module */
				WritePortUlong(ATABaseAddr[bChannelNumber], ATA_PACSTART_REGISTER_OFFSET, (U32)((PPACKET_HEAD)gSEQ_INFO[Adapter_ID][FreeSEQID].SEQID2PacketHead)->Address);
			}
#endif
		}
		else {
			/* ATAPI devices need special care */
			IssueATAPIPktCmd(bChannelNumber, FreeSEQID);
			ATA_Timeout_TimerID[bChannelNumber]=camTimer(Adapter_ID,CAMTimeoutTimes[Adapter_ID].ATAPICommand,ATA_Timeout_Function[bChannelNumber]);
			((PPACKET_HEAD)gSEQ_INFO[Adapter_ID][FreeSEQID].SEQID2PacketHead)->Flag &= ~PACFLAG_ATAPI; /* clear this */
			
			if (((PPACKET_HEAD)gSEQ_INFO[Adapter_ID][FreeSEQID].SEQID2PacketHead)->Flag & PACFLAG_ATAPI_PIO) /* PIO data transfer */ {
				/* set Sequence Counter Control Register */
				WritePortUlong(HostBaseAddr[Adapter_ID], FreeSEQID*4, SEQ_FPDMA|0);				
			}
			else {
				/* set Sequence Counter Control Register */
				WritePortUlong(HostBaseAddr[Adapter_ID], FreeSEQID*4, PacketCount);
			}
			SYNC_CPU_CACHE();
			
			/* start ata module */
			WritePortUlong(ATABaseAddr[bChannelNumber], ATA_PACSTART_REGISTER_OFFSET, (U32)((PPACKET_HEAD)gSEQ_INFO[Adapter_ID][FreeSEQID].SEQID2PacketHead)->Address);
		}
		
		if ((gBOXConfig[bChannelNumber].BOXType != camNOBOX) && (gBOXConfig[bChannelNumber].BOXType & SATAEnclosure)) {
			/* turn on LED */
			CAM_SBOX_HDAccess_LED(bChannelNumber , HDActiveON);
		}	
	}
		
	return;
}

void prepareATANCQCommand(PCAM_PACKET CAM_ATAPacket,PATA_TASK pATATask)
{
	PU8		pAtrl;
	U32 	startingSector;

	/* set bATRL */

	pAtrl = &(CAM_ATAPacket->ATAPacket.bATRL[0]);	
			
	/* select drive */
	*pAtrl++=ATRLREG_REG_SIZE_1|ATRLREG_DEVICESELECT;				
	*pAtrl++ = (U8) (0xE0);
	
	/*
	*pAtrl++=ATRLREG_REG_SIZE_1|
				ATRLREG_CONDITION_CHECK_NOT_BUSY|
				ATRLREG_DEVICESELECT;
	*pAtrl++ = (U8) (0xE0);	/ * twice, to make sure drive is ready * /
	*/
				
	*pAtrl++=ATRLREG_REG_SIZE_1|
				ATRLREG_X_3F6;
	*pAtrl++=0x08; 	/* enable interrupt */

	/* previous content */
    	*pAtrl++ = ATRLREG_REG_SIZE_1|
     				ATRLREG_FEATURE;
	*pAtrl++ = (U8)(pATATask->Cmd.ATACmd.bCount >> 8); 		/* SecCnt 15:8 */
	
	*pAtrl++ = ATRLREG_REG_SIZE_1|
    				ATRLREG_SECTOR_COUNT;
	*pAtrl++ = 0;										/* Reserved */

	startingSector=pATATask->Cmd.ATACmd.LBAl;			
 	*pAtrl++ = ATRLREG_REG_SIZE_1|
    			ATRLREG_LBA0_7 ; 
	*pAtrl++ = (U8)( startingSector >> 24 );	/* LBA 31:24 */

	startingSector=pATATask->Cmd.ATACmd.LBAh;
 	*pAtrl++ = ATRLREG_REG_SIZE_1|
     					ATRLREG_LBA8_15;   
	*pAtrl++ = (U8)( startingSector);	/* LBA 39:32 */

	*pAtrl++ = ATRLREG_REG_SIZE_1|
     				ATRLREG_LBA16_23 ;
	*pAtrl++ = (U8)( startingSector >> 8 );	/* LBA 47:40 */   	
 	
 	/* most recently written */
	*pAtrl++ = ATRLREG_REG_SIZE_1|
     				ATRLREG_FEATURE;
	*pAtrl++ = (U8)pATATask->Cmd.ATACmd.bCount; 			/* SecCnt 7:0 */
	 	
	*pAtrl++ = ATRLREG_REG_SIZE_1|
     				ATRLREG_SECTOR_COUNT;
	*pAtrl++ = 0;											/* no TAG now. */
	
	startingSector=pATATask->Cmd.ATACmd.LBAl;
	*pAtrl++ = ATRLREG_REG_SIZE_1|
     				ATRLREG_LBA0_7 ; 
	*pAtrl++ = (U8)( startingSector );   /* LBA 7:0 */ 
    
	*pAtrl++ = ATRLREG_REG_SIZE_1|
     				ATRLREG_LBA8_15;   
	*pAtrl++ = (U8)( startingSector >> 8 );	/* LBA 15:8 */

	*pAtrl++ = ATRLREG_REG_SIZE_1|
     				ATRLREG_LBA16_23;
	*pAtrl++ = (U8)( startingSector >> 16 );	/* LBA 23:16 */				


	*pAtrl++=	ATRLREG_REG_SIZE_1|
				ATRLREG_DEVICESELECT;				
	
	if (pATATask->ATACmdFlag & FORCE_MEDIA)	
		*pAtrl++ = (U8) (0xE0); /* set FUA bit */ /* lba bit must be set */
	else		
		*pAtrl++ = (U8) (0x60); /* lba bit must be set */

	*pAtrl++ = ATRLREG_REG_SIZE_1|
     				ATRLREG_COMMAND |
     				ATRLREG_X_END_1F7;
	*pAtrl++ = pATATask->Cmd.ATACmd.bCmd2Chip;  
      	   
	return;
}

void prepareATAQUEUEDCommand(PCAM_PACKET CAM_ATAPacket,PATA_TASK pATATask)
{
      	   
	return;
}

void prepareATACommand(PCAM_PACKET CAM_ATAPacket,PATA_TASK pATATask)
{
	PU8		pAtrl;
	U32 	startingSector;
	U8 		LargeLBA = checkLargeLBA(pATATask);

	/* set bATRL */

	pAtrl = &(CAM_ATAPacket->ATAPacket.bATRL[0]);	
			
	/* select drive */
	*pAtrl++=ATRLREG_REG_SIZE_1|ATRLREG_DEVICESELECT;		
	
	if (gChnlConfig[pATATask->bCh2Chip].ChnlType == CHNL_SATA) /* SATA */		
		*pAtrl++ = (U8) (0xE0);
	else
		*pAtrl++ = (U8) (0xE0 | (pATATask->bID << 4));
		
	/*
	*pAtrl++=ATRLREG_REG_SIZE_1|
				ATRLREG_CONDITION_CHECK_NOT_BUSY|
				ATRLREG_DEVICESELECT;
	if (gChnlConfig[pATATask->bCh2Chip].ChnlType == CHNL_SATA) / * SATA * /		
		*pAtrl++ = (U8) (0xE0);
	else				
		*pAtrl++ = (U8) (0xE0 | (pATATask->bID << 4));	/ * twice, to make sure drive is ready * /
	*/			
	
	*pAtrl++=ATRLREG_REG_SIZE_1|
				ATRLREG_X_3F6;
	*pAtrl++=0x08; 	/* enable interrupt */

	*pAtrl++ = ATRLREG_REG_SIZE_1|
     				ATRLREG_FEATURE;
	*pAtrl++ = pATATask->Cmd.ATACmd.bFeature; 
	
	if ( LargeLBA == TRUE )
	{
		/* previous content */
		*pAtrl++ = ATRLREG_REG_SIZE_1|
     				ATRLREG_SECTOR_COUNT;
		*pAtrl++ =(U8)(pATATask->Cmd.ATACmd.bCount >> 8); 	/* SecCnt 15:8 */

		startingSector=pATATask->Cmd.ATACmd.LBAl;			
		*pAtrl++ = ATRLREG_REG_SIZE_1|
     				ATRLREG_LBA0_7 ; 
		*pAtrl++ = (U8)( startingSector >> 24 );/* LBA 31:24 */

		startingSector=pATATask->Cmd.ATACmd.LBAh;
		*pAtrl++ = ATRLREG_REG_SIZE_1|
     					ATRLREG_LBA8_15;   
		*pAtrl++ = (U8)( startingSector);	/* LBA 39:32 */

		*pAtrl++ = ATRLREG_REG_SIZE_1|
     				ATRLREG_LBA16_23 ;
		*pAtrl++ = (U8)( startingSector >> 8 ); /* LBA 47:40 */
	}
 	/* most recently written */
 	
	*pAtrl++ = ATRLREG_REG_SIZE_1|
     				ATRLREG_SECTOR_COUNT;
	*pAtrl++ = (U8)pATATask->Cmd.ATACmd.bCount; /* SecCnt 7:0 */
	
	startingSector=pATATask->Cmd.ATACmd.LBAl;
	*pAtrl++ = ATRLREG_REG_SIZE_1|
     				ATRLREG_LBA0_7 ; 
	*pAtrl++ = (U8)( startingSector );   /* LBA 7:0 */
    
	*pAtrl++ = ATRLREG_REG_SIZE_1|
     				ATRLREG_LBA8_15;   
	*pAtrl++ = (U8)( startingSector >> 8 );	/* LBA 15:8 */

	*pAtrl++ = ATRLREG_REG_SIZE_1|
     				ATRLREG_LBA16_23;
	*pAtrl++ = (U8)( startingSector >> 16 );	/* LBA 23:16 */	
				
	*pAtrl++=	ATRLREG_REG_SIZE_1|
				ATRLREG_DEVICESELECT;				

	if (gChnlConfig[pATATask->bCh2Chip].ChnlType == CHNL_SATA) {/* SATA */
		if ( LargeLBA != TRUE )
			*pAtrl++ = (U8) (((U8)(startingSector >> 24) & 0x0F) | 0xE0); /* LBA 27:24 */
		else
			*pAtrl++ = (U8) (0xE0); /* lba bit must be set */	
	}
	else {
		if ( LargeLBA != TRUE )
			*pAtrl++ = (U8) (((U8)(startingSector >> 24) & 0x0F) | 0xE0 | (pATATask->bID << 4)); /* LBA 27:24 */
		else
			*pAtrl++ = (U8) (0xE0 | (pATATask->bID << 4)); /* lba bit must be set */		
	}
					
	*pAtrl++ = ATRLREG_REG_SIZE_1|
     				ATRLREG_COMMAND |
     				ATRLREG_X_END_1F7;
	*pAtrl++ = pATATask->Cmd.ATACmd.bCmd2Chip;  
      	   
	return;
}

void prepareATAPICommand(PCAM_PACKET CAM_ATAPacket,PATA_TASK pATATask)
{
	PU8          pAtrl;

	/* set bATRL */

	pAtrl = &(CAM_ATAPacket->ATAPacket.bATRL[0]);	

	/* send 12 bytes(6 words) command datas */	
	*pAtrl++=ATRLREG_REG_SIZE_6|
				ATRLREG_DATA|
				ATRLREG_X_END_1F0;
	
	*pAtrl++=pATATask->Cmd.Cdb[0];
	*pAtrl++=pATATask->Cmd.Cdb[1];

	*pAtrl++=pATATask->Cmd.Cdb[2];
	*pAtrl++=pATATask->Cmd.Cdb[3];

	*pAtrl++=pATATask->Cmd.Cdb[4];
	*pAtrl++=pATATask->Cmd.Cdb[5];

	*pAtrl++=pATATask->Cmd.Cdb[6];
	*pAtrl++=pATATask->Cmd.Cdb[7];

	*pAtrl++=pATATask->Cmd.Cdb[8];
	*pAtrl++=pATATask->Cmd.Cdb[9];

	*pAtrl++=pATATask->Cmd.Cdb[10];
	*pAtrl++=pATATask->Cmd.Cdb[11];
	
	return;
}

U8 checkDoTAGCommand(PATA_TASK pATATask)
{
	/*
	1. check is ATA device.
	2. check only SATA device.
	3. device support NCQ/Queued commands.
	4. command of pATATask if NCQ/Queued command 
		or set PSEUDO_TAG_CMD of ATACmdFlag.
	*/
	
	/* 1. check is ATA device. */
	if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_ATA_DEVICE)
	{
		/* 2. check only SATA device. */
		if (gChnlConfig[pATATask->bCh2Chip].ChnlType == CHNL_SATA) {
			/* 3. device support NCQ/Queued commands. */
			if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & (DEV_NCQ|DEV_QUEUED) ){
				
				if((gAdapterConfig[(pATATask->bCh2Chip)/MAX_ATA_CHNL].NoNCQAccess == TRUE)
					&& (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_NCQ)){
					return (FALSE);
				}
	
				/* 4. command of pATATask if NCQ/Queued command or set PSEUDO_TAG_CMD of ATACmdFlag. */
				if (pATATask->ATACmdFlag & PSEUDO_TAG_CMD) 
					return (TRUE);
				else {
					switch(pATATask->Cmd.ATACmd.bCmd2Chip){
						/* NCQ commands */
						case IDE_COMMAND_READ_FPDMA_QUEUED:
						case IDE_COMMAND_WRITE_FPDMA_QUEUED:
							return (TRUE);
							break;
						/* QUEUED commands */
						case IDE_COMMAND_READ_DMA_QUEUED:
						case IDE_COMMAND_READ_DMA_QUEUED_EXT:
						case IDE_COMMAND_WRITE_DMA_QUEUED:
						case IDE_COMMAND_WRITE_DMA_QUEUED_EXT:
						case IDE_COMMAND_WRITE_DMA_QUEUED_FUA_EXT:
							return (TRUE);
							break;					
					}
				}
			}
		}
	}
	return (FALSE);
}

void prepareCamPacket(PCAM_PACKET CAM_Packet,PCAM_ATA_PACKET CAM_ATAPacket,PCAM_ATA_SG CAM_ATASG,PATA_TASK pATATask)
{


	CAM_Packet->PacketHead.NPA		=	CAM_ATAPacket->PacketHead.NPA;
	CAM_Packet->PacketHead.PTA		=	CAM_ATAPacket->PacketHead.PTA;
	CAM_Packet->PacketHead.ModelType	=	ATA_PAC;
	CAM_Packet->PacketHead.Flag		=	CAM_ATAPacket->PacketHead.Flag;
	CAM_Packet->PacketHead.CountStamp	=	CAM_ATAPacket->PacketHead.CountStamp;

	CAM_Packet->PacketHead.PSG		=	(PVOID)CAM_ATASG;

	CAM_Packet->ATAPacket.bControl = 0;
	

	/* set PM Port */
	CAM_Packet->ATAPacket.bPMPort = pATATask->bID;
	
	/* set packet bControl */
	/* check is a data transfer */
	if (pATATask->ATACmdFlag & camDATA_XFER)
	{
		/* if PIO transfer command, send it as packet command, and at ISR(), must read 
		    all needed datas by PIO transfer. So set this to nondata transfer packet.
		*/
		/* PIO data transfer */
		if (pATATask->ATACmdFlag & PIO_XFER){
			CAM_Packet->ATAPacket.bControl |= ATACTRL_NONDATA;
		}/* end of PIO data transfer */
		else{		/* DMA data transfer */
					
#ifdef Encryption
		if(gAdapterConfig[pATATask->bCh2Chip/MAX_ATA_CHNL].Features & WITH_EncryKEY){
			if ((pATATask->ATACmdFlag & ATA_DEVICE) && (pATATask->ATACmdFlag & DATA_ENCRYPT))
				CAM_Packet->ATAPacket.bControl |= ATACTRL_ENCRYPT_ENABLE;
		}
#endif

			/* check is doing NCQ/QUEUED commands? */
			if (checkDoTAGCommand(pATATask) == TRUE) {
				/* device support NCQ */
				if ( gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_NCQ) { 
					CAM_Packet->ATAPacket.bControl |= ATACTRL_NONDATA;
					CAM_Packet->ATAPacket.bControl |= NCQFPDMA_CMD;
	
					/* set DMA R/W */
					if (pATATask->ATACmdFlag & camDATA_IN)	
						CAM_Packet->ATAPacket.bControl |= ATACTRL_DMAREAD;
					else 
						CAM_Packet->ATAPacket.bControl &= ~ATACTRL_DMAREAD;
						
				}
				/* end of device support NCQ */
			} /* end of check is doing NCQ/QUEUED commands? */
			else { /* normal ATA Packet Command */
				/* set DMA R/W */
				if (pATATask->ATACmdFlag & camDATA_IN)	
					CAM_Packet->ATAPacket.bControl |= ATACTRL_DMAREAD;
				else
					CAM_Packet->ATAPacket.bControl &= ~ATACTRL_DMAREAD;
			}
		}
	} /* end of check is a data transfer */
	else
	{
		CAM_Packet->ATAPacket.bControl |= ATACTRL_NONDATA;
	}
	
	CAM_Packet->ATAPacket.bSynSeqID = SEQ_NO_SEQID;	/* no SEQID now. */
	CAM_Packet->ATAPacket.bDelaySeqID = SEQ_NO_DELAY;	
	CAM_Packet->ATAPacket.ulNCA = ZERO_NULL;

	/* set ulpSG. */
	if (pATATask->SGCount){	/* have SG table */
		checkConvertSG((PCAMSG)pATATask->pSG,(PCAMSG)CAM_ATASG);
		CAM_Packet->ATAPacket.ulPSG=CamByteSwap32((U32)CAM_ATASG->Address);
	}
	else{	/* no SG table */
		CAM_Packet->ATAPacket.ulPSG=ZERO_NULL;
	}


	/* set bATRL */

	/* check is ATA or not? */
	if (pATATask->ATACmdFlag & ATA_DEVICE){		/* ATA device */
	
		/* check PRD table, last element <= (32 Dword), fix ASIC bug */
		if ( CAM_Packet->ATAPacket.ulPSG!= ZERO_NULL ){
			PCAMSG pSG = (PCAMSG)pATATask->pSG;
			PCAMSG pSGphy = (PCAMSG)CAM_ATASG;
			U16		tmpwCount;
			while(1){
				if(pSG->CtrlFlag & EOT)
					break;
				pSG++;
				pSGphy++;
			}
			if ( !pSG->wCount || (pSG->wCount > SG_COUNT_ASIC_BUG) ){
				tmpwCount = (U16)(pSG->wCount - SG_COUNT_ASIC_BUG);
								
				pSGphy->CtrlFlag &= ~EOT;
				pSGphy->wCount = CamByteSwap16((U16)tmpwCount);
								
				pSGphy++;
				pSGphy->StartAddr = CamByteSwap32((U32)(pSG->StartAddr + tmpwCount));
				pSGphy->wCount = CamByteSwap16((U16)SG_COUNT_ASIC_BUG);
				pSGphy->Reserved = 0;
				pSGphy->CtrlFlag = EOT;
			}
		}

		/* check is doing NCQ/QUEUED commands? */
		if (checkDoTAGCommand(pATATask) == TRUE) {
			/* device support NCQ */
			if ( gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_NCQ) {
				prepareATANCQCommand(CAM_Packet,pATATask);				
			}
			else
			{
				prepareATAQUEUEDCommand(CAM_Packet,pATATask);
/*======================================================================================*/
#ifdef Encryption
			if(gAdapterConfig[pATATask->bCh2Chip/MAX_ATA_CHNL].Features & WITH_EncryKEY){
				if ((pATATask->ATACmdFlag & ATA_DEVICE) && (pATATask->ATACmdFlag & DATA_ENCRYPT)){
					if(TCQ_Encrypt_Flag[pATATask->bCh2Chip] == FALSE){
						WriteIndexUchar(ATA_ExpBaseAddr[pATATask->bCh2Chip], iNDEXeXP_FPDMA_Ext_Ctrl, 0x02 );
						TCQ_Encrypt_Flag[pATATask->bCh2Chip] = TRUE;
					}
				}
				else{
					if(TCQ_Encrypt_Flag[pATATask->bCh2Chip] == TRUE){
						WriteIndexUchar(ATA_ExpBaseAddr[pATATask->bCh2Chip], iNDEXeXP_FPDMA_Ext_Ctrl, 0x00 );
						TCQ_Encrypt_Flag[pATATask->bCh2Chip] = FALSE;	
					}
				}
			}
#endif
/*======================================================================================*/
			}
		}
		else			
			prepareATACommand(CAM_Packet,pATATask);
	}
	else if (!(pATATask->ATACmdFlag & SWAPBOX_CMD)) {	/* ATAPI device */
		prepareATAPICommand(CAM_Packet, pATATask);
		if ( CAM_Packet->ATAPacket.ulPSG != ZERO_NULL ){
			/* for Big-Endian */	
			U32	TotalBytes = 0 ;
			PCAMSG pSG = (PCAMSG)pATATask->pSG;
			PCAMSG pSGphy = (PCAMSG)CAM_ATASG;
			U16		tmpwCount;
			while(1){
				TotalBytes += pSG->wCount;
				if(pSG->CtrlFlag & EOT)
					break;
				pSG++;
				pSGphy++;
			}
			if ( TotalBytes & 0x03 ){
				pSGphy->CtrlFlag &= ~EOT;
				pSGphy++;
				pSGphy->StartAddr = CamByteSwap32((U32)CAM_ATASG->Address + CAM_PRD_TBL_SIZE);
				pSGphy->wCount = CamByteSwap16((U16)4 - (U16)(TotalBytes & 0x03));
				pSGphy->CtrlFlag = EOT;
			}
			/* check PRD table, last element <= (32 Dword), fix ASIC bug */
			else if ( !pSG->wCount || (pSG->wCount > SG_COUNT_ASIC_BUG) ){
				tmpwCount = (U16)(pSG->wCount - SG_COUNT_ASIC_BUG);
				
				pSGphy->CtrlFlag &= ~EOT;
				pSGphy->wCount = CamByteSwap16((U16)tmpwCount);
								
				pSGphy++;
				pSGphy->StartAddr = CamByteSwap32((U32)(pSG->StartAddr + tmpwCount));
				pSGphy->wCount = CamByteSwap16((U16)SG_COUNT_ASIC_BUG);
				pSGphy->Reserved = 0;
				pSGphy->CtrlFlag = EOT;
			}
		}
	}
	
	return;
}


void prepareATAPacket(PCAM_ATA_PACKET CAM_ATAPacket, PATA_TASK pATATask)
{


	CAM_ATAPacket->PacketHead.NPA=(PVOID)PTR_NULL;
	CAM_ATAPacket->PacketHead.PTA=pATATask;



	
	/* set packet bControl */
	/* check is a data transfer */
	if (pATATask->ATACmdFlag & camDATA_XFER)
	{
		/* if PIO transfer command, send it as packet command, and at ISR(), must read 
		    all needed datas by PIO transfer. So set this to nondata transfer packet.
		*/
		/* PIO data transfer */
		if (pATATask->ATACmdFlag & PIO_XFER){		
			

			/* check FORCE_MEDIA */	/* ATAPI not support FORCE_MEDIA */
			if (pATATask->ATACmdFlag & ATA_DEVICE)
			{		
				/* check is need set FORCE_MEDIA  */
				switch(pATATask->Cmd.ATACmd.bCmd2Chip){
					case IDE_COMMAND_WRITE_MULTIPLE_FUA_EXT:
						pATATask->ATACmdFlag |= FORCE_MEDIA;					
						break;
					default:
						break;
				}			
			}
							
			/* check is ATA and with PSEUDO_CMD? */	/* ATAPI not support seudo CMD */
			if ( (pATATask->ATACmdFlag & ATA_DEVICE)
				&& (pATATask->ATACmdFlag & PSEUDO_CMD))
			{		
				/* check is need change command? */
				switch(pATATask->Cmd.ATACmd.bCmd2Chip){
					case IDE_COMMAND_READ_MULTIPLE:
						if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_48BIT)
							pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_READ_MULTIPLE_EXT;
						break;
					case IDE_COMMAND_READ_SECTORS:
						if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_48BIT)
							pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_READ_SECTORS_EXT;
							break;
					
					case IDE_COMMAND_WRITE_MULTIPLE:
					case IDE_COMMAND_WRITE_MULTIPLE_EXT:
					case IDE_COMMAND_WRITE_MULTIPLE_FUA_EXT:
						if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_48BIT) {
							if (pATATask->ATACmdFlag & FORCE_MEDIA)
								pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_WRITE_MULTIPLE_FUA_EXT;	
							else
								pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_WRITE_MULTIPLE_EXT;	
						}				
					case IDE_COMMAND_WRITE_SECTORS:
						if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_48BIT)
							pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_WRITE_SECTORS_EXT;
						break;					
					default:
						break;
				}
			}
		}
		/* end of PIO data transfer */
		
		/* DMA data transfer */
		else{
			
			/* check is doing NCQ/QUEUED commands? */
			if (checkDoTAGCommand(pATATask) == TRUE) {
				/* device support NCQ */
				if ( gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_NCQ) { 
	
					/* check FORCE_MEDIA */	/* ATAPI not support FORCE_MEDIA */
					switch(pATATask->Cmd.ATACmd.bCmd2Chip){
						case IDE_COMMAND_WRITE_DMA_FUA_EXT: 
						case IDE_COMMAND_WRITE_MULTIPLE_FUA_EXT:
							pATATask->ATACmdFlag |= FORCE_MEDIA;					
							break;
						default:
							break;
					}			
										
					/* check with PSEUDO_CMD? */	/* ATAPI not support seudo CMD */
					if (pATATask->ATACmdFlag & PSEUDO_TAG_CMD)
					{
						if (pATATask->ATACmdFlag & camDATA_IN)
							pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_READ_FPDMA_QUEUED;
						else
							pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_WRITE_FPDMA_QUEUED;
						/*
						switch(pATATask->Cmd.ATACmd.bCmd2Chip){
							case IDE_COMMAND_READ_DMA:
							case IDE_COMMAND_READ_DMA_EXT:
							case IDE_COMMAND_READ_MULTIPLE:
							case IDE_COMMAND_READ_MULTIPLE_EXT:
							case IDE_COMMAND_READ_SECTORS:
							case IDE_COMMAND_READ_SECTORS_EXT:
								pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_READ_FPDMA_QUEUED;
								break;
						
							case IDE_COMMAND_WRITE_DMA:
							case IDE_COMMAND_WRITE_DMA_EXT:
							case IDE_COMMAND_WRITE_DMA_FUA_EXT:
							case IDE_COMMAND_WRITE_MULTIPLE:
							case IDE_COMMAND_WRITE_MULTIPLE_EXT:
							case IDE_COMMAND_WRITE_MULTIPLE_FUA_EXT:
							case IDE_COMMAND_WRITE_SECTORS:
							case IDE_COMMAND_WRITE_SECTORS_EXT:
								pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_WRITE_FPDMA_QUEUED;					
								break;
							default:
								break;
						}
						*/		
					}
				}
				/* end of device support NCQ */
				
				/* device support QUEUED command */
				else 
				{ 
												
					/* check FORCE_MEDIA */	/* ATAPI not support FORCE_MEDIA */
					switch(pATATask->Cmd.ATACmd.bCmd2Chip){
						case IDE_COMMAND_WRITE_DMA_QUEUED_FUA_EXT: 
							pATATask->ATACmdFlag |= FORCE_MEDIA;					
							break;
						default:
							break;
					}			

					/* check with PSEUDO_CMD? */	/* ATAPI not support seudo CMD */
					if (pATATask->ATACmdFlag & PSEUDO_TAG_CMD)
					{
						if (pATATask->ATACmdFlag & camDATA_IN) {
							if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_48BIT)
								pATATask->Cmd.ATACmd.bCmd2Chip =IDE_COMMAND_READ_DMA_QUEUED_EXT;
							else
								pATATask->Cmd.ATACmd.bCmd2Chip =IDE_COMMAND_READ_DMA_QUEUED;
						}
						else {
							if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_48BIT) {
								if (pATATask->ATACmdFlag & FORCE_MEDIA)
									pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_WRITE_DMA_QUEUED_FUA_EXT;	
								else
									pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_WRITE_DMA_QUEUED_EXT;	
							}
							else
								pATATask->Cmd.ATACmd.bCmd2Chip =IDE_COMMAND_WRITE_DMA_QUEUED;							
						}
					}					
				}
			} /* end of check is doing NCQ/QUEUED commands? */
			else { /* normal ATA Packet Command */
				/* check FORCE_MEDIA */	/* ATAPI not support FORCE_MEDIA */
				if (pATATask->ATACmdFlag & ATA_DEVICE)
				{		
					/* check is need set FORCE_MEDIA  */
					switch(pATATask->Cmd.ATACmd.bCmd2Chip){
						case IDE_COMMAND_WRITE_DMA_FUA_EXT: 
						case IDE_COMMAND_WRITE_MULTIPLE_FUA_EXT:
							pATATask->ATACmdFlag |= FORCE_MEDIA;					
							break;
						default:
							break;
					}			
				}
							
				/* check is ATA and with PSEUDO_CMD? */	/* ATAPI not support seudo CMD */
				if ( (pATATask->ATACmdFlag & ATA_DEVICE)
					&& (pATATask->ATACmdFlag & PSEUDO_CMD))
				{
					if (pATATask->ATACmdFlag & camDATA_IN) {
						if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_48BIT)
							pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_READ_DMA_EXT;
						else
							pATATask->Cmd.ATACmd.bCmd2Chip =IDE_COMMAND_READ_DMA;
					}
					else {
						if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_48BIT) {
							if (pATATask->ATACmdFlag & FORCE_MEDIA)
								pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_WRITE_DMA_FUA_EXT;	
							else
								pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_WRITE_DMA_EXT;	
						}
						else
							pATATask->Cmd.ATACmd.bCmd2Chip =IDE_COMMAND_WRITE_DMA;						
					}
					
					/* 		
					switch(pATATask->Cmd.ATACmd.bCmd2Chip){
						case IDE_COMMAND_READ_DMA:
						case IDE_COMMAND_READ_DMA_EXT:
						case IDE_COMMAND_READ_MULTIPLE:
						case IDE_COMMAND_READ_MULTIPLE_EXT:
						case IDE_COMMAND_READ_SECTORS:
						case IDE_COMMAND_READ_SECTORS_EXT:						
							if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_48BIT)
								pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_READ_DMA_EXT;
							else
								pATATask->Cmd.ATACmd.bCmd2Chip =IDE_COMMAND_READ_DMA;
							break;
					
						case IDE_COMMAND_WRITE_DMA:
						case IDE_COMMAND_WRITE_DMA_EXT:
						case IDE_COMMAND_WRITE_DMA_FUA_EXT:
						case IDE_COMMAND_WRITE_MULTIPLE:
						case IDE_COMMAND_WRITE_MULTIPLE_EXT:
						case IDE_COMMAND_WRITE_MULTIPLE_FUA_EXT:
						case IDE_COMMAND_WRITE_SECTORS:
						case IDE_COMMAND_WRITE_SECTORS_EXT:
							if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_48BIT) {
								if (pATATask->ATACmdFlag & FORCE_MEDIA)
									pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_WRITE_DMA_FUA_EXT;	
								else
									pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_WRITE_DMA_EXT;	
							}
							else
								pATATask->Cmd.ATACmd.bCmd2Chip =IDE_COMMAND_WRITE_DMA;				
							break;
						default:
							break;
					}	
					*/		
				}
			}
		}
	} /* end of check is a data transfer */
	else
	{
		/* check is ATA and with PSEUDO_CMD? */	/* ATAPI not support seudo CMD */
		if ( (pATATask->ATACmdFlag & ATA_DEVICE)
			&& (pATATask->ATACmdFlag & PSEUDO_CMD))
		{		
			/* check is need change command? */
			switch(pATATask->Cmd.ATACmd.bCmd2Chip){
				case IDE_COMMAND_READ_VERIFY_SECTORS:
					if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_48BIT)
						pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_READ_VERIFY_SECTORS_EXT;
					break;
				case IDE_COMMAND_FLUSH_CACHE:
					CAM_ATAPacket->PacketHead.Flag |= PACFLAG_FLUSH;
					if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_48BIT)
						pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_FLUSH_CACHE_EXT;
					break;
				case IDE_COMMAND_READ_NATIVE_MAX_ADDRESS:
					if (gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_48BIT)
						pATATask->Cmd.ATACmd.bCmd2Chip = IDE_COMMAND_READ_NATIVE_MAX_ADDRESS_EXT;
					break;										
				default:
					break;
			}			
		}
		
		if (pATATask->ATACmdFlag & ATA_DEVICE)
		{		
			/* check is need change command? */
			switch(pATATask->Cmd.ATACmd.bCmd2Chip){
				case IDE_COMMAND_SMART_CMD:
				 	if ( gDrvConfig[pATATask->bCh2Chip][pATATask->bID].Features & DEV_SMART ) {
						if ( pATATask->Cmd.ATACmd.bFeature == 0xD8 )
							gDrvConfig[pATATask->bCh2Chip][pATATask->bID].Features |= DEV_EN_SMART;
						if ( pATATask->Cmd.ATACmd.bFeature == 0xD9 )
							gDrvConfig[pATATask->bCh2Chip][pATATask->bID].Features &= ~DEV_EN_SMART;
					}
					break;
				case IDE_COMMAND_SET_FEATURES:
				 	if ( gDrvConfig[pATATask->bCh2Chip][pATATask->bID].Features & DEV_WRITECACHE ) {
						if ( pATATask->Cmd.ATACmd.bFeature == 0x02 )
							gDrvConfig[pATATask->bCh2Chip][pATATask->bID].Features |= DEV_EN_WRITECACHE;
						if ( pATATask->Cmd.ATACmd.bFeature == 0x82 )
							gDrvConfig[pATATask->bCh2Chip][pATATask->bID].Features &= ~DEV_EN_WRITECACHE;
					}
					break;
				default:
					break;
			}			
		}
				
	}

	/* check is ATA or not? */
	if (pATATask->ATACmdFlag & ATA_DEVICE){		/* ATA device */
		/* check is doing NCQ/QUEUED commands? */
		if (checkDoTAGCommand(pATATask) == TRUE) {
			/* device support NCQ */
			if ( gDrvConfig[pATATask->bCh2Chip][pATATask->bID].DevFlag & DEV_NCQ)
				CAM_ATAPacket->PacketHead.Flag |= PACFLAG_NCQ;
			else
				CAM_ATAPacket->PacketHead.Flag |= PACFLAG_QUEUED;
		}
#ifdef PIORW		
		else if ( (pATATask->ATACmdFlag & PIO_XFER) && (pATATask->ATACmdFlag & camDATA_XFER)){ /* PIO transfer */
			CAM_ATAPacket->PacketHead.Flag |= PACFLAG_ATA_PIO;
		}
#endif		
	}
	else if (!(pATATask->ATACmdFlag & SWAPBOX_CMD)) {	/* ATAPI device */
		CAM_ATAPacket->PacketHead.Flag |= PACFLAG_ATAPI;
		if (pATATask->ATACmdFlag & PIO_XFER){		/* PIO transfer */
			CAM_ATAPacket->PacketHead.Flag |= PACFLAG_ATAPI_PIO;
		}		
	}
	
	return;
}

		
U8 CAM_SubmitATA (PATA_TASK pATATask)
{
	U8 Adapter_ID;

	PCAM_ATA_PACKET CAM_ATAPacket;

	Adapter_ID = pATATask->bCh/MAX_ATA_CHNL;
	pATATask->Cmd.ATACmd.bCmd2Chip = pATATask->Cmd.ATACmd.bCmd;	
	pATATask->bCh2Chip = CAM_ATA_Chnl_InMAP(pATATask->bCh);
	
#ifndef CHNL_BUSY_RECEIVE_TASK
	/* can Receive task? */
	if (ATAModuleReceiveTask[pATATask->bCh2Chip] != TRUE) {
		return camBUSY;
	}
#endif			
			
	/* get free packet address */
	CAM_ATAPacket = (PCAM_ATA_PACKET)CAMAllocATAPacket(pATATask->bCh2Chip);

	if (CAM_ATAPacket==(PVOID)PTR_NULL){	/* no free ata packet. */
		return camBUSY;		
	}	

	/* prevent conflict for TCQ, NCQ */
	if(!(pATATask->ATACmdFlag & camDATA_XFER)){
		if(pATATask->ATACmdFlag & PSEUDO_TAG_CMD)
			pATATask->ATACmdFlag &= ~PSEUDO_TAG_CMD;
	}

	/* prepare ata packet */
	prepareATAPacket(CAM_ATAPacket, pATATask);

/*
#ifdef _ERR_INJECT_
	if(ErrInj_Buffer_Table[pATATask->bCh2Chip].count)
		CheckErrInjection(CAM_ATAPacket, pATATask);
#endif
*/
	
	/* put atapacket to ata queue */
	Put2ATAQueue(CAM_ATAPacket,QUEUETail);

	/* start ata module */
	StartATAOperation(pATATask->bCh2Chip);

	return camACCEPTED;
}


U8 WaitOnBusy(PTR_SIZE IDEBaseAddress)
{ 
	U32	i;
	U8	Status=0;
	
	for (i=0; i<300000; i++) {
		Status = ReadIndexUchar(IDEBaseAddress, iNDEXIDEAlternateStatus);
      	if ( Status & IDE_STATUS_BUSY) {
			camUStallExecution(1);
			continue;
       	} else {
			break;
   		}
	}
	return(Status);
}

U8 WaitOnBaseBusy(PTR_SIZE IDEBaseAddress)
{ 
	U32 i;
	U8	Status=0;
	
	for (i=0; i<300000; i++) {
		Status = ReadIndexUchar(IDEBaseAddress, iNDEXIDEStatus);
		if ( Status & IDE_STATUS_BUSY) {
			camUStallExecution(1);
			continue;
       	} else {
			break;
   		}
	}
	return(Status);	
}

void AtapiSoftReset(U8 bChannelNumber, U8 DeviceNumber)
{
	PTR_SIZE IDEBaseAddress=ATABaseAddr[bChannelNumber];
	U8 i;
	U8 Status;

	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */	
		WriteIndexUchar(IDEBaseAddress, iNDEXIDEDeviceHead, 0xE0);
	else
		WriteIndexUchar(IDEBaseAddress, iNDEXIDEDeviceHead, (U8)(((DeviceNumber & 0x01) << 4) | 0xA0));
	WriteIndexUchar(IDEBaseAddress, iNDEXIDEDeviceControl, 0x02);
	camStallExecution(1);
	WriteIndexUchar(IDEBaseAddress, iNDEXIDECommand, IDE_COMMAND_ATAPI_RESET);
	camStallExecution(1);
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */		
		WriteIndexUchar(IDEBaseAddress, iNDEXIDEDeviceHead, 0xE0);
	else	
		WriteIndexUchar(IDEBaseAddress, iNDEXIDEDeviceHead, (U8)(((DeviceNumber & 0x01) << 4) | 0xA0));
	for ( i = 0; i < 10 ; i++ ) {
		/* wait 30 seconds, ATAPI Soft Reset Command and Protocol */
		Status=WaitOnBaseBusy(IDEBaseAddress);
		if (!(Status & IDE_STATUS_BUSY))
			break;
	}
	camStallExecution(1);
	ReadIndexUchar(IDEBaseAddress, iNDEXIDEStatus);
	WriteIndexUchar(IDEBaseAddress, iNDEXIDEDeviceControl, 0x08 );
	return;
}

U8 AtaSoftReset_PM(U8 bChannelNumber, U8 DeviceNumber)
{
	PTR_SIZE IDEBaseAddress=ATABaseAddr[bChannelNumber];
	U8 i;
	U8 Status=0;

	WriteIndexUchar(IDEBaseAddress, iNDEXIDEDeviceHead, 0xE0);
	WriteIndexUchar(IDEBaseAddress, iNDEXIDEDeviceControl, 0x04);
	camUStallExecution(10);
	WriteIndexUchar(IDEBaseAddress, iNDEXIDEDeviceControl, 0x08 );
	camStallExecution(2);
	for ( i = 0; i < 10 ; i++ ) {
		Status = ReadIndexUchar(IDEBaseAddress, iNDEXIDEStatus);
		if (!(Status & IDE_STATUS_BUSY))
			break;
		camUStallExecution(1);	
	}
	
	if (Status & IDE_STATUS_BUSY)
		return(camFAIL);
	else
		return(camSUCCESS);
}

U8 AtaSoftReset(U8 bChannelNumber, U8 DeviceNumber)
{
	PTR_SIZE IDEBaseAddress=ATABaseAddr[bChannelNumber];
	U8 i;
	U8 Status;

	Status = ReadIndexUchar(IDEBaseAddress, iNDEXIDEAlternateStatus);	
	if ( Status & IDE_STATUS_BUSY) 
		return(camFAIL);
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) /* SATA */		
		WriteIndexUchar(IDEBaseAddress, iNDEXIDEDeviceHead, 0xE0);
	else	
		WriteIndexUchar(IDEBaseAddress, iNDEXIDEDeviceHead, (U8)(((DeviceNumber & 0x01) << 4) | 0xA0));
	WriteIndexUchar(IDEBaseAddress, iNDEXIDEDeviceControl, 0x04);
	camUStallExecution(10);
	WriteIndexUchar(IDEBaseAddress, iNDEXIDEDeviceControl, 0x08 );
	camStallExecution(2);
	for ( i = 0; i < 10 ; i++ ) {
		/* check not bust , MAX wait 30 seconds */
		Status=WaitOnBaseBusy(IDEBaseAddress);
		if (!(Status & IDE_STATUS_BUSY))
			break;		
	}
	
	if (Status & IDE_STATUS_BUSY)
		return(camFAIL);
	else
		return(camSUCCESS);
}

U8 WaitForDrq(PTR_SIZE IDEBaseAddress) 
{ 
	U32 i; 
	U32 j;
	U8	Status=0;
	
	/* wait 3 seconds */
	for (i=0; i<300000; i++) { 
		for (j=0; j<50; j++) { 
			GetStatus(IDEBaseAddress, Status); 
			if (Status==0xFF)
				break;
			else if (Status & IDE_STATUS_DRQ) { 
				break; 
			} else if (Status & IDE_STATUS_ERROR) { 
				break; 
			}
		}
		if (Status==0xFF)
			break;
		if (Status & IDE_STATUS_BUSY) { 
			camUStallExecution(1); 
		} else if (Status & IDE_STATUS_DRQ) { 
			break; 
		} else if (Status & IDE_STATUS_ERROR) { 
			break; 
		} else { 
			camUStallExecution(1); 
		}		
	} 
	return( Status );
}

U8 WaitOnReady(U8 bChannelNumber)
{ 
	U32 i;
	U8	status;
	PTR_SIZE IDEBaseAddress=ATABaseAddr[bChannelNumber];

	/* wait 3 seconds */
	for (i=0; i<300000; i++) {
		status = ReadIndexUchar(IDEBaseAddress, iNDEXIDEStatus);
 		status &= ~(IDE_STATUS_ERROR | IDE_STATUS_INDEX | IDE_STATUS_DRQ );
       	if (!(status & IDE_STATUS_BUSY))
       		break;
       	#if 0	/* only for Octopus */
		if ((status == 0x7F) && (i > 0)) {
			if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA)
				return(camWARNING); /* no device */
		}
		#endif
			
 		camUStallExecution(1);
	}

	if ( i >=300000 )
		return( camFAIL ); /* time-out */
			
	return(camSUCCESS);	
}

/* delay 300000 == wait 3 seconds */
U8 WaitDelayOnReady(U8 bChannelNumber, U32 delay)
{ 
	U32 i;
	U8	status;
	PTR_SIZE IDEBaseAddress=ATABaseAddr[bChannelNumber];
	
	for (i=0; i<delay; i++) {
		status = ReadIndexUchar(IDEBaseAddress, iNDEXIDEStatus);
 		status &= ~(IDE_STATUS_ERROR | IDE_STATUS_INDEX | IDE_STATUS_DRQ );
       	if (!(status & IDE_STATUS_BUSY))
       		break;
       	#if 0	/* only for Octopus */
		if ((status == 0x7F) && (i > 0)) {
			if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA)
				return(camWARNING); /* no device */
		}
		#endif
			
 		camUStallExecution(1);
	}

	if ( i >= delay )
		return( camFAIL ); /* time-out */
			
	return(camSUCCESS);	
}

U8 WaitForINT(PTR_SIZE IDEBaseAddress) 
{ 
	U32 i; 
	U32 j; 
	U8	status=0;
	
	/* wait 30 sec */
	for (i=0; i<1000000; i++) {
		for (j=0; j<50; j++) { 
			status = ReadIndexUchar(IDEBaseAddress, iNDEXIDECICR+3);
			if (status & 0x08)
				break;
		}
		if (status & 0x08)
			break;
		camUStallExecution(3);	
	}
	if ( i >=1000000 )
		return( camFAIL ); /* time-out */
	
	/* clear drive INT */
	status = ReadIndexUchar(IDEBaseAddress, iNDEXIDEStatus);
    status &= ~(IDE_STATUS_ERROR | IDE_STATUS_INDEX | IDE_STATUS_DRQ );
	if ( (status != IDE_STATUS_IDLE) && (status != IDE_STATUS_DRDY) )
		return(camFAIL); /* bad device */	
			
	return( camSUCCESS );
}

U8 WaitForATAPIINT(U8 bChannelNumber) 
{ 
	U32 i; 
	U8	status;
	PTR_SIZE	IDEBaseAddress = ATABaseAddr[bChannelNumber];
	
	/* wait 30 sec */
	for (i=0; i<1000000; i++) { 
		status = ReadIndexUchar(IDEBaseAddress, iNDEXIDECICR+3);
		if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA ) {
			if (status & 0x08)
				break;			
		}
		else {
			if (status & 0x04)
				break;			
		}

		camUStallExecution(3); 
	}
	if ( i >=1000000 )
		return( camFAIL ); /* time-out */
	
	/* clear drive INT */
	status = ReadIndexUchar(IDEBaseAddress, iNDEXIDEStatus);
	return( camSUCCESS );
}

#ifdef _MMIO_
U8 ReadBuffer16(U8 bChannelNumber, PU16 Buffer, U32 Count)
{
	U32 i, j, tempdata32;
	
	for (i = 0; i< Count; i++ ) {
		if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA ){
			for(j=0; j<10; j++){
				ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEMiscInternalControlState, tempdata32 );
				tempdata32 >>= 24;
				if(!(tempdata32 & bit(2)))
					break;
			}
			if(j >= 10)
				return camFAIL;
		}

		*(Buffer+i) = camReadRegWord(ATABaseAddr[bChannelNumber]);
	}
	
	return camSUCCESS;
}
	
U8 WriteBuffer16(U8 bChannelNumber, PU16 Buffer, U32 Count)
{
	U32 i, j, tempdata32;

	for (i = 0; i< Count; i++ ) {
		if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA ){
			for(j=0; j<10; j++){
				ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEMiscInternalControlState, tempdata32 );
				tempdata32 >>= 24;
				if(tempdata32 & bit(3))
					break;
			}
			if(j >= 10)
				return camFAIL;
		}	

		camWriteRegWord(ATABaseAddr[bChannelNumber], *(Buffer+i));
	}
	
	return camSUCCESS;;
}

#else	
U8 ReadBuffer16(U8 bChannelNumber, PU16 Buffer, U32 Count)
{
	U32 i, j, tempdata32;
	
	if (gChnlConfig[bChannelNumber].ChnlType != CHNL_SATA ){
		camOutPortByte( ATABaseAddr[bChannelNumber]+ INDEX_REGISTER_OFFSET, iNDEXIDEData);
		for (i = 0; i< Count; i++ ) {
			*(Buffer+i) = camInPortWord( ATABaseAddr[bChannelNumber]+ DATA_REGISTER_OFFSET);
		}
	}
	else{
		for (i = 0; i< Count; i++ ) {
			for(j=0; j<10; j++){
				ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDEMiscInternalControlState,tempdata32);	
				tempdata32 >>= 24;
				if(!(tempdata32 & bit(2)))
					break;
			}
			if(j >= 10)
				return camFAIL;
	
			camOutPortByte( ATABaseAddr[bChannelNumber]+ INDEX_REGISTER_OFFSET, iNDEXIDEData);
			*(Buffer+i) = camInPortWord( ATABaseAddr[bChannelNumber]+ DATA_REGISTER_OFFSET);
		}
	}
	
	return camSUCCESS;
}
		
U8 WriteBuffer16(U8 bChannelNumber, PU16 Buffer, U32 Count)
{
	U32 i, j, tempdata32;
	
	
	if (gChnlConfig[bChannelNumber].ChnlType != CHNL_SATA ){
		camOutPortByte( ATABaseAddr[bChannelNumber]+ INDEX_REGISTER_OFFSET, iNDEXIDEData);
		for (i = 0; i< Count; i++ ) {
			camOutPortWord( ATABaseAddr[bChannelNumber]+ DATA_REGISTER_OFFSET, *(Buffer+i));
		}
	}
	else{
		for (i = 0; i< Count; i++ ) {
			for(j=0; j<10; j++){
				ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDEMiscInternalControlState,tempdata32);	
				tempdata32 >>= 24;
				if(tempdata32 & bit(3))
					break;
			}
			if(j >= 10)
				return camFAIL;
		
			camOutPortByte( ATABaseAddr[bChannelNumber]+ INDEX_REGISTER_OFFSET, iNDEXIDEData);
			camOutPortWord( ATABaseAddr[bChannelNumber]+ DATA_REGISTER_OFFSET, *(Buffer+i));
		}
	}
	
	return camSUCCESS;
}
#endif

#ifdef DWordPIODataPort

/* NOT use by CAM */
/* CAM use Command Packet to send ATA/ATAPI command, if use this feature 
	of DW data transfer, all command must re-write to direct-command if 
	the command access data port 1F0. So NOT implement DW Data Port for 
	PIO Read Access into CAM. Just implement DW Data Port for PIO Write Access.
*/
#if 0
U8 ReadBuffer32(U8 bChannelNumber, PU32 Buffer, U32 Count)
{
	U32 i, j, tempdata32;
	U32 ldata;
	U16 wdata;
	U8 returnValue=camSUCCESS;

	/* set PIO Dword access */
	#ifdef _MMIO_
	ReadIndexWord( ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEErrorsReportedfromLinkLayer, wdata );
	#else
	ReadIndexWord(ATABaseAddr[bChannelNumber], iNDEXIDEErrorsReportedfromLinkLayer,wdata);
	#endif
					
	wdata |=bit(0); 
		
	#ifdef _MMIO_
	WriteIndexWord(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL),iNDEXIDEErrorsReportedfromLinkLayer,wdata);
	#else
	WriteIndexWord(ATABaseAddr[bChannelNumber], iNDEXIDEErrorsReportedfromLinkLayer, wdata);
	#endif
			

	if (gChnlConfig[bChannelNumber].ChnlType != CHNL_SATA ){
		for (i = 0; i< (Count/2); i++ ) {
			#ifdef _MMIO_
			ReadIndexUlong( ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEDWordDataPortforPIOAccess, *(Buffer+i) );
			#else
			ReadIndexUlong( ATABaseAddr[bChannelNumber], iNDEXIDEDWordDataPortforPIOAccess, *(Buffer+i) );
			#endif
		}
		if ((i*2) != Count) { /* not Dword biundary */
			#ifdef _MMIO_
			ReadIndexUlong( ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEDWordDataPortforPIOAccess, ldata );
			#else
			ReadIndexUlong( ATABaseAddr[bChannelNumber], iNDEXIDEDWordDataPortforPIOAccess, ldata );
			#endif
			*((PU16)(Buffer+i)) = (U16)ldata;
		}		
	}
	else{
		for (i = 0; i< (Count/2); i++ ) {
			for(j=0; j<10; j++){
	
				#ifdef _MMIO_
				ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL),iNDEXIDEMiscInternalControlState,tempdata32);
				#else
				ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDEMiscInternalControlState, tempdata32);
				#endif
	
				tempdata32 >>= 24;
				if(!(tempdata32 & bit(2)))
					break;
			}
			if(j >= 10){
				returnValue = camFAIL;
				break;
			}
	
			#ifdef _MMIO_
			ReadIndexUlong( ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEDWordDataPortforPIOAccess, *(Buffer+i) );
			#else
			ReadIndexUlong( ATABaseAddr[bChannelNumber], iNDEXIDEDWordDataPortforPIOAccess, *(Buffer+i) );
			#endif
		}
		if(returnValue != camFAIL){
			if ((i*2) != Count) { /* not Dword biundary */
				for(j=0; j<10; j++){
	
					#ifdef _MMIO_
					ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL),iNDEXIDEMiscInternalControlState,tempdata32);
					#else
					ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDEMiscInternalControlState, tempdata32);
					#endif
	
					tempdata32 >>= 24;
					if(!(tempdata32 & bit(2)))
						break;
				}
				if(j >= 10)
					returnValue = camFAIL;
				else{
					#ifdef _MMIO_
					ReadIndexUlong( ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEDWordDataPortforPIOAccess, ldata );
					#else
					ReadIndexUlong( ATABaseAddr[bChannelNumber], iNDEXIDEDWordDataPortforPIOAccess, ldata );
					#endif
					*((PU16)(Buffer+i)) = (U16)ldata;
				}
			}
		}
	}
		
	/* re-set to PIO Word access */
	#ifdef _MMIO_
	ReadIndexWord( ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEErrorsReportedfromLinkLayer, wdata );
	#else
	ReadIndexWord(ATABaseAddr[bChannelNumber], iNDEXIDEErrorsReportedfromLinkLayer,wdata);
	#endif
					
	wdata &= ~bit(0);
		
	#ifdef _MMIO_
	WriteIndexWord(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL),iNDEXIDEErrorsReportedfromLinkLayer,wdata);
	#else
	WriteIndexWord(ATABaseAddr[bChannelNumber], iNDEXIDEErrorsReportedfromLinkLayer, wdata);
	#endif
		
	return returnValue;
}
#endif
	
U8 WriteBuffer32(U8 bChannelNumber, PU32 Buffer, U32 Count)
{
	U32 i, j, tempdata32;
	U32 ldata;
	U16 wdata;
	U8 returnValue=camSUCCESS;

	/* set PIO Dword access */
	#ifdef _MMIO_
	ReadIndexWord( ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEErrorsReportedfromLinkLayer, wdata );
	#else
	ReadIndexWord(ATABaseAddr[bChannelNumber], iNDEXIDEErrorsReportedfromLinkLayer,wdata);
	#endif
					
	wdata |=bit(0); 
		
	#ifdef _MMIO_
	WriteIndexWord(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL),iNDEXIDEErrorsReportedfromLinkLayer,wdata);
	#else
	WriteIndexWord(ATABaseAddr[bChannelNumber], iNDEXIDEErrorsReportedfromLinkLayer, wdata);
	#endif
	

	if (gChnlConfig[bChannelNumber].ChnlType != CHNL_SATA ){
		for (i = 0; i< (Count/2); i++ ) {
			#ifdef _MMIO_
			WriteIndexUlong( ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEDWordDataPortforPIOAccess, *(Buffer+i) );
			#else
			WriteIndexUlong( ATABaseAddr[bChannelNumber], iNDEXIDEDWordDataPortforPIOAccess, *(Buffer+i) );
			#endif
		}
		if ((i*2) != Count) { /* not Dword biundary */
			ldata=(U32)(*((PU16)(Buffer+i)));
			#ifdef _MMIO_
			WriteIndexUlong( ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEDWordDataPortforPIOAccess, ldata );
			#else
			WriteIndexUlong( ATABaseAddr[bChannelNumber], iNDEXIDEDWordDataPortforPIOAccess, ldata );
			#endif
		}
	}
	else{
		for (i = 0; i< (Count/2); i++ ) {
			for(j=0; j<10; j++){
	
				#ifdef _MMIO_
				ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL),iNDEXIDEMiscInternalControlState,tempdata32);
				#else
				ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDEMiscInternalControlState, tempdata32);
				#endif
	
				tempdata32 >>= 24;
				if(tempdata32 & bit(3))
					break;
			}
			if(j >= 10){
				returnValue = camFAIL;
				break;
			}
	
			#ifdef _MMIO_
			WriteIndexUlong( ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEDWordDataPortforPIOAccess, *(Buffer+i) );
			#else
			WriteIndexUlong( ATABaseAddr[bChannelNumber], iNDEXIDEDWordDataPortforPIOAccess, *(Buffer+i) );
			#endif
		}
		if(returnValue != camFAIL){
			if ((i*2) != Count) { /* not Dword biundary */
				for(j=0; j<10; j++){
	
					#ifdef _MMIO_
					ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL),iNDEXIDEMiscInternalControlState,tempdata32);
					#else
					ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDEMiscInternalControlState, tempdata32);
					#endif
	
					tempdata32 >>= 24;
					if(tempdata32 & bit(3))
						break;
				}
				if(j >= 10)
					returnValue = camFAIL;
				else{
					ldata=(U32)(*((PU16)(Buffer+i)));
					#ifdef _MMIO_
					WriteIndexUlong( ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEDWordDataPortforPIOAccess, ldata );
					#else
					WriteIndexUlong( ATABaseAddr[bChannelNumber], iNDEXIDEDWordDataPortforPIOAccess, ldata );
					#endif
				}
			}	
		}
	}
	
	/* re-set to PIO Word access */
	#ifdef _MMIO_
	ReadIndexWord( ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDEErrorsReportedfromLinkLayer, wdata );
	#else
	ReadIndexWord(ATABaseAddr[bChannelNumber], iNDEXIDEErrorsReportedfromLinkLayer,wdata);
	#endif
					
	wdata &= ~bit(0);
		
	#ifdef _MMIO_
	WriteIndexWord(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL),iNDEXIDEErrorsReportedfromLinkLayer,wdata);
	#else
	WriteIndexWord(ATABaseAddr[bChannelNumber], iNDEXIDEErrorsReportedfromLinkLayer, wdata);
	#endif
				
	return returnValue;
}

#endif

U8 ReadBuffer(U8 bChannelNumber, PU16 Buffer, U32 Count)
{
	U8 returnValue=camSUCCESS;
/* NOT use by CAM */
/* CAM use Command Packet to send ATA/ATAPI command, if use this feature 
	of DW data transfer, all command must re-write to direct-command if 
	the command access data port 1F0. So NOT implement DW Data Port for 
	PIO Read Access into CAM. Just implement DW Data Port for PIO Write Access.
*/	
#if 0
	#ifdef DWordPIODataPort
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA )
		returnValue = ReadBuffer32(bChannelNumber, (PU32)Buffer, Count);
	else
		returnValue = ReadBuffer16(bChannelNumber, Buffer, Count);
	#else
	returnValue = ReadBuffer16(bChannelNumber, Buffer, Count);
	#endif
#else
	returnValue = ReadBuffer16(bChannelNumber, Buffer, Count);
#endif
	
#ifndef _ATAPI_
	ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus); /*clear possible device INT */
#endif
	
	return returnValue;
}
	
U8 WriteBuffer(U8 bChannelNumber, PU16 Buffer, U32 Count)
{
	U8 returnValue=camSUCCESS;
	
	#ifdef DWordPIODataPort
	if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA )
		returnValue = WriteBuffer32(bChannelNumber, (PU32)Buffer, Count);
	else
		returnValue = WriteBuffer16(bChannelNumber, Buffer, Count);	
	#else
	returnValue = WriteBuffer16(bChannelNumber, Buffer, Count);
	#endif
	
#ifndef _ATAPI_
	ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus); /*clear possible device INT */
#endif

	return returnValue;
}

U8 CAM_ATA_Chnl_InMAP(U8 bChannelNumber)
{
	U8 bCh=bChannelNumber%MAX_ATA_CHNL;
	U8 Adapter_ID=bChannelNumber/MAX_ATA_CHNL;
	
	if (gChnlConfig[(Adapter_ID*MAX_ATA_CHNL)+3].bCh == camCHNL_NOT_FOUND) /* need not map */
		return (bChannelNumber);
	else {
		switch(bCh){
			case 0:
				bCh=3;
				break;
			case 1:
				bCh=1;
				break;
			case 2:
				bCh=0;
				break;
			case 3:
				bCh=2;
				break;			
		}		
	}
	return ((U8)((Adapter_ID*MAX_ATA_CHNL)+bCh));
}

U8 CAM_ATA_Chnl_OutMAP(U8 bChannelNumber)
{
	U8 bCh=bChannelNumber%MAX_ATA_CHNL;
	U8 Adapter_ID=bChannelNumber/MAX_ATA_CHNL;
	
	if (gChnlConfig[(Adapter_ID*MAX_ATA_CHNL)+3].bCh == camCHNL_NOT_FOUND) /* need not map */
		return (bChannelNumber);
	else {
		switch(bCh){
			case 0:
				bCh=2;
				break;
			case 1:
				bCh=1;
				break;
			case 2:
				bCh=3;
				break;
			case 3:
				bCh=0;
				break;												
		}		
	}
	return ((U8)((Adapter_ID*MAX_ATA_CHNL)+bCh));
}

/*=============================Remark==========================================*/
#if 0
void CAM_ATA_TASK_Chnl_InMAP(PATA_TASK pATATask)
{
	U8 bCh=pATATask->bCh%MAX_ATA_CHNL;
	U8 Adapter_ID=pATATask->bCh/MAX_ATA_CHNL;
	
	if (gChnlConfig[(Adapter_ID*MAX_ATA_CHNL)+2].ChnlType == CHNL_ATA) /* need not map */
		return;
	else {
		switch(bCh){
			case 0:
				bCh=3;
				break;
			case 1:
				bCh=1;
				break;
			case 2:
				bCh=0;
				break;
			case 3:
				bCh=2;
				break;												
		}		
	}
	pATATask->bCh=(U8)((Adapter_ID*MAX_ATA_CHNL)+bCh);
	return;
}


PATA_TASK CAM_ATA_TASK_Chnl_OutMAP(PATA_TASK pATATask)
{
	U8 bCh=pATATask->bCh%MAX_ATA_CHNL;
	U8 Adapter_ID=pATATask->bCh/MAX_ATA_CHNL;

	
	if (gChnlConfig[(Adapter_ID*MAX_ATA_CHNL)+2].ChnlType == CHNL_ATA) /* need not map */
		return(pATATask);
	else {
		switch(bCh){
			case 0:
				bCh=2;
				break;
			case 1:
				bCh=1;
				break;
			case 2:
				bCh=3;
				break;
			case 3:
				bCh=0;
				break;			
		}		
	}
	pATATask->bCh=(U8)((Adapter_ID*MAX_ATA_CHNL)+bCh);
	return(pATATask);
}
#endif
/*=============================End Remark=======================================*/
